%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxc/spectrum-spec.sl
% Title:          Some special things for HP Risc PSL compiler
% Author:         Winfried Neun , ZIB Berlin
% Created:        9 November 1991
% Status:         Experimental
% Package:        Compiler
%
% (c) Copyright 1991, Konrad-Zuse Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% a special Linke to save the NOP instruction after final BA

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (when (not (evenp DeAllocCount)) (setq DeAllocCount (iadd1 DeAllocCount)))
  (cond ((FlagP FunctionName 'ForeignFunction)
              `((*ForeignLink ,FunctionName ,FunctionType ,NumberOfArguments)
            (*Exit 0)))
        ((get functionname 'exitopencode)
              `((*DeAlloc ,DeAllocCount)
                (*JCall ,FunctionName)))
        ((get functionname 'opencode)
              `((*Call ,FunctionName)
                (*exit ,DeAllocCount)))
        ((and (freep (wtimes2 4 dealloccount))
		(InternallyCallableP functionname))
              `((BL (InternalEntry ,functionname) (reg r0))
                (nop)))
	((InternallyCallableP functionname)
	      `((ldw (displacement (reg sr0) (reg st) -20) (reg returnaddr))
		(BL (InternalEntry ,functionname) (reg r0))
                (LDO (displacement2 (reg st)
		   ,(minus (plus2 32 (times2 4 dealloccount)))) (reg st))))
	((freep (wtimes2 4 dealloccount))
		`((*JCall ,functionname)))
	(t    `((LDO (displacement2 (reg 0) (plus -8000 (idloc ,FunctionName)))
			 (reg t2))
                (ldwx (compl S)
                         (indexed (reg t2) (reg sr0) (reg symfncstart))
                         (reg t3))
		(ldw (displacement (reg sr0) (reg st) -20) (reg returnaddr))
                (ble (displacement (reg sr5) (reg t3) 0))
                (LDO (displacement2 (reg st)
                   ,(minus (plus2 32 (times2 4 dealloccount)))) (reg st))))))

% change branch to exitlabel into *exit when framesize ne 0 too

(fluid '(exitt&))

(de &jumpopt (u)
    (when (equal (cadar u) exitt&) (rplaca u '(*exit (*framesize)))))

% turn off lapopt in faslend

(fluid '(*lapopt))

(setq *lapopt t)

(when (funboundp 'old-faslend) (copyd 'old-faslend 'faslend))

(de faslend ()
    (prog (*lapopt)
       (return (old-faslend))))

% open codes for cons ncons xcons

(defcmacro *SPEC-JumpIF)

(de *SPEC-JumpIF (Label ArgOne ArgTwo Instruction)
  (prog (ResultingCode*)
    (setq label (ResolveOperand '(Reg Error) Label))
    (return `((comb (compl <) ,ArgOne ,ArgTwo ,Label)))))

(put 'cons 'opencode
          '((addil 16#10000 (reg r10)) % item to reg r1
            (stwm (reg 1) (displacement (reg sr0) (reg r10) 4))
            (stwm (reg 2) (displacement (reg sr0) (reg r10) 4))
	    (*spec-jumpif (labelgen templabel) (reg r10) (reg r11))
            %%%(comb (compl <) (reg r10) (reg r11) (label templabel))
            (*move (reg r1) (reg 1))
            (*savefastframe)  %we are short of heap we have to prepare for gc
            (*call !%reclaim)
            (*restorefastframe)
          (labelref templabel)))

(put 'xcons 'opencode
          '((addil 16#10000 (reg r10)) % item to reg r1
            (stwm (reg 2) (displacement (reg sr0) (reg r10) 4))
            (stwm (reg 1) (displacement (reg sr0) (reg r10) 4))
            (*spec-jumpif (labelgen templabel) (reg r10) (reg r11))
            %%%(comb (compl <) (reg r10) (reg r11) (label templabel))
            (*move (reg r1) (reg 1))
            (*savefastframe)  %we are short of heap we have to prepare for gc
            (*call !%reclaim)
            (*restorefastframe)
          (labelref templabel)))

(put 'ncons 'opencode
          '((addil 16#10000 (reg r10)) % item to reg r1
            (stwm (reg 1) (displacement (reg sr0) (reg r10) 4))
            (stwm (reg nil) (displacement (reg sr0) (reg r10) 4))
            (*spec-jumpif (labelgen templabel) (reg r10) (reg r11))
            %%%(comb (compl <) (reg r10) (reg r11) (label templabel))
            (*move (reg r1) (reg 1))
            (*savefastframe)  %we are short of heap we have to prepare for gc
            (*call !%reclaim)
            (*restorefastframe)
          (labelref templabel)))

(defcmacro *savefastframe)

(de  *savefastframe ()
   (cond ((null *usefastframe) '((*xalloc 1) (*move (reg 1) (xframe 1)) ))
         (T (append `((*xalloc ,(wplus2 nalloc* 1)) (*move (reg 1) (xframe 1)))
                     (pnth '((*move (reg ff4) (xframe 5))
                             (*move (reg ff3) (xframe 4))
                             (*move (reg ff2) (xframe 3))
                             (*move (reg ff1) (xframe 2))
                             (*move (reg nil) (reg nil))) % dummy for reduce
                         (wdifference 5 nalloc*))))))

(defcmacro *restorefastframe)

(de  *restorefastframe ()
   (cond ((null *usefastframe) '((*move (xframe 1) (reg 1))  (*xdealloc 1)))
        (T (append   (pnth '((*move (xframe 5) (reg ff4))
                             (*move (xframe 4) (reg ff3))
                             (*move (xframe 3) (reg ff2))
                             (*move (xframe 2) (reg ff1))
                             (*move (reg nil) (reg nil))) % dummy for reduce
                         (wdifference 5 nalloc*))
           `((*move (xframe 1) (reg 1)) (*xdealloc ,(wplus2 nalloc* 1))))) ))


(commentoutcode
(put 'bittable 'opencode
       '((extu (reg 3) (reg 2) (bitfield 32 2)) % offset
         (ld!.bu (reg 5) (indexed (reg 1) (reg 3)))
         (mak (reg 4) (reg 2) (bitfield 2 1)) % lower two bits <<1
         (sub (reg 4) (reg 4) 6)        % - fraction
         (sub (reg 4) (reg 0) (reg 4))  %   fraction
         (extu (reg 1) (reg 5) (reg 4))
         (and (reg 1) (reg 1) 3)))

(put 'putbittable 'opencode
       '((extu (reg t3) (reg 2) (bitfield 32 2)) % offset
         (ld!.bu (reg 5) (indexed (reg 1) (reg t3)))
         (mak (reg 4) (reg 2) (bitfield 2 1)) % lower two bits <<1
         (sub (reg 4) (reg 4) 6)        % - fraction
         (sub (reg 4) (reg 0) (reg 4))  %   fraction
         (*move 3 (reg t2))
         (mak (reg t2) (reg t2) (reg 4))
         (and!.c (reg 5) (reg 5) (reg t2))
         (and (reg 3) (reg 3) 3)        % value & 3
         (mak (reg t1) (reg 3) (reg 4)) % <<fraction
         (or (reg 5) (reg t1) (reg 5))  % or
         (st!.b (reg 5) (indexed (reg 1) (reg t3)))
         (*move (reg 3) (reg 1))))

)

% This code is intended for filling the 'holes' in the frame.
% e.g. in the following function
%
%  (de hugo (x) (cond ((null x) (call-another))
%                     (t (prog (a b c) ....))
%
% a frame of length 4 will be allocated, but it will not be preset (by nil)
% completely at entering the function but the frame elements for a b and c
% will be set (to nil) if the prog is entered.
% So if the (null x) case is taken we have 3 unset frame elements on the stack
% which contain random information and this is dangerous for gc.
% to resolve that problem, the following function will work on the input
% to pass-1-lap and insert nil settings where needed.

(fluid '(&fillframeholes &frame-numbers))

(de &fillframeholes (code)
   (prog (allo cmactoextern externfound ll lll nn linearcode restcode)
      (when (not &fillframeholes) (return code))
      (setq cmactoextern
                  '(*link *linke *call *jcall *cons *xcons *fastmkfltn))
             % trivial tests whether we have to work at all
      (unless (setq allo (findcmacro code '*alloc)) (return code))
      (setq ll cmactoextern)
      (while (and (not externfound) ll)
            (setq externfound (findcmacro code (car ll)))
            (setq ll (cdr ll)))
      (when (not externfound) (return code))

            % so we have to work, starting at the *alloc

      (setq nn (cadar allo))  %the frmae size
      (setq linearcode t)
      (setq restcode (cdr allo))

      (setq &frame-numbers nil)

      (while (and linearcode restcode)
         (setq linearcode (lookatinstruction (car restcode)))
         (setq restcode (cdr restcode)))

      (when linearcode (return code))

      (setq lll (setq ll (cons nil nil)))

             % search frame that have not been set, prepare code,
             % put into place after *alloc

      (for (from i 1 nn 1) (do
            (unless (memq i &frame-numbers)
                  (rplaca ll `(*move (reg nil) (frame ,i)))
                  (push 'nil ll))))

      (when (eq lll ll) (return code))
      (rplacd lll (cdr allo))

      (rplacd allo (cdr ll))
      (return code)))

(de lookatinstruction (instr)
      (cond ((atom instr) t)
            ((and (eq (car instr) '*move)
                  (eqcar (caddr instr) 'frame))
                       (push (cadr (caddr instr)) &frame-numbers) 17)
            ((memq (car instr) '(*link *call *linke *jcall *jumpintype
                 *jumpeq *jump  *JumpType *JumpNotType *JumpNotInType
                 *jumpnoteq *jumpwlessp *jumpwgreaterp *jumpwleq *jumpwgeq
                 *jumpon)) nil)
            (t t)))


(de findcmacro (list pat)
     (cond ((atom list ) nil)
           ((eqcar (car list) pat) list)
           (t        (findcmacro (cdr list) pat ))))

(setq &fillframeholes t)
