%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:SPARC-CMAC.sl
% Title:          Patterns and predicates for Sparc PSL cmacro expansion
% Author:         Winfried Neun, using Chris Burdorf's sparc-cmacros
% Created:        December 15, 1988
% Modified:
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions
%  2. Jan 1990 (Winfried Neun)
%  Included immediatequote stuff
%  20 April 1989 (Winfried Neun)
%  Cleaning up, better car and cdr expansions
%  30 Jan 1989 (Winfried Neun)
%  Optimized loading of items from memory addresses in code (*movex)
%  Better *JUMPON now, using jmpl instruction
%  Revised alloc scheme in  order to make stack frames self explanatory for
%  gc and backtrace

(on fast-integers)

(compiletime
  (progn (on eolinstringok)
     (dm * (u)  nil)
     (setq *syslisp t)
     (load if-system)
     nil))

(fluid '(AddressingUnitsPerItem
     CharactersPerWord
     StackDirection
     *ImmediateQuote
     AddressingUnitsPerFunctionCell))

(loadtime (off ImmediateQuote))

(flag '(Immediate UnImmediate Indirect displacement predecrement
    label extrareg postincrement Indexed absolute deferred reglist)
    'TerminalOperand)

% MkItem may be used when evaluating WConst expressions.

(de MkItem (TagPart InfPart) % ZIB
  (lor (lsh TagPart 59) (land InfPart 16#7ffffff)))

(de freep (asize) (and (eq 0 asize) (not &links-present&)))

(de FiveP (x)  (eq x 5))

(de FiftyNinep (x)  (eq x 59))

(de evenp (x) (and (fixp x) (eq 0 (land x 1))))

(de indexedp (x) (eqcar x 'indexed))

(de DeferrableP (Expression)
  (and (PairP Expression)
       (not (memq (first Expression)
              '(Deferred $Fluid $Global Fluid Global)))))

(de RegisterDeferredP (Expression)
  (and (EqCar Expression 'Deferred) (EqCar (second Expression) 'Reg)))


(de FakeRegisterNumberP (Expression)
  (and (IntP Expression) (GreaterP Expression 5)))

% NegativeImmediateP is for numbers whose negative is a one byte operand.

(de negp (value) (and (numberp value) (lessp value 0)))

(de displacementp (x) (and (pairp x) (eq (car x) 'displacement)))

(de NegativeImmediateP (Expression)
  (and (FixP Expression) (minusp Expression) (GreaterP Expression -16#fffffff)))

% expansions of addressing modes.

(DefAnyreg CAR  ZIBanyregCAR)  % ZIB

(de ZIBAnyregCAR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. caar
               (member (cadr source) *nextreglist*))
             (setq areg (cadr source))

             (setq areg (nexttempreg))
	     (when (equal aReg reg)
                   (setq aReg (nexttempreg))))

      (if (Xregp source)  % its a reg or a frame reg
          (push `(andn ,source (reg i5) ,areg) Resultingcode*)

          (push `(*move ,source , areg) Resultingcode*)
          (push `(andn ,areg (reg i5) ,areg) Resultingcode*))
      (return `(displacement ,areg 0))))

(DefAnyreg CDR  ZIBanyregCDR)  % ZIB

(de ZIBAnyregCDR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. cddar
               (member (cadr source) *nextreglist*))
             (setq areg (cadr source))

             (setq areg (nexttempreg))
	     (when (equal aReg reg)
                   (setq aReg (nexttempreg))))
      (if (Xregp source)  % its a reg or a frame reg
          (push `(andn ,source (reg i5) ,areg) Resultingcode*)

          (push `(*move ,source , areg) Resultingcode*)
          (push `(andn ,areg (reg i5) ,areg) Resultingcode*))
      (return `(displacement ,areg 8))))

(fluid '(*nextreg* *nextreglist*))

(setq *nextreg* 0        % Four registers should be sufficient
      *nextreglist* '((reg t2) (reg t3) (reg t4) ))

(de nexttempreg (x)   % ZIB  :This saves Regs , non static assignment
     (setq *nextreg* (wplus2 *nextreg* 1))
     (when (eq *nextreg* 4) (setq *nextreg* 1))
     (nth *nextreglist* *nextreg*))

(DefAnyreg Memory  ZIBanyregMemory)  % ZIB

(de ZIBAnyregMemory (reg source offset)
  (prog (aReg cadrsource)
   (setq source (resolveoperand reg source))
   (setq offset (resolveoperand reg offset))
   (cond ((and (registerp source) (registerp offset))
             (return `(indexed ,source ,offset))) % SUN 4 has double index
         ((registerp source)
      (return (if (eq 0 offset) `(Deferred ,source)
                     (if (sparc-short-immediate-p offset)
			     `(displacement ,source ,offset)
			     (progn
			       (setq aReg (nexttempreg))
				(when (equal aReg reg)
					 (setq aReg (nexttempreg)))
			                      % take a new TempReg
			       (push `(*move ,offset ,aReg) Resultingcode*)
			      `(displacement ,source ,aReg) )))))
    % what shall that ?
      ((and (AddressConstantP source)(ZeroP offset))
	    (return  `(unimmediate ,SOURCE )))
%%%%%%%   ((and (DeferrableP source) (ZeroP offset))
%%%%%%%    (return `(Deferred ,SOURCE) ))
      ((and (RegisterDeferredP source )(ZeroP offset))
         (return `(deferred (Displacement (UnDeferred ,SOURCE) 0)))))
    (setq aReg (if (and (displacementp source)
                    (member (setq cadrsource (cadr source))
			          *nextreglist*))
               cadrsource      (nexttempreg))) % take a new TempReg
    (when (equal aReg reg) (setq aReg (nexttempreg)))
    (push `(*move ,source ,aReg) Resultingcode*)
    (return `(displacement ,aReg  ,offset))))


(DefAnyreg QUOTE
       AnyregQUOTE
       ((Null) (REG NIL))
       ((EqTP) (indexed (reg g7) -72)) %%FLUID T))
       ((InumP) SOURCE)
       (       (QUOTE SOURCE)))

(fluid '(&frame-regs&))

(setq &frame-regs& [ nil l1 l2 l3 l4 l5 l6 l7 i0 i1 i2 i3 i4 ])

(DefAnyreg FRAME ZIBAnyregFrame)  % ZIB

(de ZIBAnyregFRAME (dum number)
    (cond ((wgeq number 13)
             `(displacement (reg st) ,(plus2 2231 (times2 8 (difference number 13)))))
% bizarre calculation : 2239 - 8 (which is 1959 + 272) why??
      (t `(reg ,(indx &Frame-regs& number)))))

(DefAnyreg REG ZIBAnyregREG)

(de ZIBanyregReg (reg source)
   (cond ((Fakeregisternumberp source)
          `(indexed (reg g7) ,(times2 8 (plus2 source -6))))
     ((and (idp source) (get source 'viv-number)) % fast access var?
          `(indexed (reg g7) ,(times2 2 (get source 'viv-number))))
     (t       `(REG ,SOURCE))))

(loadtime (remflag '(fluid global $fluid $global) 'Terminaloperand))

(defanyreg Fluid ZIBanyregfluid)

(defanyreg $Fluid ZIBanyregfluid)

(fluid '(*syslisp ))

(de ZIBanyregFluid (reg source)
   (cond ((and (idp source) (get source 'viv-number) % fast access var?
         ) % (or *syslisp (flagp source '*user-variable))) % safety !!
                `(indexed (reg g7) ,(times2 2 (get source 'viv-number))))
     (t       `($Fluid ,SOURCE))))

(defanyreg $VivFluid ZIBanyregVivfluid)  % for usage for VIV in non syslisp
			              % e.g. lambind, cons ... etc
(de ZIBanyregVIVFluid (reg source)
  (let ((getviv (get source 'viv-number)))
     (cond (getviv  `(indexed (reg g7) ,(times2 2 getviv)))
       (t       `(*Cerror " Wrong VivFluid" ,SOURCE)))))

(defanyreg Global ZIBanyregglobal)

(defanyreg $Global ZIBanyregglobal)

(de ZIBanyregGlobal (reg source)
   (cond ((and (idp source) (get source 'viv-number) % fast access var?
           )% (or *syslisp (flagp source '*user-variable))) % safety !!
                `(indexed (reg g7) ,(times2 2 (get source 'viv-number))))
     (t       `($global ,SOURCE))))

%-----------------------------------------------------------------------------

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (if (FlagP FunctionName 'ForeignFunction)
      (list (list '*ForeignLink FunctionName FunctionType NumberOfArguments)
        '(*Exit 0))
      (list (list '*DeAlloc DeAllocCount)
        (list '*JCall FunctionName))))

(DefCMacro *Call
       ((InternallyCallableP)  (call (InternalEntry ArgOne))
			    (NOP))
        ( (*move  ($fluid symfnc) (reg t1))
          (*move (immediate (idloc ArgOne)) (reg t4))
          (sll (reg t4) 3 (reg t4))
          (*wplus2 (reg t1) (reg t4))
          (*move (displacement (reg t1) 0) (reg t4))
          (jmpl (displacement (reg t4) 0) (reg 15))
          (sethispec (immediate (idloc ArgOne)) (reg t3))))

(DefCMacro *JCall
       ((InternallyCallableP) (BA (InternalEntry ArgOne))
			     (nop))
        ( (*move  ($fluid symfnc) (reg t1))
          (*move (immediate(idloc ArgOne)) (reg t4))
          (sll (reg t4) 3 (reg t4))
          (*wplus2 (reg t1) (reg t4))
          (*move (displacement (reg t1) 0) (reg t4))
          (jmpl (displacement (reg t4) 0) (reg 0))
          (sethispec (immediate (idloc ArgOne)) (reg t3))
))


(de quote-fix-p (x)
    (and (eqcar x 'quote) (fixp (cadr x))))

(de displacement-p (x)
    (and (eqcar x 'displacement)
     (eqn (length x) 3) (sparc-register-p (cadr x))))

(de deferred-p (x)
    (and (pairp x) (or (sparc-register-p (car x))
                   (eq (car x) 'deferred))))

(remprop '*move 'optfn)

(DefCMacro *Move
       (Equal)
       ((Xregp Xregp)               (mov ArgOne argtwo))
       ((schreckenshugop Xregp)     (*schrecken ArgOne argtwo))
       ((indexedp Xregp)            (ldx ArgOne argtwo))
       ((Xregp indexedp)            (stx ArgOne argtwo))
       ((Xregp anyp)                (*movex ArgOne argtwo))
       ((anyp Xregp)                (*movex ArgOne argtwo))
       (			     (*move ArgOne (reg t2))
			             (*move (reg t2) Argtwo)))

% for gc: put a hskip item  into frame in correct length

(de schreckenshugop (x) (and (eqcar x 'immediate)
			  (pairp (cadr x))
			  (eqcar (cadr x) 'schreckenshugo)))

(de *schrecken (ArgOne argtwo)
    (setq ArgOne (cadr (cadr ArgOne))) % (immediate (schreckenshugo n)
    `((sethi ,(wor 16#360000 ArgOne) ,argtwo))) % its a forward tag

(defcmacro *schrecken)

(de *ALLOC (framesize)

    (setq NAlloc!* framesize)
 % use standard allocation scheme % ZIB
 % we have to leave 7 words unused for foreigncall!! see stack scheme in
 % Sparc architecture manual
    (cond ((freep) NIL)
      ((wgeq framesize 13)
        (if (evenp framesize)
        `((save (reg st) ,(minus (times 8 (plus2 14 framesize))) (reg st))
              (sethi 16#3E0000 (reg i5)) % frame 13
	      (sllx (reg i5) 32 (reg i5))
          (*move (schreckenshugo ,framesize) (reg l0)))
        `((save (reg st) ,(minus (times 8 (plus2 15 framesize))) (reg st))
              (sethi 16#3E0000 (reg i5)) % frame 13
	      (sllx (reg i5) 32 (reg i5))
          (*move (schreckenshugo ,framesize) (reg l0)))
       )) % must be a multiple of 8 !!
      (t `((save (reg st) ,-192 (reg st))
              (sethi 16#3E0000 (reg i5)) % frame 13
	      (sllx (reg i5) 32 (reg i5))
          (*move (schreckenshugo ,framesize) (reg l0))))
)   )

(defcmacro *ALLOC)

(DefCMacro *DeAlloc
       ((FreeP))
       (         (restore  (reg 0) ArgOne (reg 0))))

(DefCMacro *Exit
     % leaf routine first
       ((Freep)  (jmpl (displacement (reg 15) 8) (reg 0))
                 (nop))
       (         (jmpl (displacement (reg returnaddr) 8) (reg 0))
                 (restore (reg 0) 0 (reg 0))))

(de deferredp (x) (and (pairp x) (eq (car x) 'deferred)))

(DefCMacro *Jump
       ((deferredp)        (*move (undeferred ArgOne) (reg t5))
			    (jmpl (deferred (reg t5)) (reg 0))
			        (or (reg 0) 0 (reg 0))) % NO NOP!!!!! FOR LAPOPT
       ((displacementp)    (jmpl ArgOne (reg 0))
			    (nop))
       (                   (ba ArgOne)
			    (nop)))


(DefCMacro *Lbl           (ArgOne))

(de Xregp (x) (or (eqcar x 'Reg)
              (and (eqcar x 'frame) (leq (cadr x) 12))))

(de *3op (a1 a2 instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) A1)
                 (ResolveOperand '(Reg t2) A2)
                  Instruction)
                 (get '*3op 'CMacroPatternTable)))))

(DefCMacro *3op
       ((Xregp Xregp) (ARGTHREE ArgOne ARGtwo ArgOne))
       ((Xregp sparc-short-immediate-p)
                      (ARGTHREE ArgOne ARGTWO ArgOne))
       ((Xregp AnyP)  (*move argtwo (reg T5))
                      (*3op ArgOne (reg t5) argthree))
       ((Anyp Xregp)  (*move ArgOne (reg T5))
                      (ARGTHREE (reg t5) argtwo (reg t5))
                      (*move (reg t5) ArgOne))
       ((Anyp sparc-short-immediate-p)
                      (*move ArgOne (reg T6))
                      (argthree (reg t6) argtwo (reg t5))
                      (*move (reg t5) ArgOne))
       (              (*move argtwo ($fluid auxviv1))
		      (*move argone ($fluid auxviv0))
		      (save (reg st) -192 (reg st))
		      (*move ($fluid auxviv0) (reg T1))
                      (*move ($fluid auxviv1) (reg T2))
                      (argthree (reg t1) (reg t2) (reg i5))
		      (restore  (reg 0) 0 (reg 0))
                      (*move (reg o5) ArgOne)))

(DefCMacro *WPlus2
       ((*3op ArgOne Argtwo Add)))

(de *wplus3(arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wplus3))

(DefCMacro *WPlus3
           ((xregp xregp xregp) (add argthree argtwo ArgOne))
           ((xregp sparc-short-immediate-p xregp)
			         (add argthree argtwo ArgOne))
           ((xregp xregp sparc-short-immediate-p)
			         (add argtwo argthree ArgOne))
           ((xregp sparc-short-immediate-p anyp)
			        (*move argthree (reg t5))
			        (add (reg t5) argtwo ArgOne))
           ((xregp anyp sparc-short-immediate-p)
			        (*move argtwo (reg t5))
			        (add (reg t5) argthree ArgOne))
           ((xregp anyp anyp)  (*move argtwo ArgOne)
			        (*wplus2 ArgOne argthree))
           ((anyp xregp anyp)  (*move argthree (reg t5))
			        (*wplus2 (reg t5) argtwo)
			        (*move (reg t5) ArgOne))
           (                   (*move argtwo (reg t5))
			        (*move argthree (reg t6))
			        (add (reg t5) (reg t6) (reg t5))
			        (*move (reg t5) ArgOne)))

(DefCMacro *WDifference
       ((*3op ArgOne ARGTWO Sub)))

(de *wdifference3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wdifference3))

% format ArgOne = argtwo - argthree

(DefCMacro *Wdifference3
           ((xregp xregp xregp) (sub argtwo argthree ArgOne))
           ((xregp xregp sparc-short-immediate-p)
			        (sub argtwo argthree ArgOne))
           ((xregp anyp xregp) (*move argtwo (reg t5))
			        (sub (reg t5) argthree ArgOne))
           ((xregp anyp anyp)   (*move argtwo ($fluid auxviv1))
	                        (*move argthree ($fluid auxviv0))
				(save (reg st) -192 (reg st))
				(*move ($fluid auxviv1) (reg t1))
			        (*move ($fluid auxviv0) (reg t2))
			        (sub (reg t1) (reg t2) (reg i5))
				(restore  (reg 0) 0 (reg 0))
				(*move (reg o5) ArgOne))
           ((anyp xregp anyp)  (*move argthree (reg t5))
			        (sub  argtwo (reg t5) (reg t5))
			        (*move (reg t5) ArgOne))
           (                   (*move argtwo ($fluid auxviv1))
                                (*move argthree ($fluid auxviv0))
                                (save (reg st) -192 (reg st))
                                (*move ($fluid auxviv1) (reg t1))
                                (*move ($fluid auxviv0) (reg t2))
			        (sub (reg t1) (reg t2) (reg i5))
				(restore  (reg 0) 0 (reg 0))
			        (*move (reg o5) ArgOne)))

(defcmacro *wtimes2)

(de *wtimes2 (arg1 arg2) (*wtimes3 arg1 arg1 arg2))

(de *wtimes3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wtimes3))

% format ArgOne = argtwo * argthree

(defcmacro *wquotient)

(de *wquotient (arg1 arg2) (*wquotient3 arg1 arg1 arg2))

(de *wquotient3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wquotient3))

% format ArgOne = argtwo * argthree

(DefCMacro *wquotient3
	   ((anyp anyp onep)  (*move argtwo ArgOne))
	   ((xregp fixp fixp)  (*move (quotient ArgTwo ArgThree) ArgOne))
           ((xregp xregp powerof2p)
                    (*wquotientshift ArgOne argtwo argthree))
           (        (*move argtwo   ($fluid arithargs0))
                    (*move argthree ($fluid arithargs1))
                    (*move ($fluid arithargs0) (reg f0))
                    (*move ($fluid arithargs1) (reg f1))
                    (fxtod (reg f0) (reg f2))
                    (fitod (reg f1) (reg f4))
                    (fdivd (reg f2) (reg f4) (reg f6))
                    (fdtox (reg f6) (reg f7))
                    (stx (reg f7) (indexed (reg st) 72)) % this is unused
                    (*move (indexed (reg st) 72) ArgOne)))

(de *wquotientshift (arg1 arg2 arg3)
        (setq arg3 (powerof2p arg3))
        `((srlx ,arg2 ,(difference 64 arg3) (reg t5))
          (*wplus2 (reg t5) ,arg2)
          (srax (reg t5) ,arg3 ,arg1)))

(defcmacro *wquotientshift)

(DefCMacro *WAnd
       ((*3op ArgOne ARGTWO and)))

(DefCMacro *WOr
       ((*3op ArgOne ARGTWO or)))

(DefCMacro *WXOr
       ((*3op ArgOne ARGTWO xor)))

(de fixplusp (x) (and (fixp x) (not (minusp x))))

(DefCMacro *AShift
    ((AnyP ZeroP))
    ((Xregp negp)     (srax ArgOne (minus ARGTWO) ArgOne))
    ((Xregp fixplusp) (sllx ArgOne ARGTWO ArgOne))
    ((Xregp Xregp)    (subcc ARGTWO (reg 0) (reg 0))
                      (bl!,a templabel)
                      (sub (reg 0) ARGTWO (reg t6)) % delay slot
                      (sllx ArgOne ARGTWO ArgOne)
                      (ba templabel2)
                      (nop)
                     templabel
                      (srax ArgOne (reg t6) ArgOne)
                     templabel2)
    ((Xregp anyp)     (*move argtwo (reg t5))
                      (*ashift ArgOne (reg t5)))
    ((Anyp fixp)      (*move ArgOne (reg t5))
                      (*ashift (reg t5) argtwo)
			(*move (reg t5) ArgOne))
    (                 (*move ArgOne (reg t1))
                      (*ashift (reg t1) argtwo)
                      (*move (reg t1) ArgOne)))


(DefCMacro *WShift
        ((AnyP ZeroP))
    ((Xregp negp)     (srlx ArgOne (minus ARGTWO) ArgOne))
    ((Xregp fixplusp) (sllx ArgOne ARGTWO ArgOne))
    ((Xregp Xregp)    (subcc ARGTWO (reg 0) (reg 0))
                      (bl!,a templabel)
                      (sub (reg 0) ARGTWO (reg t6)) % delay slot
                      (sllx ArgOne ARGTWO ArgOne)
                      (ba templabel2)
                      (nop)
                     templabel
                      (srlx ArgOne (reg t6) ArgOne)
                     templabel2)
    ((Xregp anyp)     (*move argtwo (reg t5))
                      (*wshift ArgOne (reg t5)))
    ((Anyp fixp)      (*move ArgOne (reg t5))
                      (*wshift (reg t5) argtwo)
			(*move (reg t5) ArgOne))
    (                 (*move ArgOne (reg t1))
                      (*wshift (reg t1) argtwo)
                      (*move (reg t1) ArgOne)))

(DefCMacro *WNot
       ((Xregp Xregp) (Xnor ARGTWO (reg g0) ArgOne))
       ((Xregp Anyp)  (*move argtwo ArgOne)
                      (Xnor ArgOne (reg g0) ArgOne))
       (              (*move argtwo (reg t5))
                      (*wnot (reg t5) (reg t5))
                      (*move (reg t5) ArgOne)))

(DefCMacro *WMinus
           ((Xregp Xregp) (neg ARGTWO ArgOne))
           ((Xregp Anyp)  (*move argtwo ArgOne)
			   (neg ArgOne ArgOne))
           (              (*move argtwo (reg t5))
			   (*wminus (reg t5) (reg t5))
			   (*move (reg t5) ArgOne)))

(DefCMacro *MkItem
       ((Xregp fixp)      (andn ArgOne (reg i5) ArgOne)
			   (*move argtwo (reg t5))
			   (sllx (reg t5) 59 (reg t5))% sethi with tag to t5
			   (or ArgOne (reg t5) ArgOne))
       ((Xregp Xregp)     (andn ArgOne (reg i5) ArgOne)
			   (sllx argtwo 59 (reg t5))
			   (or ArgOne (reg t5) ArgOne))
       ((Xregp anyp)      (*move argtwo (reg t5))
			   (*Mkitem ArgOne (reg t5)))
       (                  (*move ArgOne (reg t1))
			   (*mkitem (reg t1) argtwo)
			   (*move (reg t1) ArgOne)))

(de *specsethi (num)          % avoid shift of constant at compiletime
        `((sethi ,(wshift num 17) (reg t5))))

(defcmacro *specsethi)  % support for *mkitem

(de smallinump (x) (sparc-short-immediate-p x))

(de immediate-p (x)    (fixp x))

(DefCmacro *tag
      ((Xregp Xregp)(srlx argtwo 59 ArgOne))
      ((Anyp  Xregp)(*move argtwo (reg t1))
                    (srlx (reg t1) 59 (reg t1))
                    (*move (reg t1) ArgOne))
      ((xregp anyp) (*move argtwo (reg t1))
                    (srlx (reg t1) 59 ArgOne))
      (             (*move argtwo (reg t1))
                    (*tag ArgOne (reg t1))))

(de *tag (ar1 ar2)
     (expand2operandCmacro ar1 ar2 '*tag))

(DefCMacro *JumpType
      (             (*Tag (Reg t5) ArgOne)
                    (*JumpEQ ArgThree (Reg t5) ArgTwo)))

(DefCMacro *JumpNotType
      (             (*Tag (Reg t5) ArgOne)
                    (*JumpNOTEQ ArgThree (Reg t5) ArgTwo)))

(DefCMacro *JumpInType
       ((Xregp Anyp) (srlx ArgOne 59 (reg t5))
                     (subcc (reg t5) ARGTWO (reg 0))
                     (ble ARGTHREE)
                     (subcc (reg t5) 31 (reg 0)) % negint-tag)
                     (be ARGTHREE)
                     (nop))
       (             (*move ArgOne (reg t5))
                     (srlx (Reg t5) 59 (reg t5))
                     (subcc (reg t5) ARGTWO (reg 0))
                     (ble ARGTHREE)
                     (subcc (reg t5) 31 (reg 0)) % negint-tag)
                     (be ARGTHREE)
                     (nop)))

(DefCMacro *JumpNotInType
       ((Xregp Anyp) (srlx ArgOne 59 (reg t5))
                     (subcc (reg t5) 31 (reg 0))
                     (be templabel)
                     (subcc (reg t5) argtwo (reg 0))
                     (bg ARGTHREE)
                     (nop)
                    templabel)
       (             (*move ArgOne (reg t5))
                     (srlx (reg t5) 59 (reg t5))
                     (subcc (reg t5) 31 (reg 0))
                     (be templabel)
                     (subcc (reg t5) argtwo (reg 0))
                     (bg ARGTHREE)
                     (nop)
                    templabel))


%---------------------------------------------------------------------
%   General Purpose Jumper.
%   FORMAT: Operand1 Operand2 Address JumpOpCode

(DefCMacro *JumpIf
     ((Xregp Xregp)      (subcc ArgOne argtwo (reg g0))
			  (argfour argthree)
			  (nop))
     ((XRegP SmallInumP) (subcc ArgOne argtwo (reg g0))
			  (argfour argthree)
			  (nop))
     ((Xregp anyp)       (*move argtwo (reg t5))
			  (subcc ArgOne (reg t5) (reg g0))
			  (argfour argthree)
			  (nop))
     ((anyp  Xregp)      (*move ArgOne (reg t5))
			  (subcc (reg t5) argtwo  (reg g0))
			  (argfour argthree)
			  (nop))
     ((anyp  SmallInumP) (*move ArgOne (reg t5))
			  (subcc (reg t5) argtwo  (reg g0))
			  (argfour argthree)
			  (nop))
     (                   (*move ArgOne (reg t5))
			  (*move argtwo (reg t4))
			  (subcc (reg t5) (reg t4) (reg g0))
			  (argfour argthree)
			  (nop))
)

(de *JumpIF (ArgOne ArgTwo Label Instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                     (ResolveOperand '(Reg Error) Label)
                  Instruction)
                 (get '*JumpIF 'CMacroPatternTable)))))


(defcmacro *jumpeq)

(de *jumpeq (Lbl ArgOne ArgTwo)
       (*jumpif ArgOne argtwo lbl 'BE!,a))

(defcmacro *jumpnoteq)

(de *jumpnoteq(Lbl ArgOne ArgTwo)
        (*jumpif ArgOne argtwo Lbl 'BNE!,a))

(defcmacro *jumpwlessp)

(de *jumpwlessp (Lbl ArgOne ArgTwo)
        (*jumpif ArgOne argtwo lbl 'BL!,a))

(defcmacro *jumpwgreaterp)

(de *jumpwgreaterp (Lbl ArgOne ArgTwo)
        (*jumpif ArgOne argtwo lbl 'BG!,a))

(defcmacro *jumpwleq)

(de  *jumpwleq(Lbl ArgOne ArgTwo)
        (*jumpif ArgOne argtwo lbl 'BLE!,a))

(defcmacro *jumpwgeq)

(de *jumpwgeq (Lbl ArgOne ArgTwo)
        (*jumpif ArgOne argtwo lbl 'BGE!,a))


(de fluid-arg-p (x)
    (and (pairp x)
     (member (car x) '(!$fluid !$global fluid global))))


(DefCMacro *Push
       ( (*cerror "NO push ")))

(DefCMacro *Pop
       ( (*cerror "NO pop ")))

(de take-car (x) (car x))
(de take-cadr (x) (cadr x))   % functions to strip out deferred and
			   % displacement parts.
(de take-caddr (x) (caddr x))


(de *loc (ArgOne argtwo)
   (let ((resolvedarg (resolveoperand '(reg t2) argtwo)))
      (cond ((deferred-p resolvedarg)
         `((*move (undeferred ,resolvedarg) ,ArgOne))) % get the address.
         ((displacement-p resolvedarg)
            `((add ,(cadr resolvedarg) ,(caddr resolvedarg) ,ArgOne)))
         ((eqcar resolvedarg 'cdr)
            `((add ,(cadr resolvedarg) 8 ,ArgOne)))
         (t `((*movex ,resolvedarg ,ArgOne))))))

(DefCMacro *loc)

%t6 is used, because it is not used in lap code, but it destroyed in cmacros.
%look at test file mask.s in ~burdorf/sparc for example.

(de X5p (u) (eq u 5))

(DefCMacro *Field

  ((Xregp Xregp zerop fixp)  (srlx ARGTWO (difference 64 ARGFOUR) ArgOne))
  ((Xregp Xregp X5p Fiftyninep) (andn ARGTWO (reg i5) ArgOne))
  ((Xregp Xregp fixp fixp)   (sllx argtwo argthree ArgOne)
			  (srlx ArgOne (difference 64 ARGFOUR) ArgOne))
  ((Xregp anyp fixp fixp)    (*move argtwo (reg t5))
			  (sllx (reg t5) argthree (reg t5))
			  (srlx (reg t5) (difference 64 ARGFOUR) ArgOne))
  ((anyp Xregp fixp fixp)    (*field (reg t5) argtwo argthree argfour)
			  (*move (reg t5) ArgOne))
  (			   (*move argtwo (reg t6))
			  (*field (reg t5) (reg t6) argthree argfour)
			  (*move (reg t5) ArgOne)))

(DefCMacro *SignedField

  ((Xregp Xregp fixp fixp)   (sllx argtwo argthree ArgOne)
			  (srax ArgOne (difference 64 ARGFOUR) ArgOne))
  ((Xregp anyp fixp fixp)    (*move argtwo (reg t5))
			  (sllx (reg t5) argthree (reg t5))
			  (srax (reg t5) (difference 64 ARGFOUR) ArgOne))
  ((anyp Xregp fixp fixp)    (*Signedfield (reg t5) argtwo argthree argfour)
			  (*move (reg t5) ArgOne))
  (			   (*move argtwo (reg t6))
			  (*Signedfield (reg t5) (reg t6) argthree argfour)
			  (*move (reg t5) ArgOne)))

% following is used by garbage collector only. Not very useful to
% optimize other cases than put inf field or put tag field (with constant)


(DefCMacro *PutField

 ((Xregp XregP fiveP FiftynineP) (andn ArgOne (reg i5) (reg t6))
			            (and   (reg i5) argtwo argtwo)
			            (or    argtwo (reg t6) ARGTWO))
 ((Xregp Xregp Zerop fivep)        (sllx   ArgOne 59 (reg t5))
			        (andn  argtwo (reg i5) argtwo)
			        (or    (reg t5) argtwo argtwo))
 ((inump Xregp Zerop fivep)        (mov   ArgOne (reg t5))
			        (sllx   (reg t5) 59 (reg t5))
			        (andn  argtwo (reg i5) argtwo)
			        (or    (reg t5) argtwo argtwo))
 (          (mov -1  (reg t5))
        (*move   ArgOne (reg t4))
        (sllx     (reg t5) (difference 64 argfour) (reg t5))
        (*wshift (reg t5) (minus argthree)) % for 0 opt
        (sllx   (reg t4) (difference 64 (plus argthree argfour)) (reg t4))
        (and   (reg t5) (reg t4) (reg t4))
        (*move argtwo (reg t6))
        (andn  (reg t6) (reg t5) (reg t6))
        (or   (reg t4) (reg t6) (reg t5))
        (*move (reg t5) argtwo)))


(deflist '((Byte        ((ldsb (displacement (reg 2) (reg 1)) (reg 1))))
       (PutByte     ((stb (reg 3) ((reg 1) (reg 2)))))
       (HalfWord    ((sllx (reg 2) 1 (reg t1))
                     (ldsh (displacement (reg t1)  (reg 1)) (reg 1))))
       (PutHalfWord ((sllx (reg 2) 1 (reg t1))
                     (sth (reg 3) (displacement (reg 1) (reg t1))))))
  'OpenCode)

(&OneReg '(Byte PutByte HalfWord PutHalfWord))

(on fast-integers)

(for (in function '(idapply0 idapply1 idapply2 idapply3 idapply4))
     (from register 1)
     (do
      (put function 'opencode
        `((*move ($fluid symfnc) (reg t1))
          (sllx (reg ,register) 5 (reg ,register))
          (srlx (reg ,register) 2 (reg t2))
          (ldx (displacement (reg t2) (reg t1)) (reg 15))
        (jmpl (displacement (reg 15) 0) (reg 15))
              (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      (put function 'exitopencode
        `((*move ($fluid symfnc) (reg t1))
          (sllx (reg ,register) 5 (reg ,register))
          (srlx (reg ,register) 2 (reg t2))
          (ldx (displacement (reg t2) (reg t1)) (reg t1))
        (jmpl (displacement (reg t1) 0) (reg 0))
          (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      ))

(for (in function '(addressapply0 addressapply1 addressapply2
                addressapply3 addressapply4))
     (from register 1)
     (do
      (put function 'opencode
         `((jmpl (deferred (reg ,register)) (reg 15))
               (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      (put function 'exitopencode
         `((jmpl (deferred (reg ,register)) (reg 0))
             (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      ))


% *feq, *fgreaterp and *flessp can only occur once in a function.

(deflist '((*wfix ((ldd (deferred (reg 1)) (reg f0))
               (fdtox (reg f0) (reg f2))
               (std  (reg f2) (displacement (reg sp)  -9))
               (ldx  (displacement  (reg sp) -9) (reg 1))))
       (*wfloat ((stx (reg 2) (displacement (reg sp) -9))
                 (ldd (displacement (reg sp) -9) (reg f0))
                 (fxtod (reg f0) (reg f2))
                 (std (reg f2) (deferred (reg 1)))))
       (*feq ((ldd  (deferred (reg 1)) (reg f0))
              (ldd (deferred (reg 2)) (reg f2))
              (fcmpd (reg f0) (reg f2))
              (fbe  *donefeq*)
              (nop)
              (mov (reg nil) (reg 1))
              *donefeq*))
       (*fgreaterp ((ldd  (deferred (reg 1)) (reg f0))
                    (ldd (deferred (reg 2)) (reg f2))
                    (fcmpd (reg f0) (reg f2))
			 (*move (quote t) (reg 1))
                    (fbg  *donefgreaterp*)
                    (nop)
                    (mov (reg nil) (reg 1))
                    *donefgreaterp*))
       (*flessp ((ldd  (deferred (reg 1)) (reg f0))
                 (ldd (deferred (reg 2)) (reg f2))
                 (fcmpd (reg f0) (reg f2))
                     (*move (quote t) (reg 1))
                 (fbl *doneflessp*)
                 (nop)
                 (mov (reg nil) (reg 1))
                 *doneflessp*))
       (*fplus2 ((ldd  (deferred (reg 2)) (reg f0))
                 (ldd (deferred (reg 3)) (reg f2))
                 (faddd (reg f0) (reg f2) (reg f0))
                 (std (reg f0) (deferred (reg 1)))))
       (*fdifference ((ldd  (deferred (reg 3)) (reg f0))
                      (ldd (deferred (reg 2)) (reg f2))
                      (fsubd (reg f2) (reg f0) (reg f0))
                      (std (reg f0) (deferred (reg 1)))))
       (*ftimes2 ((ldd  (deferred (reg 2)) (reg f0))
                  (ldd (deferred (reg 3)) (reg f2))
                  (fmuld (reg f0) (reg f2) (reg f0))
                  (std (reg f0) (deferred (reg 1)))))
       (*fquotient ((ldd  (deferred (reg 3)) (reg f0))
                    (ldd (deferred (reg 2)) (reg f2))
                    (fdivd  (reg f2) (reg f0) (reg f0))
                    (std (reg f0) (deferred (reg 1))))))
     'opencode)

(defcmacro *fast-apply-load
   ( (*move ArgOne ($fluid *second-value*)) ))

(put 'fast-idapply    'opencode
     '((*move ($fluid symfnc) (reg t5)) %($fluid symfnc) returns address of
			                %the symfnc table.
       (*move ($fluid *second-value*) (reg t1))
       (sllx (reg t1) 5 (reg t1))
       (srlx (reg t1) 2 (reg t1)) %multiply by 8 to get offset into symfnc table
       (ldx  (displacement (reg t1) (reg t5)) (reg 15))
	(jmpl (displacement (reg 15) 0) (reg 15))
       (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT

(put 'fast-codeapply    'opencode
     '((*move ($fluid *second-value*) (reg t1))
       (andn (reg t1) (reg i5) (reg t1))
			     %codepointer passed, so just strip off tag
       (jmpl (deferred (reg t1)) (reg 15))
       (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT

(put 'fast-idapply    'exitopencode
     '((*move ($fluid *second-value*) (reg t1))
       (*move ($fluid symfnc) (reg t5)) %($fluid symfnc) returns address of
			                %the symfnc table.
       (sllx (reg t1) 5 (reg t1))
       (srlx (reg t1) 2 (reg t1)) %multiply by 8 to get offset into symfnc table
       (ldx (displacement (reg t1) (reg t5)) (reg t1))
	(jmpl (displacement (reg t1) 0) (reg 0))
       (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT

(put 'fast-codeapply    'exitopencode
     '((*move ($fluid *second-value*) (reg t1))
       (andn (reg t1) (reg i5) (reg t1))
			  %codepointer passed, so just strip off tag
       (jmpl (deferred (reg t1)) (reg 0))
       (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT

(De *LamBind (Regs Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cregs cfluids cadrcfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Regs (rest Regs))   % Remove "Registers" from the front
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Regs) 16)) % two words per BndStk entry *4

   %looking for free register to preload values of fluid s

      (setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
                   (function (lambda (x)
                                (cond ((member x regs) nil)
                                       (t (cons x nil))
      )              ) )         )  )
      (setq freeregs (nconc (list '(reg T3) '(reg T4)) freeregs))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
                (nconc cadrcfluids (cons (car freeregs) nil)))
        (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist

      (setq list `((*move ($VIVfluid BndStkPtr) (Reg t1))
               (*move ($VIVfluid BndstkUpperBound) (Reg t5))
               (*wplus3 (Reg t2) ,lng (Reg t1))
               (subcc (reg t5)(reg t2)(reg g0))
               (bg!,a   ,genlabel)
               (nop)
               (*call Bstackoverflow)
               ,genlabel
               (*move (Reg t2) ($VIVfluid BndstkPtr))  )) %start of code

      (setq list (append initload list))

  loop
      (setq cregs (car Regs))
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
           (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 16))
      (Setq list (append list
     `((*move ,(car freeregs) (displacement (Reg t1) ,n))
       (*move (quote ,Cadrcfluids)
                   (displacement (reg t1) ,(wplus2 n -8)))
       (*move ,cregs ,cfluids)
      )          ))
      (setq fluids (cdr Fluids))
      (setq freeregs (cdr freeregs))
      (cond ((setq regs (cdr Regs)) (go loop)))
      (return list)
)    )

(defcmacro !*lambind)

(De *ProgBind (Fluids)

 % be careful, code is somewhat tricky

(prog (n lng list cfluids cadrcfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 16)) % two words per BndStk entry
                                         % * 8 addressingunits
      (setq freeregs '((reg t3)(reg T4)(reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
                (nconc cadrcfluids (cons (car freeregs) nil)))
        (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist

      (setq list `((*move ($VIVfluid BndStkPtr) (Reg t1))
               (*wplus3 (Reg t2) ,lng (Reg t1))
               (*move ($VIVfluid BndstkUpperBound) (Reg t5))
               (subcc (reg t5)(reg t2)(reg g0))
	       (bg!,a ,genlabel)
		(nop)
               (*call Bstackoverflow)
               ,genlabel
               (*move (Reg t2) ($VIVfluid BndstkPtr))  )) %start of code

     (setq list (append initload list))

 loop
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
           (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 16))
      (Setq list (append list
             `((*move ,(car freeregs) (indexed (Reg t1) ,n))
               (*move (quote ,Cadrcfluids)
			    (indexed (reg t1)  ,(wplus2 n -8)))
               (*move (quote nil) ,cfluids)
      )          ))
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *progbind)

(De *FreeRstr (Fluids)

 (prog (n lng list cfluids listfluids lab initload freeregs genlabel)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
      (setq freeregs '((reg t3)(reg t4)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
      (setq n (wtimes2 8 (wdifference 2 lng)))
      (setq lng (wtimes2 lng 8)) % * addressingunitperitem
      (setq initload (list '(*move ($VIVfluid Bndstkptr) (reg t1))))

preload  (setq initload
       (progn (setq listfluids
               (if freeregs
                (nconc listfluids (cons (car freeregs) nil))
                (nconc listfluids (cons nil nil))) )
        (nconc initload
         (if freeregs
     `((*move (indexed (reg t1) ,n) ,(car freeregs))) nil)
       )   ))
       (setq n (wplus2 n 16))
       (when freeregs (setq freeregs (cdr freeregs)))
       (setq cfluids (cdr cfluids))

       (cond (cfluids (go preload)))

       (setq freeregs listfluids ) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and nil if not enough regs available

      (setq list `((sub   (Reg t1) ,lng (reg t2))
               (*move ($VIVfluid BndstkLowerBound) (Reg t5))
               (subcc (reg t2)(reg t5) (reg g0))
               (bge   ,genlabel)
               (*call  Bstackunderflow)
               ,genlabel
               (*move (Reg t2) ($VIVfluid BndstkPtr))  )) %start of code

     (setq list (append initload list))
     (setq n 0)

 loop
      (setq cfluids (car Fluids))
      (setq n (wplus2 n 16))

  % insert reloaded register or memory reference

      (setq list (append list
           (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
                `((*move (indexed (Reg t2) ,n) (reg t1))
		  (*move (reg t1) ,cfluids )))

      )          )
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )

(defcmacro *freerstr)

(setq *unsafebinder t)   % has to save Registers across calls

(global '(*writingasmfile))

(de !*xjumpon (register lowerbound upperbound labellist)
    (for (from x lowerbound upperbound)
         (join
          `((subcc ,register ,x (reg g0))
            (be ,(pop labellist))
            (nop)))))

(de !*jumpon (register lowerbound upperbound labellist)
   (if *writingasmfile
       (*xjumpon register lowerbound upperbound labellist)
     (PROG (X LL LL2)
       (setq ll  (gensym))
       (setq ll2 (stringgensym))
       (SETQ X
        (if (and (weq lowerbound 0) (weq upperbound 31) *syslisp)
                  % jumpon on tags (most probably)
      `((sllx ,register 3 (reg t2))  % 8 bytes per jumptable entry
        (move-relocatable-object ,ll2 (reg t1))
        (jmpl (displacement (reg t1) (reg t2)) (reg 0))
        (nop)
       ,ll2)
      `((subcc ,register ,upperbound (reg 0))
        (be!,a ,(lastcar labellist))
        (nop)
        (bpos!,a (label ,ll))
        (nop)
        (subcc ,register ,Lowerbound (reg t5))
        (bneg!,a (label ,ll))
        (nop)
        (sllx (reg t5) 3 (reg t5))  % 8 bytes per jumptable entry
    (move-relocatable-object ,ll2 (reg t1))
        (jmpl (displacement (reg t1) (reg t5)) (reg 0))
        (nop)
       ,ll2) ) )
      Loop  (Setq x (nconc X `((BA ,(car Labellist)) ,(list 'nop))) )
            (setq Labellist (cdr Labellist))
            (cond (Labellist (go loop)))

            (setq x (nconc x `((*lbl (Label ,ll))) ))
            (return x)
)  )  )

(defcmacro !*jumpon)


% Call foreign fns (loaded by oload) by putting SymFunc cell addr in a reg,
% then doing "calls" instr indirect via the target address of the jmp
% instruction stored there.  Use old direct link for asm (kernel) compile.

(de *foreignlink (functionname functiontype numberofarguments)

% Following Sparc Architecture Manual, the Parameter past the six have
% to be passed in stack. We therefore need some extra space, we get it
% by doing an extra alloc.

 (if (wgreaterp numberofarguments 6)
     `((*alloc 30)            % This is a rare case (hopefully)
       (*move (reg 15) (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 14) (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 13) (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 12) (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 11) (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 10) (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 9)  (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 8)  (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 7)  (reg o5)) (stx (reg o5) (indexed (reg sp) 124))
       (*move (reg 6) (reg o5))
       (*move (reg 5) (reg o4))
       (*move (reg 4) (reg o3))
       (*move (reg 3) (reg o2))
       (*move (reg 2) (reg o1))
       (*move (reg 1) (reg o0))
       (call (foreignentry ,functionname) ,numberofarguments)
       (nop)
       (*move (reg o0) (reg 1))
       (*dealloc 30))

   (let ((sequence nil))
     (setq sequence
               (cond ((equal numberofarguments 0) nil)
                     (t (pnth '((*move (reg 6) (reg o5))
			      (*movex argfive  (reg o4))
			         (*movex argfour  (reg o3))
			         (*movex argthree (reg o2))
			         (*movex argtwo   (reg o1))
			         (*movex ArgOne   (reg o0)))
			       (difference 7  numberofarguments)))))
     (setq sequence
       (append sequence
          (list (list 'call (list 'foreignentry functionname)
			              numberofarguments)
			  (list 'nop))))
     (setq sequence
       (append sequence
                  '((mov (reg o0) (reg 1))   %store return value in r1
                   ))))))

(defcmacro !*foreignlink)

(defcmacro *movex)
   % Movex is of type cmacro

(de *movex (src dst)
     (let (tmp)
      (cond ((sparc-register-p dst)
          (cond ((sparc-register-p src)  `((mov ,src ,dst)))
                ((and (sparc-fluid-p src) (equal dst '(reg fsr))
			   *writingasmfile)
			  `((sethi (hi ,src) (reg t5))
			    (ldx (displacement (reg t5) (lo ,src)) ,dst)))
                ((eqcar src 'quote)
                     (if *writingasmfile           %% asmout and lap generate
			  `((sethi (hh ,src) ,dst)  %% different code.
			    (sethi (lm ,src) (reg t6))  %% different code.
			   (or ,dst (hm ,src) ,dst)
			   (sllx ,dst 32 ,dst)
			   (or ,dst (reg t6) ,dst)
			    (ldx  (displacement ,dst (lo ,src)) ,dst))
			  `((move-relocatable-object ,src ,dst)
			   (ldx  (displacement ,dst 0) ,dst))))
                ((or (sparc-absolute-p src) (sparc-fluid-p src))
                 (if *writingasmfile           %% asmout and lap generate
			 `((sethi (hh ,src) ,dst)  %% different code.
			    (sethi (lm ,src) (reg t6))  %% different code.
			    (or ,dst (hm ,src) ,dst)
			    (sllx ,dst 32 ,dst)
			    (or ,dst (reg t6) ,dst)
			    (ldx  (displacement ,dst (lo ,src)) ,dst))
                     `((move-relocatable-object-and-load ,src ,dst))))
                ((sparc-short-immediate-p src) `((mov ,src ,dst)))
                ((or (sparc-relocatable-immediate-p src)
                     (sparc-foreignentry-p src))
                 (if *writingasmfile           %% asmout and lap generate
			 `((sethi (hh ,src) ,dst)  %% different code.
			    (sethi (lm ,src) (reg t6))  %% different code.
			    (or ,dst (hm ,src) ,dst)
			    (sllx ,dst 32 ,dst)
			    (or ,dst (reg t6) ,dst)
			    (add ,dst (lo ,src) ,dst))
                     `((move-idloc-object ,src ,dst))))
                ((and (sparc-immediate-p src) (negativeimmediatep src))
                     `((sub ,dst ,dst ,dst)
			(sethi (lm ,(minus src)) ,dst)
			(add ,dst (lo ,(minus src)) ,dst)
			 (sub (reg g0) ,dst ,dst)))
                ((and (sparc-immediate-p src) (fixp src) (lessp src 0))
                     `((sub ,dst ,dst ,dst)
                        (or ,dst (hm ,(minus src)) (reg t6))
                        (sllx (reg t6) 32 (reg t6))
                        (sethi (lm ,(minus src)) ,dst)
                        (add ,dst (lo ,(minus src)) ,dst)
                        (or (reg t6) ,dst ,dst)
                        (sub (reg g0) ,dst ,dst)))
                ((sparc-immediate-p src)
                     `((sub ,dst ,dst ,dst)
			(or ,dst (hm ,src) (reg t6))
			(sllx (reg t6) 32 (reg t6))
			(sethi (lm ,src) ,dst)
			(add ,dst (lo ,src) ,dst)
			(or (reg t6) ,dst ,dst)))
                ((sparc-huge-displacement-p src)
                      `((sub ,dst ,dst ,dst)
			(sethi (lm ,(caddr src)) ,dst)
			(add ,dst (lo ,(caddr src)) ,dst)
			(ldx (,(cadr src) ,dst) ,dst)))
                ((and (sparc-deferred-p src) (sparc-indirect-p (cadr src)))
			(car (load-src-arg src dst)))
                ((sparc-indirect-p src) `((ldx ,src ,dst)))
                (t  (print-error-status)
                    (error 0
                 (list "Unknown addressing mode for *movex:" src dst)))))
         ((sparc-register-p src)
          (cond ((and 
		  (or (sparc-absolute-p dst)
                     (sparc-fluid-p dst)
                     (sparc-foreignentry-p src))
		  (not (member src '((reg t5) (reg t6)))))
                 (if *writingasmfile
                     `((sethi (hh ,dst) (reg t6))  %% different code.
			(sethi (lm ,dst) (reg t5))  %% different code.
			(or (reg t6) (hm ,dst) (reg t6))
			(sllx (reg t6) 32 (reg t6))
			(or (reg t6) (reg t5) (reg t6))
			(stx ,src (displacement (reg t6) (lo ,dst))))
                     `((move-relocatable-object-and-store
			              ,dst (reg t6) ,src))))
		((or (sparc-absolute-p dst)
                     (sparc-fluid-p dst)
                     (sparc-foreignentry-p src))
                 (if *writingasmfile
                     `((sethi (hh ,dst) (reg t6))  %% different code.
                        (sethi (lm ,dst) (reg t4))  %% different code.
                        (or (reg t6) (hm ,dst) (reg t6))
                        (sllx (reg t6) 32 (reg t6))
                        (or (reg t6) (reg t4) (reg t6))
                        (stx ,src (displacement (reg t6) (lo ,dst))))
                     `((move-relocatable-object-and-store
                                      ,dst (reg t4) ,src))))
                ((and (sparc-deferred-p dst) (sparc-indirect-p (cadr dst)))
                 (setq tmp (load-src-arg (cadr dst) '(reg t6)))
                 (append (car tmp) `((stx ,src (deferred (reg t6))))))
                ((sparc-huge-displacement-p dst)
                 `((sethi (hi ,(caddr dst)) (reg t6))
                   (add (reg t6) (lo ,(caddr dst)) (reg t6))
                   (stx ,src (,(cadr dst) (reg t6)))))
                ((sparc-indirect-p dst) `((stx ,src ,dst)))
                (t   (print-error-status)
                     (error 0
			 (list "Unknown addressing mode:" src dst)))))
         (t (print (list " Unsaeglich: Move " src dst))
))))
