%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         CRAY-FORMS.SL
% Description:  Cray lap support forms.
% Author:       Robert Kessler, Utah Pass Project
% Created:      27-Sep-84
% Modified:     13-Dec-84 06:59:22 (Robert Kessler)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, University of Utah, all rights reserved
% (c) Copyright 1987, Konrad Zuse Zentrum Berlin, for modifications
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 12-Jan-87 W.Neun - Cleaning up, done Makeregs as DM
% 30-May-86 W.Neun - Added new B and T Reg's and their numbers.
%               added some properties for gc eval and binding
%               removed LNOT, is in arithmetic now
% 7-Apr-86 H.Melenk ,ZIB Berlin
%     Corrected  add-forward-internal to handle multiple forward refs.
% 10-Mar-86 W.Neun  - Added new B Reg's and their numbers.
% 10-May-85 Kessler - Added new T Reg's and their numbers.
% 08-Feb-85 Kessler
% Uncommented out the idloc kludge of the year.
% 27-Dec-84 Kessler
%  Put in fixed version of valuecell that knows about fasl files.
% 13-Dec-84 06:58:58 (Robert Kessler)
%  Changed parceladdressvalue again, to consider internalentry's that
%  are generated when doing r2i.
% 11-Dec-84 15:35:26 (Robert Kessler)
%  Added fix to parceladdressvalue, so it only adds to the forward
%  internal reference list when appropriate, using proc:
%  add-forward-internal.  Also, changed order of elements in this
%  list, so the name is first (for doing an assoc).
% 29-Nov-83 08:21:57 (Robert Kessler)
%  Fixed bug in putbittable that had parcel-address used instead of
%  offset in the computation of the pnumber.
% 28-Nov-83 13:30:44 (Robert Kessler)
%  Changed parceladdressvalue so it looks for the special case of a
%  function cell address....  See the code for more details.
% 28-Nov-83 13:05:55 (Robert Kessler)
%  Added *test-only* fluid, so we can access the test-only flag from
%  addressvalue and parceladdressvalue from operandvaluefunctions
%  (like binxidloc).
% 28-Nov-83 08:31:21 (Robert Kessler)
%  Added another argument to parceladdressvalue and addressvalue.  If
%  T, it will not write the bit table, if nil it will write the bit
%  table.  Thus in T mode, we are testing only.
% 27-Nov-83 11:43:33 (Robert Kessler)
%  Changed parceladdressvalue and addressvalue, so they simply add the
%  next bit for the bittable, instead of actually calling
%  updatebittable.  That way, things will only happen when the
%  putparcel is done.  Also changed crayputparcel to updatebittable.
%  Defined a putbittable function to store the bits into the word.
% 27-Sep-84 10:22:28 (Robert Kessler)
%  Changed addressvalue and parceladdressvalue to update the bit table
%  when either one is generated.  The only other thing that needs to
%  be changed is the operandvalue functions, so they do the correct
%  thing.
% 3/22/84 - Kessler
%  Lnot added, was missing.
%  Put valuecelllocation, setf of symfnc and idloc back in, since it seems to
%   need it in the running system.
%  New version of sregp that allows reg number to be an sreg
%  Insert dskin of tags.sl
%  Inserted addressvalue function definition
%  Modify to ijkpart, so it picks off the correct 6 bits.
%  Modify parceladdressvalue, so it knows about negative number operands.
%  Added *pcode flag to tell oneparcel, twoparcel and crayputword to
%   print what they are dumping.  Very good for debugging the binary code.
% 9/7/83 - Kessler
%  Commented out idloc, valuecelllocation and setf of symfnc, because of
%  weirdies in the running kernel.
% 9/5/83 - Kessler
%  Fixed bug in Crayputword, which put the address instead of the word.
% 8/16/83 - Kessler
%  Added ParcelToWord Function, so AddressValue will return a word
%  address instead of a parcel one.
% 8/15/83 - Kessler
%  Changed printf in the oneparce, twoparcel and crayputword, so it
%  prints offset, real address and parcels.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (load sys-consts))
(compiletime (setq *syslisp t)) % for symval etc.
 
(fluid '(symval symfnc symprp symnam ArgumentBlock))
 
(compiletime(flag '(oneparcel twoparcel true-ijkpart)
                                                    'internalfunction))
 
% Do the definitions.
 
(compiletime(progn
 
(put 'reloc_word          'const 1)
(put 'reloc_id_number     'const 2)
(put 'escape_to_second    'const 3)
(put 'parcel_label        'const 0)
(put 'word_label          'const 1)
(put 'reloc_value_cell    'const 2)
(put 'reloc_function_cell 'const 3)
nil))
 
 % establish as compiler constants
 
(put 'reloc_word          'compiler-constant? t)
(put 'reloc_id_number     'compiler-constant? t)
(put 'escape_to_second    'compiler-constant? t)
(put 'parcel_label        'compiler-constant? t)
(put 'word_label          'compiler-constant? t)
(put 'reloc_value_cell    'compiler-constant? t)
(put 'reloc_function_cell 'compiler-constant? t)
 
(put 'reloc_word          'compiler-constant-value 1)
(put 'reloc_id_number     'compiler-constant-value 2)
(put 'escape_to_second    'compiler-constant-value 3)
(put 'parcel_label        'compiler-constant-value 0)
(put 'word_label          'compiler-constant-value 1)
(put 'reloc_value_cell    'compiler-constant-value 2)
(put 'reloc_function_cell 'compiler-constant-value 3)
 
(de  LapPlus2(x y)
  (Plus2 x y))
 
(de  LapLand(x y)
  (Land x y))
 
(de  LapLor(x y)
  (Lor x y))
 
(de  LapAdd1(x)
  (Add1 x ))
 
(de  LapSub1(x)
  (Sub1 x ))
 
(de  LapLsh(x y)
  (Lshift x y))
 
(df DefOpcode (a)
 (Prog (op args body)
       (setq Op (first a))
       (setq args (second a))
       (setq Body (cddr a))
  (put op 'OpCodeFunction
     (list 'lambda args
        (cons 'cond (append body
                      (list (list 'T (list 'OpFailed (Mkquote op)
                                        (cons 'list args))))))))
 (return op)
))
 
(de OpFailed (op args)
  (printf "*** CRAY op %p failed for arguments %p %n" op args)
  (op-ijk 0 0))
 
(de RegP  (R)
   (eqCar R 'Reg))
 
(compiletime (ds Regp-inline (r) (eqcar r 'reg)))
 
(de AregP (R)
  (and (RegP-inline R) (get (second R) 'Areg)))
 
(de BregP (R)
  (and (RegP-inline R) (get (second R) 'Breg)))
 
(de Sregp (r)
  (prog (secr)
   (cond ((not (regp-inline r)) (return nil))
         ((numberp (setq secr (second r))) (return secr)))
   (return (get secr 'sreg))))
 
(de Sreg0P (R)
  (and (regP-inline R) (eq (second R) 'S0)))
 
(de VregP (R)
  (and (RegP-inline R) (get (second R) 'Vreg)))
 
(de TregP (R)
  (and (RegP-inline R) (get (second R) 'Treg)))
 
(compiletime
(dm MakeRegs (u)
  (prog (i type liste)
      (setq i 0)
      (setq type (cadr u))
      (foreach x in (caddr u)
         do (setq liste (cons (list 'put (mkquote x) (mkquote type) i)
                              liste))
                   (setq i (add1 i)))
      (return (cons 'progn liste)))))
 
(MakeRegs Areg (A0 A1 A2 A3 A4 A5 A6 A7))
(MakeRegs Sreg (S0 S1 S2 S3 S4 S5 S6 S7))
(MakeRegs Vreg (V0 V1 V2 V3 V4 V5 V6 V7))
 
(MakeRegs Breg (B00 B01 B02 B03 B04 B05 B06 B07
                B10 B11 B12 B13 B14 B15 B16 B17
                B20 B21 B22 B23 B24 B25 B26 B27
                B30 B31 B32 B33 B34 B35 B36 B37
                B40 B41 B42 B43 B44 B45 B46 B47
                B50 B51 B52 B53 B54 B55 B56 B57
                B60 B61 B62 B63 B64 B65 B66 B67
                B70 B71 B72 B73 B74 B75 B76 B77))
 
(makeRegs Treg (T00 T01 T02 T03 T04 T05 T06 T07
                T10 T11 T12 T13 T14 T15 T16 T17
                T20 T21 T22 T23 T24 T25 T26 T27
                T30 T31 T32 T33 T34 T35 T36 T37
                T40 T41 T42 T43 T44 T45 T46 T47
                T50 T51 T52 T53 T54 T55 T56 T57
                T60 T61 T62 T63 T64 T65 T66 T67
                T70 T71 T72 T73 T74 T75 T76 T77))
 
% Special registers not included in the above.
 
(makeregs treg (nil t1 t2 t3 t4 t5 t6 t7))
 
(put 'stacklowerbound 'treg 8#11)
(put 'stackupperbound 'treg 8#12)
(put 'heaplowerbound  'treg 8#13)
(put 'heapupperbound  'treg 8#14)
(put '&symflag&       'treg 8#15)
(put 'arithargs0      'treg 8#16)
(put 'arithargs1      'treg 8#17)
(put 'heaptrapped     'treg 8#20) %?????????????
(put 't               'treg 8#21)
(put 'betahi*         'treg 8#22)
(put 'betalow*        'treg 8#23)
(put 'currentscantable* 'treg 8#24)
(put 'iobuffer        'treg 8#25)
(put 'prevframesize   'treg 8#26)
(put 'returnaddr      'treg 8#27)
 
% the following registers up to t76 are for frame resp. fast-frame
 
(put 'nargreg       'treg 8#77)
 
(put 'st            'areg 4)
 
% registers B10 to B27 are for vector codes
% registers B30 to B43 are for io warrays and tokenscanner
 
(put 'nextposition  'breg 8#30)
(put 'lineposition  'breg 8#31)
(put 'writefunction 'breg 8#32)
(put 'readfunction  'breg 8#33)
(put 'maxline       'breg 8#34)
(put 'chtokentype   'breg 8#35)
(put 'currentchar   'breg 8#36)
(put 'bufferlength  'breg 8#37)
 
(put 'tokch         'breg 8#40)
(put 'symget        'breg 8#41)
 
(put 'unreadbuffer        'breg 8#43)
(put 'lambda-link-address 'breg 8#44)
(put 'codeptr*            'breg 8#45)
(put 'undefined-function-address 'breg 8#46)
(put 'catchstackptr 'breg 8#47)
 
(put 'symval        'breg 8#50)
(put 'symfnc        'breg 8#51)
(put 'symprp        'breg 8#52)
(put 'symnam        'breg 8#53)
(put 'stkupbnd      'breg 8#54)
(put 'HashTable     'breg 8#55)
(put 'Bndstkptr     'breg 8#56)
(put 'TokenBuffer   'breg 8#57)
(put 'heaplast      'breg 8#60)
(put 'heaptrapbound 'breg 8#61)
(put 'bndstkupperbound   'breg 8#62)
(put 'argumentblock 'Breg 8#63)
(put 'bndstklowerbound   'breg 8#64)
(put 'codeargs      'breg 8#65)
(put 'symcnt        'breg 8#66)
(put 'lambindargs*  'breg 8#67)
(put 'in*           'breg 8#70)
(put 'out*          'breg 8#71)
(put 'big-preallocated* 'breg 8#72)
(put '*bigcompress-threshold* 'breg 8#75)
 
% some fluids which are not a kernel-structure or are altered
% have to be updated by garbage collector because they
% point into heap. some fluids may be set by the user
 
(put 'arithargs0        '*known-to-gc 1) % value is vectorposition !!
(put 'arithargs1        '*known-to-gc 2)
(put 'currentscantable* '*known-to-gc 3)
(put 'iobuffer          '*known-to-gc 4)
(put '&symflag&         '*known-to-gc 5)
                                     % if more than 10 vars are used
                                     % enlarge vector in compacting-gc
 
(flag '(currentscantable* in* out* symget
        *bigcompress-threshold*)
                                    '*user-variable)
 
(flag '(nextposition lineposition writefunction readfunction
        maxline bufferlength unreadbuffer codeargs lambindargs*)
       '*warray-pointer)
 
 
% Assembly Functions
 
(fluid '(CurrentOffset*))
(de OpValue (op)
  (Wshift op 9))
 
(de IjkPart (e)    %this function is a jkpart ??? Neun
   (wshift (wand e 16#3f0000) -16))
 
(de True-IjkPart (e)
   (wshift (wand e 16#1ff0000) -16))
 
(De Mpart (e)
   (wand e 16#FFFF))
 
(de Op-ijk (Op ijk)
  (Oneparcel (wor (OpValue op) ijk)))
 
(de Op-i-jk (Op i jk)
  (Oneparcel (wor (OpValue op)
               (wor (wshift i 6) jk))))
 
(de Op-i-j-k (op i j k)
  (oneParcel (wor (opvalue op)
             (wor (wshift i 6)
               (wor (wshift j 3) k)))))
 
(de Op-i-e  (op i e)
 (setq e (AddressValue e nil))
 (TwoParcel (wor (OpValue op)
            (wor (wshift i 6) (IjkPart e)))
          (Mpart e)))
 
(de Op-h-i-e (op h i e)
  (setq e (ParcelAddressValue e nil))
  (TwoParcel (wor (wshift op 12)
             (wor (wshift h 9)
             (wor (wshift i 6) (IjkPart e))))
             (Mpart e)))
 
(de JumpOp (op e)
  (Setq e (ParcelAddressValue e nil))
  (TwoParcel (wor (OpValue op) (True-IjkPart e)) (Mpart e)))
 
% Kludge of the year, just to avoid having IDLOC defined during compilation
 
(CompileTime (fluid '(MACRO)))
%
(setf MACRO 'MACRO)
 
(PutD 'IDLoc MACRO (function (lambda (x) (id2int (cadr x)))))
%
% Parcel gets former warray's address and turns into a parcel.
(de parcel (u) (wshift u 2))
 
% Add the argument ref-name to the internal list if necessary.
%  This is written for clarity, instead of using nots and an and.
(de add-forward-internal (Ref-Name)
  (cond
   ((assoc Ref-Name labeloffsets*)
    nil) 
   ((get Ref-Name 'internalentryoffset)
    nil) 
   (t (setq forwardinternalreferences*
      (cons (cons Ref-Name currentoffset*)
            forwardinternalreferences*)))))
 
% This fluid is used to pass to guys like xbinxidloc, whether or not
%  we will in the add-next-bits.
(fluid '(*test-only*))
 
% The actual location of an address.
% This change is really convoluted.  We get passed in a jump to a
% symfnc cell something that looks like (plus2 (parcel symfnc) 5884),
% which came from a (j (plus2 (parcel symfnc) (idloc foo))).  So,
% since I'm not sure where the idloc is getting eliminated, what we
% will do is reconstruct it and set the appropriate bits so it can be
% a relocated function cell.
%%%% This code is terrible.  Rewrite and use some local variables!! %%%
 
(de parceladdressvalue (e *test-only*)
  (cond
   % We'll just use the regular mode when compiling in core and if we
   % are just looking to test for padv.
   ((and (not *test-only*)
       *writingfaslfile
       (pairp e)
       (pairp (cdr e))
       (equal (cadr e) '(parcel symfnc))
       (pairp (cddr e))
       (numberp (caddr e)))
       % Wouldn't a pattern matcher be nice here??  KLUDGE OF THE YEAR
    (add-next-bit (const escape_to_second))
    (add-next-bit (const reloc_function_cell))
    % Boy what sleze!!!
    (findidnumber (int2id (wshift (caddr e) -2))))
   ((minusexpp e) e)
   ((posintp e) e)
   ((labelp e) (cond ((and *writingfaslfile (not *test-only*))
                  (add-next-bit (const escape_to_second))
                  (add-next-bit (const parcel_label))))
    (labelvalue e))
   ((eqcar e 'internalentry)
    (cond ((and *writingfaslfile (not *test-only*))
         (add-next-bit (const escape_to_second))
         (add-next-bit (const parcel_label))
         (add-forward-internal (cadr e))))
    (cond ((and *writingfaslfile
            (or (and (assoc (cadr e) labeloffsets*)
                   (setq e (cdr (assoc (cadr e) labeloffsets*))))
                (setq e (get (cadr e) 'internalentryoffset))))
         e)
        % Aha, we don't yet know what the address is, so we
        % need to set things to fix it up later.
        (*writingfaslfile 0)
        ((assoc (cadr e) labeloffsets*)
         (wplus2 codebase* (cdr (assoc (cadr e) labeloffsets*))))
        (t (functioncelllocation (cadr e)))))
   ((posintp (eval e)) (eval e))
   ((minusexpp (eval e)) (eval e))))
 
(de addressvalue (e *test-only*)
   (cond ((minusexpp e) e)
         ((posintp e) e)
         ((labelp e) (cond (*writingfaslfile
                       (cond ((not *test-only*)
                            (add-next-bit (const escape_to_second))
                            (add-next-bit (const word_label))))
                       (parceltoword (labelvalue e)))
                     (t (parceltoword (labelvalue e)))))
         ((not (regp e)) (operandvalue e))))
 
% (dm idloc (u) (id2int (cadr u)))
 
% The location of the value cell.
(de valuecelllocation (u)
  (cond ((not *writingfaslfile)
       (wplus2 symval (id2int u)))
      (t 
       (cond ((not *test-only*)
            (add-next-bit (const escape_to_second))
            (add-next-bit (const reloc_value_cell))))
       (findidnumber u))))
 
(de ExtraRegLocation (X)
  (progn 
   (setq X (second X))
   (cond ((not *WritingFaslFile)
        (wplus2 ArgumentBlock
             (wdifference X (wplus2 MaxRealRegs 1))))
       (t
        (cond ((not *test-only*)
             (add-next-bit (const escape_to_second))
             (add-next-bit (const reloc_value_cell))))
        (wplus2 X 8150))))  )
 
(de posintP (e)
  (and (fixp e) (wgeq e 0)))
 
(de ijksize(x)
     (if (eq x (wand x 8#777)) x nil))
 
(de JKsizeP(x)
 %(or (and (fixp x) (wgeq x 0) (wleq x 8#77) x)
 %    (and (fixp (setf x (get x 'wconst))) (JKsizeP x))))
    (cond ((null x) nil)
          ((numberp x) (and (posintp x) (wleq x 8#77) x))
          ((idp x) (jksizep (get x 'wconst)))
          (t nil)))
 
(de MinusExpP (x)
  (and (fixp x) (minusp x)))
 
% Convert a Parcel address to a word address, first checking that the
%  Parcel address is in fact on a word boundary, signaling an error
%  if it is not.
(de ParcelToWord (ParcelValue)
  (cond ((WordBoundaryP ParcelValue) (wrshift ParcelValue 2))
      (T (Error 999 "ParcelToWord - Parcel not on a Word Boundary"))))
 
% ---- Lowest level Stubs for debugging
% Use Parcel Addresses as BPS addresses
% Word=[P1,P2,P3,P4]
 
(de parcel4(W)
      (wand W 16#FFFF))
 
(de parcel3(W)
      (wand 16#FFFF (wshift W -16)))
 
(de parcel2(W)
      (wand 16#FFFF (wshift W -32)))
 
(de parcel1(W)
      (wand 16#FFFF (wshift W -48)))
 
(global '(*pcode))
 
(de putbittable (Parcel-address Offset Bits)
  (Prog (WAddress PNumber ShiftSize)
    % Update the bittable.
    (setq WAddress (wplus2 (wrshift Parcel-Address 2) (wrshift Offset 5)))
    (setq PNumber (wand Offset 31))
    (if (eq PNumber 0)
     (putmem WAddress
            (wlshift Bits 62))
     (progn
      (setq ShiftSize (wdifference 62 (wshift PNumber 1)))
      (putmem WAddress
             (wor (wlshift Bits ShiftSize)
                      (wand (wnot (wlshift 16#3 ShiftSize))
                          (getmem WAddress))))))
     (when  *pcode
             (return (printf "Bittable, Adr: %w Offset: %w Value: %w%n"
                    Parcel-address Offset Bits)))))
 
% CrayGetParcel  already defined in the kernel
 
% Return the parcel letter of the address.
(de ParcelLetter (Address)
   (cadr (Assoc (wand Address 3) '((0 "a") (1 "b") (2 "c") (3 "d")) )))
 
(de OneParcel (x)
  (Prog (NewAddr OutputBase*)
      (setq OutputBase* 8)
        (setf NewAddr (CalculateAddress))
      (cond (*pcode
           (PrintF "%p%t%pp%w%t%p%n" CurrentOffset* 10
                 (wrshift NewAddr 2) (ParcelLetter NewAddr) 25 x)))
      (CrayPutParcel NewAddr x)
        (setq CurrentOffset* (wPlus2 CurrentOffset* 1))))
 
 
(de TwoParcel (x y)
  (Prog (NewAddr OutputBase*)
      (setq OutputBase* 8)
        (setf NewAddr (CalculateAddress))
      (cond (*pcode (PrintF "%p%t%pp%w%t%p%t%p%n" CurrentOffset* 10
            (wrshift NewAddr 2) (ParcelLetter NewAddr) 25 x 40 y)))
      (CrayPutParcel NewAddr x)
      (CrayPutParcel (wplus2 NewAddr 1) y)
        (setq CurrentOffset* (wPlus2 CurrentOffset* 2))))
 
(de CrayPutWord (x y)
  (Prog (OutputBase* NewAdr)
        (setq OutputBase* 8)
        (when (not (intp y)) (setq y (int2sys y)))
       (cond (*pcode
         (PrintF "%p%t%pp%w%t%p%n" CurrentOffset* 10 (wrshift x 2)
             (ParcelLetter x) 25 y)))
        (setq NewAdr (CalculateAddress))
        (CrayPutParcel
          NewAdr (parcel1 y))
        (CrayPutParcel
          (Wplus2 NewAdr 1) (parcel2 y))
        (CrayPutParcel
          (Wplus2 NewAdr 2) (parcel3 y))
        (CrayPutParcel
          (Wplus2 NewAdr 3) (parcel4 y))
        (setq CurrentOffset* (wPlus2 CurrentOffset* 4))))
 
 
(de NotDone (x)
  (printF "*** CRAY LAP %p not implemented yet ***%n" x)
  (Op-ijk 0 0))
 
(de CalculateAddress ()
   (when (and *writingfaslfile (wgeq (wplus2 currentoffset* 4)
                                     (wshift maxfasloffset* 2)))
         (when (not (add-fasl-space))
          (fatalerror "BPS exhausted during Faslout")))
   (wplus2 Codebase* CurrentOffset*))
 
(compiletime (setq *syslisp nil)) % for symval etc.
