%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         CRAY-INSTRS.SL
% Description:  Cray Assembler Tables, using CMACLIB names.
% Author:       M. L. Griss and Robert Kessler, Utah Pass Project
% Created:      19-Nov-82
% Modified:     28-Nov-83 08:35:11 (Robert Kessler)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, University of Utah, all rights reserved
% (c) Copyright 1987, Konrad Zuse Zentrum Berlin, for modifications
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 2 -Feb-88 - W. Neun
%             Revised semaphores and shared REGS, INCLUDED EMA PLOC
% 4 -Jan-88 - W. Neun
%             Added Semaphore instructions and Shared B and T regs
% 10-Nov-87 - H. Melenk and W. Neun
%             Added DRI/ERI and DFI/EFI instructions
% ATTENTION DRI/ERI are not supported on a CRAY1
% 15-Nov-86 - H. Melenk and W. Neun
%             Added some Pseudo instructions for optimization control
% 15-Apr-86 - W. Neun added movsbit (Si   Sj!Si&SB) as variant of merge
%                     Added SCONST04, for CAL's 0.4 instr.
% 15-May-85 - William F. Galway
%  Added SCONST06, for CAL's 0.6 instr.
% 27-Dec-84 - Kessler
%  Added function PARCELP (maybe should be some place else?).  FIxed to
%  lapmovelength to know about parcels.
% 28-Nov-83 08:33:07 (Robert Kessler)
%  Changed parceladdressvalue and addressvalue to use the new 2nd argument.
% 8/16/83 - Kessler
%  Fixed AddressValue problem.
%  To handle the problem with Load and Store having either 2 or 3
%  arguments, we made it an opcodefexprfunction, and break up the
%  operands as appropriate.
% D-Dawson, 11 August 1983
%  Op-Code tables completed.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Note form of DefOpcode, similar to DefAnyreg or DefCmacro
%      A trailing OpFailure will be appended unless (T is used)
%      Currently TESTs require arguments and explicit AND
 
% (DefOpCode OP (ARGS)
%   ( (tests) (actions))
%   ( (tests) (actions))
% ....
% )
 
(fluid '(*EMA))
 
(compiletime (load if-system))  % we need to distinguish between
                                % Cray1 and Cray X-MP
 
   % load of relocateable parceladdress into a-register
 
(defOpCode PLOC (r1 r2)
  ((and (aregp r1) (parceladdressvalue r2 t) (not *EMA))
        (Op-i-e 8#020 (aregp r1) (parceladdressvalue r2 nil)))
  ((and (aregp r1) (parceladdressvalue r2 t)) % EMA
        (Jumpop (plus2 8#010 (aregp r1))
                (plus2 (lshift 1 24) (parceladdressvalue r2 nil)))))
 
(put      'PLOC 'LapLength 2)
 
(defopcode  RMASK (r1 n)
   ((and(sregp r1)(jksizep n))
           (op-i-jk 8#042 (sregp r1)(jksizep (difference 64 n))))
)
(defOpCode    DLSHIFT  (r1 r2 a)
  ((and (Sregp r1)(Sregp r2)(Aregp a))
       (op-i-j-k 8#056 (Sregp r1)(Sregp r2) (Aregp a)))
  ((and (Sregp r1)(Sregp r2)(Onep a))
       (op-i-j-k 8#056 (Sregp r1)(Sregp r2) 0))     )
 
(defOpCode    DRSHIFT  (r1 r2 a)
  ((and (Sregp r1)(Sregp r2)(Aregp a))
       (op-i-j-k 8#057 (Sregp r1)(Sregp r2) (Aregp a)))
  ((and (Sregp r1)(Sregp r2)(Onep a))
       (op-i-j-k 8#057 (Sregp r1)(Sregp r2) 0))     )
 
% Mask Generation
%/ Check > or <
(defOpCode      LMASK    (r1 n)
 ((and (SregP r1) (JKsizeP n)) (Op-i-jk 8#043 (SregP r1) (JKsizeP n)))
)
 
% Sign Bit
 
(defOpCode      SBIT    (r1)
%      r1      SB
((sregP r1) (op-i-j-k 8#051 (sregP r1) 0 0))
)
(defOpCode      CMPSBIT    (r1)
%      r1      #SB
((sregP r1) (op-i-j-k 8#047 (sregP r1) 0 0))
)
 
% Sign Extend
 
(defOpCode      MOVESGN   (r1 r2)
%      r1      +r2
((and (SregP r1) (AregP r2))
 (op-i-j-k 8#071 (SregP r1) 1 (AregP r2)))
)
 
% Floating Move
 
(defOpCode      MOVEFL    (r1 r2)
%      r1      +Fr2
((and (SregP r1) (AregP r2))
 (op-i-j-k 8#071 (SregP r1) 2 (AregP r2)))
)
 
% Transmit constant 0.75*2**48 into Si, called 0.6 in CAL (0.75 = 0.6 octal)
(defOpCode      SCONST06    (r1)
%      r1      0.6
 ((SregP r1)
  (op-i-j-k 8#071 (SregP r1) 3 0))
)
 
(defOpCode      SCONST04    (r1)
%      r1      0.4 = 0.5
 ((SregP r1)
  (op-i-j-k 8#071 (SregP r1) 4 0))
)
 
(defOpCode      SCONST10    (r1)
%      r1      1.0
 ((SregP r1)
  (op-i-j-k 8#071 (SregP r1) 5 0))
)
 
(defOpCode      SCONST20    (r1)
%      r1      2.0
 ((SregP r1)
  (op-i-j-k 8#071 (SregP r1) 6 0))
)
 
(defOpCode      SCONST40    (r1)
%      r1      4.0
 ((SregP r1)
  (op-i-j-k 8#071 (SregP r1) 7 0))
)
 
%  Move Negative and Sign Extend
 
(defOpCode      MOVENEG    (r1 r2)
%      r1      -r2
((and (AregP r1) (AregP r2))
 (op-i-j-k 8#031 (AregP r1) 0 (AregP r2)))
((and (SregP r1) (SregP r2))
 (op-i-j-k 8#061 (SregP r1) 0 (SregP r2)))
)
 
% %%%%%%%%%%%%  Register Shifts %%%%%%%%%%%%%%
 
% r1 is r2 or S0, e could also be Ak
 
(defOpCode      LSHIFT  (r1 r2 e)
%      r1      r2<e
((and (Sreg0P r1) (SregP r2) (JKsizeP e))
 (Op-i-jk 8#052 (Sregp r2) (JKsizeP e)))
((and (SregP r1) (equal r1 r2) (JKsizeP e))
 (Op-i-jk 8#054 (Sregp r1) (JKsizeP e)))
((and (SregP r1) (equal r1 r2) (AregP e))
 (Op-i-j-k 8#056 (Sregp r1) 0 (AregP e)))
)
 
(defOpCode      RSHIFT  (r1 r2 e)
%      r1      r2>e
((and (Sreg0P r1) (SregP r2) (JKsizeP e))
 (Op-i-jk 8#053 (Sregp r2) (difference 64  (JKsizeP e))))
((and (SregP r1) (equal r1 r2) (JKsizeP e))
 (Op-i-jk 8#055 (Sregp r1) (difference 64  (JKsizeP e))))
((and (SregP r1) (equal r1 r2) (AregP e))
 (Op-i-j-k 8#057 (Sregp r1) 0 (AregP e)))
)
 
% %%%%%%%%%%%%%% Integer Register Arithmetic %%%%%%
 
 
(defOpCode      ADD  (r1 r2 r3)
%      r1      r2+r3
  ((and (Aregp r1) (AregP r2) (AregP r3))
       (Op-i-j-k 8#030 (Aregp r1) (AregP r2) (AregP r3)))
  ((and (AregP r1) (AregP r2) (OneP r3))
       (Op-i-j-k 8#030 (AregP r1) (AregP r2) 0))
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#60 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      SUB  (r1 r2 r3)
%      r1      r2-r3
  ((and (Aregp r1) (AregP r2) (AregP r3))
       (Op-i-j-k 8#031 (Aregp r1) (AregP r2) (AregP r3)))
  ((and (AregP r1) (AregP r2) (OneP r3))
       (Op-i-j-k 8#031 (AregP r1) (AregP r2) 0))
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#061 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      MUL  (r1 r2 r3)
%      r1      r2*r3
  ((and (Aregp r1) (AregP r2) (AregP r3))
       (Op-i-j-k 8#032 (Aregp r1) (AregP r2) (AregP r3)))
 
)
 
% Floating Register Arithmetic
 
 
(defOpCode      FADD  (r1 r2 r3)
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#062 (Sregp r1) (SregP r2) (SregP r3)))
 
%      r1      r2+Fr3
)
 
(defOpCode      FSUB  (r1 r2 r3)
%      r1      r2-Fr3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#063 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      FMUL  (r1 r2 r3)
%      r1      r2*Fr3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#064 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      HMUL  (r1 r2 r3)
%      r1      r2*Hr3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#065 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      RMUL  (r1 r2 r3)
%      r1      r2*Rr3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#066 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      IMUL  (r1 r2 r3)
%      r1      r2*Ir3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#067 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      HRECIP  (r1 r2)
%      r1      /Hr2
  ((and (Sregp r1) (SregP r2) )
       (Op-i-j-k 8#070 (Sregp r1) (SregP r2) 0))
 
)
 
% %%%%%%%%%%%%%%%  Logical Operations
% OR
 
(defOpCode      OR  (r1 r2 r3)
%      r1      r2!r3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#051 (Sregp r1) (SregP r2) (SregP r3)))
)
 
% AND
 
(defOpCode      AND (r1 r2 r3)
%      r1      r2&r3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#044 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
% One's Complement of an S register
 
(defOpCode      NOT  (r1 r2)
%      r1      #r2
  ((and (Sregp r1) (SregP r2) )
       (Op-i-j-k 8#047 (Sregp r1) 0 (SregP r2) ))
)
 
% Complement and AND, only Si # sk & Sj
 
(defOpCode      CMPAND  (r1 r2 r3)
%      r1      #r2&r3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#045 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      XOR  (r1 r2 r3)
%      r1      r2\r3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#046 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
% Complemented
 
(defOpCode      CMPXOR  (r1 r2 r3)
%      r1      #r2\r3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#047 (Sregp r1) (SregP r2) (SregP r3)))
 
)
 
(defOpCode      MERGE  (r1 r2 r3)
%      r1      r2!r1&r3
  ((and (Sregp r1) (SregP r2) (SregP r3))
       (Op-i-j-k 8#050 (Sregp r1) (SregP r2) (SregP r3)))
)
 
(defOpCode      MOVSBIT (r1 r2)
%      r1      r2!r1&SB
  ((and (Sregp r1) (SregP r2) )
       (Op-i-j-k 8#050 (Sregp r1) (SregP r2) 0))
)
 
% Some others with VM ?
 
% %%%%%%%%%%%%%  Loads and Stores %%%%%%%%%%%%%%%%%%%%%%%
 
(put 'STORE 'OpCodeFexprFunction 'storeOpCodeFexpr)
 
(de storeOpCodeFexpr (AllArgs)
 (prog (r1 e a2)
   (setq r1 (first AllArgs))
   (cond ((equal (length AllArgs) 2)
               (cond ((eqcar (second AllArgs) 'Indexed)
                        (setq a2 (second (second AllArgs)))
                        (setq e (third (second AllArgs))))
                     (T (setq a2 '(reg a0))
                        (setq e (second AllArgs)))))
         (T (setq e (second AllArgs))
                (setq a2 (third AllArgs))))
 (return 
  (cond
    %      e,a2   r1
    ((and (AregP r1) (AregP a2) (AddressValue e T))
      (Op-h-i-e 8#11 (AregP a2) (AregP r1) (AddressValue e nil)))
    ((and (SregP r1) (AregP a2) (AddressValue e T))
      (Op-h-i-e 8#13 (AregP a2) (SregP r1) (AddressValue e nil)))
    (T (OpFailed 'Store))))))
 
(put 'STORE 'LapLength 2)
 
(put 'LOAD 'OpCodeFexprFunction 'loadopCodeFexpr)
 
(de LoadOpCodeFexpr (AllArgs)
  (prog (r1 e a2)
    (setq r1 (first AllArgs))
    (cond ((equal (length AllArgs) 2)
           (cond ((eqcar (second AllArgs) 'Indexed)
                       (setq a2 (second (second AllArgs)))
                       (setq e (third (second AllArgs))))
                 (T (setq a2 '(reg a0))
                    (setq e (second AllArgs)))))
         (T (setq e (second AllArgs))
            (setq a2 (third AllArgs))))
 (return 
  (cond
    %      r1    e,a2
    ((and (AregP r1) (AregP a2) (AddressValue e T))
      (Op-h-i-e 8#10 (AregP a2) (AregP r1) (AddressValue e nil)))
    ((and (SregP r1) (AregP a2) (AddressValue e T))
      (Op-h-i-e 8#12 (AregP a2) (SregP r1) (AddressValue e nil)))
    (T (OpFailed 'Load))))))
 
(put 'LOAD 'LapLength 2)
 
% block load store of T and B
 
(defOpCode      BLKSTO  (a bt)
%      0,A0    bt,a
( (and (AregP a) (BregP bt))
 (op-i-jk 8#035 (AregP a) (BregP bt)))
( (and (AregP a) (TregP bt))
 (op-i-jk 8#037 (AregP a) (TregP bt)))
)
 
(defOpCode      BLKLOD  (a bt)
%      bt,a    0,A0
( (and (AregP a) (BregP bt))
 (op-i-jk 8#034 (AregP a) (BregP bt)))
( (and (AregP a) (TregP bt))
 (op-i-jk 8#036 (AregP a) (TregP bt)))
)
 
 
(defOpCode   ERR (ijk)
 ((IJKsize ijk)   (Op-ijk 8#000 ijk))
)
 
% Omit 001
 
(defOpCode   EFI ()
      (t   (Op-ijk 8#002 8#100)))
(defOpCode   DFI ()
      (t   (Op-ijk 8#002 8#200)))
 
(if_system Cray1 NIL   % not supported
   (progn
     (defOpCode   ERI ()
           (t   (Op-ijk 8#002 8#300)))
     (defOpCode   DRI ()
           (t   (Op-ijk 8#002 8#400)))
)  )
 
(defOpCode   SMTS (jk)        % test and set semaphore
 ((semaphorep jk) (Op-i-jk 8#003 4 (semaphorep jk))))
 
(defOpCode   SMCL (jk)        % clear semaphore
 ((semaphorep jk) (Op-i-jk 8#003 6 (semaphorep jk))))
 
(defOpCode   SMST (jk)        % set semaphore
 ((semaphorep jk) (Op-i-jk 8#003 7 (semaphorep jk))))
 
(defOpCode   EX (ijk)
 ((IJKsize ijk) (Op-ijk 8#004 ijk)))
 
(defOpCode   J (eb)
 ((BregP eb) (Op-ijk 8#005 (BregP eb)))
 ((ParcelAddressValue eb T) (JumpOp 8#006 eb)))
 
(put 'J 'LapLength 'LapJumpLength)
 
(de LapJumpLength (U)
  (prog (eb)
    (setq eb (second U))
    (return
      (cond
 ((BregP eb) 1)
 ((or (idp eb) (posintp eb) (stringp eb) (eqcar eb 'internalentry)
  (posintp (eval eb))) 2)
 (T (printf "Can't find length of case: %p%n" U)))
      )
    )
  )
 
(defOpCode   R (e)
 ((ParcelAddressValue e T) (JumpOp 8#007 e))
)
 
(defOpCode   JAZ (e)
 ((ParcelAddressValue e T) (JumpOp 8#010 e))
)
 
(defOpCode   JAN (e)
 ((ParcelAddressValue e T) (JumpOp 8#011 e))
)
 
(defOpCode   JAP (e)
 ((ParcelAddressValue e T) (JumpOp 8#012 e))
)
 
(defOpCode   JAM (e)
 ((ParcelAddressValue e T) (JumpOp 8#013 e))
)
 
(defOpCode   JSZ (e)
 ((ParcelAddressValue e T) (JumpOp 8#014 e))
)
 
(defOpCode   JSN (e)
 ((ParcelAddressValue e T) (JumpOp 8#015 e))
)
 
(defOpCode   JSP (e)
 ((ParcelAddressValue e T) (JumpOp 8#016 e))
)
 
(defOpCode   JSM (e)
 ((ParcelAddressValue e T) (JumpOp 8#017 e))
)
 
(put 'R   'LapLength 2)
(put 'JAZ 'LapLength 2)
(put 'JAN 'LapLength 2)
(put 'JAP 'LapLength 2)
(put 'JAM 'LapLength 2)
(put 'JSZ 'LapLength 2)
(put 'JSN 'LapLength 2)
(put 'JSP 'LapLength 2)
(put 'JSM 'LapLength 2)
 
% Register r2 to register r1, or const
 
(defOpCode      MOVE    (r1 r2)
%      r1      r2
 ((and (Aregp r1) (JKSizeP r2))
                (Op-i-jk 8#022 (Aregp r1) r2))
 ((and (Aregp r1) (MinusExpP r2))
                (Op-i-e 8#021 (Aregp r1) (TwosComp r2)))
 ((and (Aregp r1) (AddressValue r2 T))
                (Op-i-e 8#020 (AregP r1) (AddressValue r2 nil)))
 ((and (Aregp r1) (Sregp R2))
         (Op-i-j-k  8#023 (AregP r1) (SregP r2) 0))
 ((and (Aregp r1) (Bregp R2))
         (Op-i-jk  8#024 (AregP r1) (BregP r2) ))
 ((and (Bregp r1) (Aregp R2))
         (Op-i-jk  8#025 (AregP r2) (BregP r1) ))
 ((and (Aregp r1) (Aregp r2))
                (Op-i-j-k 8#030 (Aregp r1) 0 (Aregp r2)))
 ((and (Sregp r1) (Onep r2))
                (Op-i-jk 8#042 (Sregp r1) 8#77))
 ((and (Sregp r1) (MinusOneP r2))
                (Op-i-jk 8#042 (Sregp r1) 8#00))
 ((and (Sregp r1) (Zerop r2))
                (Op-i-jk 8#043 (Sregp r1) 8#00))
 ((and (Sregp r1) (MinusExpP r2))
                (Op-i-e 8#041 (Sregp r1) (TwosComp r2)))
 ((and (Sregp r1) (AddressValue r2 T))
                (Op-i-e 8#040 (Sregp r1) (AddressValue r2 nil)))
 ((and (Sregp r1) (Sregp r2))
                (Op-i-j-k 8#051 (Sregp r1) 0 (Sregp r2)))
 ((and (Sregp r1) (Aregp r2))
                (Op-i-j-k 8#071 (Sregp r1) 0 (Aregp r2)))
 ((and (Sregp r1) (Tregp r2))
                (Op-i-jk 8#074 (Sregp r1) (Tregp r2)))
 ((and (Tregp r1) (Sregp r2))
                (Op-i-jk 8#075 (Sregp r2) (Tregp r1)))
)
 
(put 'MOVE 'LapLength 'LapMoveLength)
 
% End of MOVE definition table
%
% Bit count
 
(defOpCode      ONEPOP  (a s)
 ((and (AregP a) (SregP s))
           (op-i-j-k 8#026 (aregP a) (sregp s) 0))
)
 
%      a       Ps
 
(defOpCode      PARITYPOP (a s)
 ((and (AregP a) (SregP s))
           (op-i-j-k 8#026 (aregP a) (sregp s) 1))
)
 
%     a        QS
 
(defOpCode      ZEROPOP  (a s)
 ((and (AregP a) (SregP s))
           (op-i-j-k 8#027 (aregP a) (sregp s) 0))
%      a       Zs
)
%%%%%%%%%%%%%% shared registers  and semaphores %%%%%%%%%%%%%%%%%%%%%%%
 
(defOpCode      GETSM  (i)             %  Si  SM
 ((sregp i) (op-i-j-k 8#072 (Sregp i) 0 2)))
 
(defOpCode      SETSM  (i)             %  SM  Si
 ((Sregp i) (op-i-j-k 8#073 (Sregp i) 0 2)))
 
(defOpCode      GETSB  (i j)           %  Ai  SBj
 ((and (aregp i) (shared-bregp j))
                    (op-i-j-k 8#026 (aregp i) (shared-bregp j) 7)))
 
(defOpCode      SETSB  (i j) %  SBj Ai
 ((and (shared-bregp i) (aregp j))
                    (op-i-j-k 8#027 (aregp j) (shared-bregp i) 7)))
 
(defOpCode      GETST  (i j) %  Si  STj
 ((and (Sregp i) (shared-Tregp j))
                    (op-i-j-k 8#072 (Sregp i) (shared-Tregp j) 3)))
 
(defOpCode      SETST  (i j) %  STj Si
 ((and (shared-Tregp i) (sregp j))
                    (op-i-j-k 8#073 (Sregp j) (shared-Tregp i) 3)))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(de ParcelP (u)
  (eqcar u 'Parcel))
 
(de LapMoveLength (U)
 (prog (r1 r2)
  (setq r1 (second U))
  (setq r2 (third U))
  (return
   (cond 
     ((and (Aregp r1) (JKSizeP r2))
          1)
     ((and (Aregp r1) (MinusExpP r2))
          2)
     ((and (Aregp r1) (ParcelP r2))
          2)
     ((and (Aregp r1) (AddressValue r2 T))
          2)
     ((and (Aregp r1) (Sregp R2))
          1)
     ((and (Aregp r1) (Bregp R2))
   1)
     ((and (Bregp r1) (Aregp R2))
          1)
     ((and (Aregp r1) (Aregp R2))
          1)
     ((and (Sregp r1) (Onep r2))
          1)
     ((and (Sregp r1) (MinusOneP r2))
          1)
     ((and (Sregp r1) (Zerop r2))
          1)
%     ((and (Sregp r1) (MinusExpP r2))
%          2)
     ((and (Sregp r1) (ParcelP r2))
          2)
     ((and (Sregp r1) (AddressValue r2 T))
        2)
     ((and (Sregp r1) (Sregp r2))
          1)
     ((and (Sregp r1) (Aregp r2))
          1)
     ((and (Sregp r1) (Tregp r2))
          1)
     ((and (Tregp r1) (Sregp r2))
          1)
     (T (printf "Can't find length of case: %p%n" U))
     )
   )
   )
 )
 
(de TwosComp (u)
    (wplus2 (wminus u ) -1))
%   (sub1 (minus u)))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %
% Pseudo instructions to control lapopt %
% %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(defOpCode  Lapopt-fence (eb)
 ((anyp eb) (anyp eb))
)
 
(put 'lapopt-fence 'LapLength 0)
 
