%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         CRAY-LAPOPT.SL
% Description:
% Author:       H. Melenk , ZIB Berlin
% Created:      10-Aug-86
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LAP pass 1 appendix : instruction schedule for CRAY
% dependent of global switch *LAPOPT
%              and options set in the list *LapOptOptions*
%
% option values:
%      1x : control of unrolling
%            10 no unrolling
%            11 partial unrolling
%            12 full unrolling level 1
%      2x : control of load optimization
%            20 no load optimization
%            21 full load optimization
%      3x : control of instrcution scheduling
%            30 no instrcution scheduling
%            31 scheduling of load instrcutions
%            32 full scheduling
%
% default: (setq *LapOptOptions* 11 21 32)
% requires loadtime loading of CRAY-LAPOPT-2
%                              CRAY-LAPOPT-3
%                              CRAY-LAPOPT-4
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (off lapopt))

(compiletime (load carcdrnil))

(imports '(cray-lapopt-2 cray-lapopt-3 cray-lapopt-4))
 
(fluid '(*LAPOPT PeeplapDirection* *VectorInstructions* *LapOptTrace
         &Opt-Tregs& &opt-highregs& &softjumps&
         *R-CONS-RECLAIM* *LapOptOptions*))
 
(setq *LAPOPT T)          % default: Lap optimization active
 
(setq *LapOptOptions* '(
              11          % do partial unrolling
              21          % do load optimization
              32 ))       % do full instruction scheduling
 
(setq *LapOptTrace nil)   % default: no trace printed
 
                          % registers potentially free for local
                          % optimizations.
                          % t60 - 767 not included; they are
                          % used for partially global optimizations
                          % e.g. inside the code of compacting-gc
(setq &Opt-Tregs& '(
(Reg T30)(Reg T31)(Reg T32)(Reg T33)(Reg T34)(Reg T35)(Reg T36)(Reg t37)
(Reg T40)(Reg T41)(Reg T42)(Reg T43)(Reg T44)(Reg T45)(Reg T46)(Reg t47)
(Reg T50)(Reg T51)(Reg T52)(Reg T53)(Reg T54)(Reg T55)(Reg T56)(Reg t57)
))
 
(setq &Opt-Highregs& '(
(Reg T71)(Reg T72)(Reg T73)(Reg T74)(Reg T75)(Reg T76)))
 
(setq *R-CONS-RECLAIM*
   `(R (plus (parcel symfnc) ,(wtimes2 (id2int 'cons-reclaim) 4))))
 
(setq &softjumps&
  `((R (plus2 (parcel symfnc) ,(wtimes2 (id2int 'cons-reclaim) 4)))
    (R (plus2 (parcel symfnc) ,(wtimes2 (id2int 'bstackoverflow) 4) ))
    (R (plus2 (parcel symfnc) ,(wtimes2 (id2int 'bstackunderflow) 4)))))
 
 
% routines for code optimization
%
% this routines takes an instduction list as input and returns
% a reordered one. The pass is inserted between PASS1LAP and
% deposit Instructions
 
(compiletime (progn (flag '(
                           ImportantOperands OperandCollisionP
                           limit1p
                           LapOptLoads1 LapOptLoads2
                           LapOpt1aux
                           limit2p PushInstr) 'iInternalFunction) nil))
 
% load optimization : auxilary function for register history
 
(ds optregp (x) (and (eqcar x 'reg) (not (equal x '(Reg S0)))
                                    (not (equal x '(Reg A0)))))
 
                         % store register contents:
                         % - delete old contents
                         % - add new contents
                         % if reference is to a S/T register, take the
                         % contents of that register too
(de set-val (reg val db)
         (if
          (not (optregp reg))
          db
          (progn
             (setq db (delete-vals reg db))
             (setq db (cons (cons reg val) db))
             (when (optregp val)
                (when (find-val-of-reg val db)
                   (setq db (cons
                              (cons reg (car (find-val-of-reg val db))) db))))
                (setq db (append db (list (cons val reg) )))
             db
)        ))
 
 
                                    % same as set-val without deleting
                                    % old register contents; e.g. used
                                    % for store (store does not destroy
                                    % the source register)
(de set-val-preserving (reg val db)
         (if
          (not (optregp reg))
          db
          (progn
             (setq db (cons (cons reg val) db))
             (when (optregp val)
                (when (find-val-of-reg val db)
                   (setq db (cons
                              (cons reg (car (find-val-of-reg val db))) db))))
                (setq db (append db (list (cons val reg) )))
             db
)        ))
 
(de delete-vals (reg db)
       (cond ((null db) nil)
             ((membl reg (caar db)) (delete-vals    reg (cdr db)))
             ((membl reg (cdar db)) (delete-vals    reg (cdr db)))
             (t (cons (car db)(delete-vals reg (cdr db)))) ))
 
(de DeleteAregRefs (db)
       (cond ((null db) nil)
             ((MembAreg (caar db)) (deleteAregrefs (cdr db)))
             ((MembAreg (cdar db)) (deleteAregrefs (cdr db)))
             (t (cons (car db)(deleteAregRefs  (cdr db)))) ))
 
(de membl (reg db)
       (cond ((null db) nil)
             ((atom db) (eq reg db))
             ((equal reg db) t)
             ((membl reg (car db)) t)
             (t        (membl reg (cdr db)))  ))
 
(de membAreg (db)
       (cond
             ((atom db) nil)
             ((Aregp db) t)
             ((regp db) nil)
             ((membAreg (car db)) t)
             (t        (membAreg (cdr db)))  ))
 
(de find-val-of-reg (reg db)
       (cond ((null db) nil)
             ((equal (caar db) reg) (list (cdar db)))
             (t (find-val-of-reg  reg (cdr db)))) ))
(de find-sreg-with-val (val db)
       (cond ((null db) nil)
             ((and (equal (cdar db) val) (sregp (caar db))) (caar db))
             (t (find-sreg-with-val val (cdr db)))) ))
(de find-areg-with-val (val db)
       (cond ((null db) nil)
             ((and (equal (cdar db) val) (aregp (caar db))) (caar db))
             (t (find-areg-with-val val (cdr db)))) ))
 
 
% lap optimization 1: delete unnecessary register moves
%                 and storage accesses
%
% technique:   tracing the register contents in a register database
%              which is updated for each operation modifying
%              a register and which is deleted if the register
%              becomes unknown (e.g. label as target of a jump)
 
(de LapOptLoads1 (code)
  (prog (
 
        Rcode         % instructions processed already
        Instr         % instruction under investigation
        NextInstr     % next instruction
        PrevInstr     % previous instruction
        Ops           % operands of current instruction
        Dest          % destination of current instruction
        Source        % source of current instruction
        AltSource     % alternate source proposal
        Db            % database for register bookkeeping
        FluidDb       % database of fluids held in T regs
        AllTregs      % list of generally available T registers
        TregsLeft     % T registers not yet in use
        LabelBackup   % stack for backups
        x             % for local usage
        LabelDb       % databases valid from label positions on
        Trace         % trace status
        Count         % counting
        )
 
     (setq trace *LapOptTrace)
     (setq TregsLeft (setq AllTregs (FreeTregs code)))
     (setq Count 0)
 
 LOOP
     (cond ((null code)(go ready)))
     (setq PrevInstr Instr)
     (setq Instr (car Code))           % instruction under investigation
     (setq Code (cdr Code))
     (cond
        ((atom instr)(go label))
        ((eq (setq x (car instr)) 'MOVE) (go move))
        ((eq x                    'STORE) (go store))
        ((eq x                    'LOAD) (go load))
        ((eq x                    'R   ) (go nonlocal-jump))
        ((eq x            'Lapopt-fence) (go nonlocal-jump))
        ((memq x '(J JAZ JAN JAP JAM JSZ JSN JSP JSM))
             (if (atom (cadr Instr))
                                         (go local-jump)
                                         (go nonlocal-jump)))
        ((eq x                    'BLKLOD) (go blklod))
        (t           (go continue)))   % default: no action
 
 MOVE
                                       % first test special case
                                       % (MOVE (REG S6) ....)
                                       % (MOVE (REG A..)(REG S6))
     (setq NextInstr (car Code))
     (setq source(caddr instr))
     (setq dest (cadr instr))
     (when (and
              (Sregp Dest)
              (eq (car NextInstr) 'MOVE)
              (Aregp   (cadr NextInstr))
              (optregp (cadr NextInstr)) % fixes a bug regaring A0
              (equal Dest (caddr NextInstr))
              (setq altsource (Find-Areg-With-Val Source db))
              (SregNotUsed (cdr Code) Dest)    )
           (go special-move)   )
     (setq x (setq altsource nil))
     (when (optregp dest)
        (when (aregp dest)
           (setq altsource (find-areg-with-val source db))
           (when (aregp source) (go do-not))
           (when (and (not altsource) (find-val-of-reg source db))
                 (setq altsource (find-areg-with-val
                                 (car(find-val-of-reg source db))db))))
        (when (sregp dest)
           (setq altsource (find-sreg-with-val source db))
           (when (sregp source) (go do-not))
           (when (and (not altsource) (find-val-of-reg source db))
                 (setq altsource (find-sreg-with-val
                                 (car(find-val-of-reg source db))db))))
     )
     (when altsource
       (if (equal altsource source)      % no replacement necessary
         (setq x code)
         (setq x (catch 'lapopt1 (lapopt1aux code dest altsource)))))
 
     (when x (setq code x) % replacement done
             (when trace
                 (when (not (equal source altsource))
                       (prin2 "-----------------------replaced---")
                       (prin2 dest)
                       (prin2 "--by--")
                       (prin2t altsource) )
                  (prin2 "----one instruction deleted------->>>  ")
                  (prin2t instr))
             (go loop))
 do-not                                % no action for move/load
     (setq db (set-val dest source db))
     (go continue-2)
 
 special-move                            % move to areg via sreg
                                         % full testing done already
     (setq Dest (cadr NextInstr))
 
     (if (equal Dest AltSource)
 
      (setq x (cdr code))        % no replacement necessary
 
      (when
       (and
        (setq x (catch 'lapopt1 (lapopt1aux (cdr code) dest
                                                       altsource)))
        (Sregp (caddr Instr)) )
                           % replace intermediate reg too
       (setq x (catch 'lapopt1 (lapopt1aux x (cadr Instr)
                                             (caddr Instr))))
     ))
     (when x (setq code x) % replacement done
             (when trace
                 (when (not (equal Dest AltSource))
                       (prin2 "-----------------------replaced---")
                       (prin2 dest)
                       (prin2 "--by--")
                       (prin2t altsource))
                 (prin2 "-----two instructions deleted----->>>  ")
                 (prin2t Instr)
                 (prin2 "---------------------------------->>>  ")
                 (prin2t NextInstr) )
             (go loop))
 LOAD
     (setq source(cddr instr))
     (setq dest (cadr instr))
     (setq x (setq altsource nil))
 
     (when (aregp dest)
           (cond
              ((setq altsource (find-areg-with-val source db)))
              ((setq altsource (find-sreg-with-val source db)))
              ((setq x (assoc (car Source) FluidDb))
                               (go load-areg-from-fluid-treg))
              ((find-val-of-reg source db)
                 (setq altsource (find-areg-with-val
                                 (car(find-val-of-reg source db))db))
     )     )  )
 
     (when (sregp dest)
           (setq altsource
               (if
                    (setq x (assoc (car source) FluidDb))
                    (cdr x)
                    (find-sreg-with-val source db)))
           (when (and (not altsource) (find-val-of-reg source db))
                 (setq altsource (find-sreg-with-val
                                 (car(find-val-of-reg source db))db))))
     (when altsource
            (setq Count (Wplus2 Count 1))
            (when trace
                     (prin2 "--instruction-->")(prin2 instr))
                            % do not generate (MOVE (REG 1)(REG 1))
            (when (equal altsource dest)
                  (when trace (prin2t "--- deleted"))
                  (go loop))
            (setq instr (list 'MOVE dest altsource))
            (setq source altsource)
            (when trace (prin2t "--replaced by following instruction"))
                                                    )
 
     (when                                  % load of fluid: store local
         (and (not AltSource)
              (pairp (car Source))
              (eq (caar Source) '$FLUID)
              (Sregp Dest)
              TregsLeft
              (TestUsabilityOfFluidValue code 10 source))
         (setq x (car TregsLeft))
         (setq TregsLeft (cdr TregsLeft))
         (setq code
            (cons
               (list 'MOVE x dest)
               code))
         (setq FluidDb
               (cons
                  (cons (car source) x)
                  FluidDb))
         (when trace
               (prin2 "--inserted instruction -->")
               (prin2 (car code))
               (prin2t "-- after next")   )    )
 
load-done
 
     (setq db
         (if (membl dest source)
             (delete-vals dest db)          % clear recursive refs
             (set-val dest source db)))     % store new ref
 
     (go continue-2)
 
load-areg-from-fluid-treg
 
     (when (not TregsLeft) (go load-done))  % replacement impossible
     (setq rcode                            % rcode backwards
           (nconc
             (list
               (list 'move dest '(Reg s7))
               (list 'move '(Reg s7) (cdr x))
               (list 'move (car TregsLeft) '(Reg s7)) %save S7
             )
             rcode))
     (when trace
           (prin2 "--- replaced --->>")
           (prin2 Instr)
           (prin2t "--- by following 4 instructions")
           (prin2t (caddr Rcode))
           (prin2t (cadr Rcode))
           (prin2t (car Rcode)))  % last will be printed by standard
     (setq Instr(list 'move '(reg s7) (car TregsLeft)))% restore S7
     (go load-done)
 
 BLKLOD
                                       % blklod overwrites the low
                                       % t registers (frame)
     (when FluidDb
          (setq FluidDb
             (mapcan
                 FluidDb
                 (Function (lambda (x)
                              (if
                                 (member (cdr x) &Opt-Highregs&)
                                 (cons x nil)
                                 nil
     )    )  )   )         )  )
     (go continue-2)
 
 STORE
                                       % remove old references to this
                                       % memory cell
     (setq db (delete-vals (cddr instr) db))
                                       % mark the source reg to hold
                                       % the value of the memory cell
     (setq db (set-val-preserving (cadr instr) (cddr instr) db))
     (setq Source (cadr instr))
     (setq Dest (caddr instr))
 
                                       % known fluid overwritten?
                                       % if yes, update Treg too
     (when (assoc Dest FluidDb)
           (setq code
               (cons
                  (list 'MOVE (cdr (assoc Dest FluidDB)) Source)
                  code))
           (when trace
               (prin2 "--inserted instruction -->")
               (prin2 (car code))
               (prin2t "-- after next")   )    )
 
     (go continue-2)
 
 NONLOCAL-JUMP
 
     (setq Db (DeleteAregRefs Db))
     (when (member Instr &softjumps&) (go continue-2))
     (setq Db nil)
     (setq FluidDb nil)
     (setq TregsLeft AllTregs)
     (go continue-2)
 
 LOCAL-JUMP                            % save databases for target
 
     (when (setq x (atsoc (cadr Instr) LabelBackup))
           (go critical-jump-backwards))
 
     (setq LabelDb
        (cons
          (cons (cadr Instr)           % label value
                (list Db               % register database
                      FluidDb          % fluids database
                      TregsLeft))      % free Tregs
          LabelDb))
     (go continue-2)
 
critical-jump-backwards
                                       % test compatibility of
                                       % FluidDbs
     (when
         (compatiblep (cadr x) FluidDb)% anything ok
         (go continue-2))
               % B A C K U P           % BACKUP necessary
                                       % with reduced FluidDb
                                       % restore operating env
     (setq Instr (car x))
     (setq Code (caddr x))
     (setq Rcode (cadddr x))
     (setq x (cddddr x))
     (setq LabelDb (car x))
     (setq LabelBackup (cadr x))
     (setq Count (caddr x))
                                             % rerun with reduced
     (setq FluidDb nil)                      % fluids database
     (setq Db nil)
     (setq TregsLeft AllTregs)
     (when trace (prin2 " ===== Backup necessary ===== from ")
                 (prin2t Instr))
 
     (go continue-2)
                                       % label
 LABEL
 
     (if                               % no backwards jump: restore dbs
       (setq x (CompressLabelDb Instr
                                (if
                                   (eqcar PrevInstr 'J)
                                   nil       % no db from above
                                   (list Db FluidDb TregsLeft)  )
                                LabelDb))
       (progn
             (setq Db (car x))
             (setq FluidDb (cadr x))
             (setq TregsLeft (caddr x)) )
       (progn
             (setq Db nil)
             (setq FluidDb nil)
             (setq TregsLeft AllTregs)  )
     )
 
     (when (not (NoJumpTo Instr (cdr code)))
           (go label-with-backward-reference))
 
     (go continue-2)
 
 label-with-backward-reference
 
     (setq Db nil)                     % not worth to be saved
     (when (not FluidDb) (go continue-2))
                                       % save the whole environment
                                       % to be backed up later
                                       % in case of a jump to this
                                       % label with incompatible
                                       % FluidDb
     (setq LabelBackup
        (cons
             (list
                 Instr                 % label
                 FluidDb
                 Code
                 Rcode
                 LabelDb
                 LabelBackup           % looks recursive, is not
                 Count)
             LabelBackup))
     (go continue-2)
 
 CONTINUE                              % default: proceed without action
     (when (cdr instr) (setq db (delete-vals (cadr instr) db)))
 
 continue-2
                                       % put actual instr on to output
     (setq rcode (cons instr rcode))
     (when trace (prin2t instr))
     (go loop)
 
 
 ready                                 % reverse instr list and delete
     (return (reverse rcode))
))
 
     % replace source by altsource whenever possible
     % if source needed longer than basic block or altsource
     % is destroyed before source will be overwritten, nil will
     % be returned immediately
 
(de lapopt1aux (code source altsource)
     (cond ((null code)           % failure if result register
                    (if (equal source '(reg 1))
                        (throw 'lapopt1 nil)
                        nil))
 
           ((equal (car code)'(J (Reg B00))) % failure if result-register
                    (if   (equal source '(reg 1))
                          (throw 'lapopt1 nil)
                          code ))
 
      %    ((equal (car code) *R-CONS-RECLAiM*) % all registers critical r
      %                   (throw 'lapopt1 nil))
      %
 
           ((limit1p (car code))  % failure if "LISP"-register
                    (if   (member source
                                '((Reg 1)(Reg 2)(Reg 3)
                                  (Reg 4)(Reg 5)))
                          (throw 'lapopt1 nil)
                          code ))
 
                                        % lshift and rshift work in place
                                        % if source or alts. involved,
                                        % alternative no longer available
           ((and (memq (caar code) '(LSHIFT RSHIFT DLSHIFT DRSHIFT))
                 (or (equal (cadar code) source)
                     (equal (cadar code) altsource) ))
                          (throw 'lapopt1 nil)        )
 
                                        % end of register range reached
           ((and(equal (cadar code) source)
                (not (eq (caar code) 'STORE)))
                  (cons
                        (cons (caar code)
                              (cons (cadar code)
                                    (subst altsource source (cddar   code)) ))
                        (cdr code) ))
 
                                        % alternative no longer available
           ((equal (cadar code) altsource) (throw 'lapopt1 nil))
 
                                        % continue
           (t (cons (subst altsource source (car code))
                    (lapopt1aux (cdr code) source altsource))) ))
 
 
                   % test, if the code depends on the current
                   % contents of the sreg
(de SregNotUsed (code sreg)
     (cond ((null code)           % failure if result register
                    (if (equal sreg '(reg 1))
                        nil
                        t                ))
 
           ((equal (car code)'(J (Reg B00))) % failure if result-register
                    (if   (equal sreg '(reg 1))
                          nil
                          t                 ))
 
           ((equal (car code) *R-CONS-RECLAiM*) % all registers critical r
                          nil)
 
 
           ((limit1p (car code))  % failure if "LISP"-register
                    (if   (member sreg
                                '((Reg 1)(Reg 2)(Reg 3)
                                  (Reg 4)(Reg 5)))
                          nil
                          t                  ))
 
                                        % lshift and rshift work in place
                                        % if sreg  involved,
                                        % alternative no longer available
           ((and (memq (caar code) '(LSHIFT RSHIFT DLSHIFT DRSHIFT))
                 (equal (cadar code) sreg))
                          nil )
 
                                        % end of possible register
                                        % range found: reg is overwritten
           ((and(equal (cadar code) sreg)
                (not (eq (caar code) 'STORE)))
                          t  )
 
                                        % register cited
           ((member sreg (cdar code)) nil)
 
                                        % continue
           (t       (SregNotUsed (cdr code) sreg )) ))
 
 
                   % test, if a fluid will be be needed once more
 
 
                            % test end of basic block for optimization
                            % of type 1
(de limit1p (instr)
   (or   (atom instr)            % a label
         (memq   (car instr) % a critical instruction
                 '(*ENTRY LABEL BLKLOD BLKSTO
                    J R JAZ JAN JAM JSP JSM
                        JSZ JSN
                        lapopt-fence)
         )
         ) )
 
(fluid '(spec-Nalloc*))
 
(de FreeTregs (code)
                              % use fast frames first
  (prog (regs)
    (when  (and (wlessp spec-Nalloc* 6)
                (wgeq spec-Nalloc* 0))
             (setq regs (pnth &opt-highregs&
                                 (wplus2 spec-NAlloc* 1))))
    (setq spec-Nalloc* 4711)
    (return regs)))
 
                              % build a list of free T registers
(commentoutcode               % for temporary storage purposes
(de FreeTregs (code)
   (prog (n instr registers)
                              % use fast frames first
         (when  (and (wlessp Nalloc* 6)
                     (wgeq Nalloc* 0))  % NOT SAVE with new Modifyallocs
                (setq registers (pnth &opt-highregs&
                                              (wplus2 NAlloc* 1))) )
 
search                        % search framesize
         (when (null code) (return registers))
         (setq instr (car code))
         (when
            (and
                (eqcar instr 'move)
                (equal (cadr instr) '(Reg A3))
                (numberp (setq n (caddr instr))) )
                              % framesize found
            (setq registers
              (append
                registers
                (if (greaterp n 20)
                    nil
                    (pnth &Opt-Tregs& (wplus2 1 n))  )))
            (return registers) )
         (setq code (cdr code))
         (go search)))
) % end comment out
                            % investigate the code to be free of
                            % count jumps to the label
(de NoJumpTo (Label Code)
       (cond
         ((null Code) t)
         ((atom (car Code)) (NoJumpTo Label (cdr Code)))
         ((eq label (cadar code)) nil)
         (t           (NoJumpTo Label (cdr Code))) ))
 
 
                    % find intersection of databases from jups to
                    % one label
                    % db=nil on input: no prerequisits for dbs
                    % otherwise the actual databases are in db
(de CompressLabelDb (label db labeldb)
  (let (db1 db2 db3 x)
       (cond
         ((null labeldb) db)
         ((eq label (caar labeldb))
             (cond
               ((null db)
                (CompressLabelDb label (cdar LabelDb) (cdr LabelDb)))
               (t
                (setq x (cdar LabelDb))
                (setq db1 (intersection (car x) (car db)))
                (setq db2 (intersection (cadr x) (cadr db)))
                (setq db3 (intersection (caddr x) (caddr db)))
                (CompressLabelDb label (list db1 db2 db3)
                                       (cdr LabelDb))
         )   ) )
         (t (CompressLabelDb label db (cdr LabelDb)))  )))
 
(de compatiblep (db1 db2)  % is in fact a test for subset property
   (cond 
       ((null db1) t)
       ((member (car db1) db2) (compatiblep (cdr db1) db2))
       (t nil) ))
 
      % does the situation justify the saving of a fluid value?
(de TestUsabilityOfFluidValue (code n object)
 (let (instr opcode)
   (cond 
     ((null code) nil)
     ((eq n 0) t)
     ((atom (setq instr (car code)))              t) % label
     ((and (eq (setq opcode (Car instr)) 'LOAD)      % a second load
           (equal (cddr instr) object))           t) % of the same obj.
     ((eq opcode 'R)                              nil) % returnjump
     ((and (eq opcode 'J)                              % exit jumps
           (or (eq (caadr instr) 'plus2)
               (equal (cadr instr) '(Reg B00)) )) nil)
     (t (TestUsabilityOfFluidValue (cdr code)(wplus2 n -1) object))  )))
 
% load optimization 2: move load instructions backwards as
% far as possible and the instructions using the loaded information
% forwards
 
(de LapOptLoads2 (code)
  (prog (rcode instr ops)
 
 loop
     (cond ((null code)(go ready)))
     (setq instr (car code))           % instruction under investigation
     (setq code (cdr code))
     (cond
        ((atom instr)(go continue))    % label: no action
        ((eq (car instr) 'LOAD) (go load))  % case of interes
        (t           (go continue)))   % default: no action
 
 
 load
     (setq ops (ImportantOperands instr)) % relevant ops of instr
     (setq PeeplapDirection* 'backward)
     (setq rcode                       % move load backwards
           (PushInstr (list (cons 'backward instr)) rcode ops))
     (setq PeeplapDirection* 'forward)
     (setq ops (list (cadr instr)))    % target of load op
     (cond                             % uses following instr same regs?
       ((and
         (not (limit2p (car code)))
         (OperandCollisionP ops (car code)) )
         (setq code                    % move this instr forwards
              (PushInstr (list (cons 'forward (car code)))
                         (cdr code)
                         (ImportantOperands (car code))
     )  ))    )
     (go loop)
 
 continue                              % default: proceed without action
     (setq rcode (cons instr rcode))
     (go loop)
 
 
 ready                                 % reverse instr list and delete
     (cond ((null rcode)(return code)))% forward and backward marks
     (setq instr (car rcode))
     (cond ((atom instr)(setq code (cons instr code)))
           ((memq   (car instr) '(forward backward))
                        (setq code (cons (cdr instr) code)))
           (t           (setq code (cons instr code)))  )
     (setq rcode (cdr rcode))
     (go ready)
))
 
% select the register operands from an instruction
 
(de ImportantOperands (l)
  (cond
     ((null l) nil)
     ((atom (car l)) (ImportantOperands (cdr l)))
     ((eq (caar l) 'REG) (cons (car l) (ImportantOperands (cdr l))))
     ((eq (caar l) 'INDEXED)
            (cons (cadar l) (ImportantOperands (cdr l))))
     (t (ImportantOperands (cdr l)))           ))
 
% test an instruction against collision with collected operands
 
(de OperandCollisionP (ops instr)
  (cond
    ((null instr) nil)
    ((atom instr) nil)
    ((atom (car instr)) (OperandCollisionP ops (cdr instr)))
    ((member (car instr) ops) t)
    ((and
       (eq (caar instr) 'indexed)
       (member (cadar instr) ops)
     )          t)
    (t (OperandCollisionP ops (cdr instr)))   ))
 
 
% test if an instruction is a segment limit against movement
 
(fluid '(AllocJAPinstr))
 
(setq AllocJAPinstr
    (list 'JAP
          (list 'PLUS2
                '(Parcel SYMFNC)
                (times2 (idloc stack-blown-up) 4)   )))
 
(de limit2p (instr)
   (or   (atom instr)            % a label
         (memq   (car instr) % a critical instruction
                 '(*ENTRY LABEL BLKLOD BLKSTO STORE
                    J R JAZ JAN JAM JSP JSM
                        JSZ JSN
                        lapopt-fence)
         )
         (memq   (car instr) *VectorInstructions*)
         (and
             (eq (car instr) 'JAP)  % JAP within ALLOC sequence:no limit
             (not (equal instr AllocJAPinstr))
         )
         (eq (car instr) peepLapDirection*) % do not overmove an instr
                                            % already moved
         (and (memq   (car instr) '(forward backward))
              (limit2p (cdr instr))         % case marked instr
         ) ))
 
% move an instruction group as far as possible
 
(de PushInstr (instrs code ops)
   (cond ((or
             (null code)
             (limit2p (car code))    % no more movement possible
          )                          % insert here
          (append (reverse instrs) code)   )
                                     % case instr uses critical operand
        ((OperandCollisionP ops (car code))
         (PushInstr
              (cons (car code) instrs)   % instr added to move group
              (cdr code)                 % insert into rest of code
              (union ops                 % new operand packet
                     (ImportantOperands (car code))
        ))    )
        (t                           % move around this instruction
            (cons (car code)
                  (PushInstr instrs (cdr code) ops)
)  )    )   )
 
% create connection to PASS1LAP
 
(copyd 'Pass1LapOnly 'Pass1Lap)
 
(remd 'Pass1Lap)
 
(de Pass1Lap (u)
  (prog()
   (setq u (Pass1LapOnly u))
   (cond ((not *LapOpt) (return u)))
 
   (if   (memq 11 *LapOptOptions*)
         (setq u (LapOptUnroll (LapOptReformJumps u) 0))
         (when (memq 12 *LapOptOptions*)
               (setq u (LapOptUnroll (LapOptReformJumps u) 1))))
 
   (when (memq 21 *LapOptOptions*)
      (setq u (LapOptLoads1 u)) )    % remove redundant loads
 
   (if   (memq 32 *LapOptOptions*)
         (setq u (LapOptSched u))      % total instructions schedule
 
         (when (memq 31 *LapOptOptions*)
             (setq u (LapOptLoads2 u)) ))   % partial schedule
   (return u)
) )
