%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PC:OPENCODEDFUNCTIONS.SL
%
% Description:  open-coded functions in PSL compiler (CRAY PSL3.4)
%
% Author:       W.Neun , ZIB Berlin
%
% Created:      28-May-1986
%
% Status:       Experimental (Do Not Distribute)
%
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 10 Nov 1987 W. Neun. Put in destroys nil for EFI and DFI
% 25 Sep 1987 W. Neun. Put in destroys nil for ERI and DRI
% These functions where tagged as open coded in the CRAY PSL3.4
%  BYTE
%  STRBYT
%  HALFWORD
%  PUTBYTE
%  PUTSTRBYT
%  PUTHALFWORD
 
%  LONGTIMES2         %Functions using 46-bit Integer Arithmetic
%  LONGQUOTIENT
%  LONGDIVIDE
%  LONGDIVIDE-QUOTIENT
%  LONGDIVIDE-REMAINDER
 
%  Clear-Extraregs
%  catchsetup
%  !%uncatch
%  idapplye
%  idapplye2
%  (enable-operand-range-interrupt)   % XMP only
%  (disable-operand-range-interrupt)  % XMP only
%  (enable-floatingpoint-interrupt)
%  (disable-floatingpoint-interrupt)
 
(compiletime (load if-system))
 
(deflist '((longtimes2                % Integer Multiplication
      ((sconst06  (reg 3))            % with 46-bit Integers
       (sub  (reg 2) (reg 3) (reg 2)) % using FloatingPoint Unit
       (sub  (reg 1) (reg 3) (reg 1))
       (fsub (reg 2) (reg 3) (reg 2))
       (fsub (reg 1) (reg 3) (reg 1)) % reg 1 and reg 2 are float now
       (fmul (reg 1) (reg 2) (reg 1))  % *
       (fsub (reg 1) (reg 3) (reg 1))
       (sub  (reg 1) (reg 3) (reg 1))
      ) )
 
          (longquotient               % Integer quotient
      ((sconst06  (reg 3))            % with 46-bit Integers
       (sconst04  (reg 5))
       (sub  (reg 2) (reg 3) (reg 2)) % using FloatingPoint Unit
       (sub  (reg 1) (reg 3) (reg 1))
       (fsub (reg 2) (reg 3) (reg 2))
       (fsub (reg 1) (reg 3) (reg 1)) % reg 1 and reg 2 are float now
       (hrecip (reg 4) (reg 2))
       (movsbit (reg 5) (reg 1))
       (fadd (reg 1) (reg 1) (reg 5))
       (imul (reg 2) (reg 2) (reg 4))
       (fmul (reg 2) (reg 2) (reg 4))
       (rmul (reg 1) (reg 2) (reg 1))
       (fsub (reg 1) (reg 3) (reg 1))
       (sub  (reg 1) (reg 3) (reg 1))
      )) 
 
          (longdivide                 % Integer divide
      ((sconst06  (reg 3))            % with 46-bit Integers
       (sconst04  (reg 5))
       (Move (reg S7) (Reg 1))
       (sub  (reg 2) (reg 3) (reg 2)) % using FloatingPoint Unit
       (sub  (reg 1) (reg 3) (reg 1)) %results: quotient and remainder
       (fsub (reg 2) (reg 3) (reg 2)) % put into T1 and T2
       (fsub (reg 1) (reg 3) (reg 1)) %
       (hrecip (reg 4) (reg 2))
       (movsbit (reg 5) (reg 1))
       (fadd (reg 1) (reg 1) (reg 5))
       (imul (reg S6) (reg 2) (reg 4))
       (fmul (reg S6) (reg S6) (reg 4))
       (rmul (reg 1) (reg S6) (reg 1))
       (fsub (reg 1) (reg 3) (reg 1))
       (sub  (reg 1) (reg 3) (reg 1))
       (move (reg T1) (Reg 1))         % quotient
       (sub  (reg 5) (reg 3) (reg 1))  % refloat fixed quotient
       (fsub (reg 5) (reg 3) (reg 5))
       (fmul (Reg 5) (Reg 5) (Reg 2))
       (fsub (reg 5) (reg 3) (reg 5))
       (sub  (reg 5) (reg 3) (reg 5))
       (sub  (reg 5) (reg S7) (reg 5))
       (move (reg T2) (Reg 5))         % remainder
      )) 
 
          (longdivide-quotient
      ((move (reg 1) (reg T1))))
 
          (longdivide-remainder
      ((move (reg 1) (reg T2))))
 
        (PUTBYTE
      ((*MOVE        (REG 1) (REG A6))
       (*MOVE        (REG 2) (REG 4))
       (*WSHIFT      (REG 4) (WCONST -3))
       (*MOVE        (Reg 4) (Reg A7))
       (*WPLUS2      (REG A6) (REG A7))
       (*Move        (Wconst 7) (REG S7))
       (LOAD         (REG 1) 0 (REG A6))
       (*WAND        (REG 2) (Reg S7))
       (*MOVE        (WCONST 7) (REG 4))
       (*WDIFFERENCE (REG 4) (REG 2))
       (*MOVE        (WCONST 255) (REG 2))
       (*Wand        (Reg 3) (REG 2))
       (*WSHIFT      (REG 4) (WCONST 3))
       (*Move        (Reg 4) (Reg A5))
       (LSHIFT       (REG 2) (REG 2) (REG A5))
       (LSHIFT       (REG 3) (REG 3) (REG A5))
       (*WNOT        (REG 2) (REG 2))
       (*WAND        (REG 1) (REG 2))
       (*WOR         (REG 1) (REG 3))
       (STORE        (REG 1) 0 (REG A6))
       (*MOVE        (QUOTE NIL) (REG 1))
      )) 
 
             (BYTE
      ((*MOVE        (REG 1) (REG A6))
       (*MOVE        (REG 2) (REG 4))
       (*WSHIFT      (REG 4) (WCONST -3))
       (*Move        (Reg 4) (Reg A7))
       (*WPLUS2      (REG A6) (REG A7))       % ADDRESS
       (*Move        (WCONST 7) (Reg S7))
       (LOAD         (REG 1) 0 (REG A6))
       (*WAND        (REG 2) (Reg S7))
       (*MOVE        (WCONST 7) (REG 4))
       (*WDIFFERENCE (REG 4) (REG 2))
       (*WSHIFT      (REG 4) (WCONST 3))   % (7 - I) * 8 TO THE RIGHT
       (*Move        (Reg 4) (Reg A5))
       (RSHIFT       (REG 1) (REG 1) (REG A5))
       (*WAND        (REG 1) (WCONST 255))
       ))
 
        (PUTSTRBYT
      ((*MOVE        (REG 1) (REG A6))
       (*MOVE        (REG 2) (REG 4))
       (*WSHIFT      (REG 4) (WCONST -3))
       (*MOVE        (Reg 4) (Reg A7))
       (*WPLUS2      (REG A6) (REG A7))
       (*Move        (Wconst 7) (REG S7))
       (LOAD         (REG 1) 1 (REG A6)) %instead of strbase
       (*WAND        (REG 2) (Reg S7))
       (*MOVE        (WCONST 7) (REG 4))
       (*WDIFFERENCE (REG 4) (REG 2))
       (*MOVE        (WCONST 255) (REG 2))
       (*Wand        (Reg 3) (REG 2))
       (*WSHIFT      (REG 4) (WCONST 3))
       (*Move        (Reg 4) (Reg A5))
       (LSHIFT       (REG 2) (REG 2) (REG A5))
       (LSHIFT       (REG 3) (REG 3) (REG A5))
       (*WNOT        (REG 2) (REG 2))
       (*WAND        (REG 1) (REG 2))
       (*WOR         (REG 1) (REG 3))
       (STORE        (REG 1) 1 (REG A6))
       (*MOVE        (QUOTE NIL) (REG 1))
      )) 
 
             (STRBYT
      ((*MOVE        (REG 1) (REG A6))
       (*MOVE        (REG 2) (REG 4))
       (*WSHIFT      (REG 4) (WCONST -3))
       (*Move        (Reg 4) (Reg A7))
       (*WPLUS2      (REG A6) (REG A7))       % ADDRESS
       (*Move        (WCONST 7) (Reg S7))
       (LOAD         (REG 1) 1 (REG A6))  % this replaces strbase
       (*WAND        (REG 2) (Reg S7))
       (*MOVE        (WCONST 7) (REG 4))
       (*WDIFFERENCE (REG 4) (REG 2))
       (*WSHIFT      (REG 4) (WCONST 3))   % (7 - I) * 8 TO THE RIGHT
       (*Move        (Reg 4) (Reg A5))
       (RSHIFT       (REG 1) (REG 1) (REG A5))
       (*WAND        (REG 1) (WCONST 255))
       ))
 
               (PUTHALFWORD      % CRAY-VERSION : HALFWORDS = PARCELS
      ((*MOVE        (REG 1) (REG A6))
       (*MOVE        (REG 2) (REG 4))
       (*WSHIFT      (REG 4) (WCONST -2))     % DIVIDE BY 4
       (*Move        (Reg 4) (Reg A7))
       (*WPLUS2      (REG A6) (REG A7))
       (*Move        (WCONST 3) (Reg S7))     % MODULO 4
       (LOAD         (REG 1) 0 (REG A6))
       (*WAND        (REG 2) (Reg S7))     % MODULO 4
       (*MOVE        (WCONST 3) (REG 4))
       (*WDIFFERENCE (REG 4) (REG 2))          % (3 - I MOD 4) * 16 ^
       (*MOVE        (WCONST 65535) (REG 2))   % MASK OF 16 BITS
       (*WSHIFT      (REG 4) (WCONST 4))
       (*Move        (Reg 4) (Reg A5))
       (LSHIFT       (REG 2) (REG 2) (REG A5))
       (LSHIFT       (REG 3) (REG 3) (REG A5))
       (*WAND        (REG 3) (REG 2))
       (*WNOT        (REG 2) (REG 2))
       (*WAND        (REG 1) (REG 2))
       (*WOR         (REG 1) (REG 3))
       (STORE        (REG 1) 0 (REG A6))
       (*MOVE        (QUOTE NIL) (REG 1))
       ))
 
               (HALFWORD   % CRAY-HALFWORDS =  PARCELS
      ((*MOVE        (REG 1) (REG A6))
       (*MOVE        (REG 2) (REG 4))
       (*WSHIFT      (REG 4) (WCONST -2))
       (*Move        (Reg 4) (Reg A7))
       (*WPLUS2      (REG A6) (REG A7))       % ADDRESS
       (*Move        (WCONST 3) (Reg S7))
       (LOAD         (REG 1) 0 (REG A6))
       (*WAND        (REG 2) (Reg S7))
       (*MOVE        (WCONST 3) (REG 4))
       (*WDIFFERENCE (REG 4) (REG 2))
       (*WSHIFT      (REG 4) (WCONST 4))   % (3 - I) * 16 TO THE RIGHT
       (*Move        (Reg 4) (Reg A5))
       (RSHIFT       (REG 1) (REG 1) (REG A5))
       (*WAND        (REG 1) (WCONST 65535))
       (*MOVE        (WCONST 15) (REG A6))         % SIGNEXTEND^
       (DRSHIFT      (REG 2) (REG 1) (REG A6))
       (RSHIFT       (REG 1) (REG 1) (WCONST 15))
       (MOVENEG      (REG 1) (REG 1))
       (DLSHIFT      (REG 1) (REG 2) (REG A6))
      )) 
    (clear-extraregs
     ((*VsetVl  10)
      (*Move 0 (Reg S6))
      (*Move (Reg Argumentblock) (Reg A0))
      (*Vset 1 (Reg S6))
      (*VStore 1 (Reg A0) 1)
    ))
    ) 'opencode
)
(put 'idapplye 'exitopencode
                             '((Move (Reg s7) (Reg 2))
                               (Move (Reg a6) (Reg symfnc))
                               (Move (Reg a7) (Reg s7))
                               (Add (Reg a6) (Reg a6) (Reg a7))
                               (load (reg 5) (indexed (reg a6) 0))
                               (lshift (Reg s7) (Reg s7) 4)
                               (*move 1 (reg s6))
                               (or (Reg s6) (Reg s6) (Reg s7))
                               (Move (Reg nargReg) (Reg s6))
                               (*wshift (reg 5) -32)
                               (*Move (reg 5) (Reg b07))
                               (j (Reg b07))))
 
(put 'idapplye2 'exitopencode
                             '((Move (Reg s7) (Reg 3))
                               (Move (Reg a6) (Reg symfnc))
                               (Move (Reg a7) (Reg s7))
                               (Add (Reg a6) (Reg a6) (Reg a7))
                               (load (reg 5) (indexed (reg a6) 0))
                               (*move 2 (reg S6))
                               (lshift (Reg s7) (Reg s7) 4)
                               (or (Reg s6) (Reg s6) (Reg s7))
                               (Move (Reg nargReg) (Reg s6))
                               (*wshift (reg 5) -32)
                               (*Move (reg 5) (Reg b07))
                               (j (Reg b07))))
 
(put '!%uncatch 'opencode
      '((*move (reg 1) (reg catchstackptr))
        (*move (reg T00)(Reg 1))
        (*move (Reg 1) ($fluid throwsignal*))))
 
(put 'catchsetup 'opencode
    '(
      (*ploc (reg a1) (labelgen ganz-hinten))
      (*move (reg a1) (reg 2))
      (*Move (reg st) (reg 3))
      (*Move (reg A3) (reg 4))
      (*wplus2 (reg 3) (reg 4))
      (*move (reg st) (reg a0))
      (blksto (reg a3) (reg prevframesize))
      (*move (reg catchstackptr) (Reg 5))
      (*move (reg catchstackptr) (reg a6))
      (*wplus2 (reg a6) (wconst 5))
      (*move (reg a6) (reg catchstackptr))
      (*move (reg 1) (indexed (reg a6) 0)) % tag
      (*move (reg 2) (indexed (reg a6) 1)) % pc
      (*move (reg 3) (indexed (reg a6) 2)) % sp
      (*move (reg bndstkptr) (reg a7))
      (*move (reg a7) (indexed (reg a6) 3)) % env
      (*move (reg 4) (indexed (reg a6) 4)) % framesize
      (*move catchstacksize (reg a7))
      (*move 5 (reg a1))
      (*wtimes2 (reg a7) (Reg a1))
      (*move ($fluid catchstack) (reg a1))
      (*wplus2 (reg a1) (reg a7))
      (*jumpwlessp (labelgen templabel) (reg a6) (reg a1))
      (*move (quote "Catch-throw stack overflow") (reg 1))
      (*Jcall  StdError)
      (*lbl (label (labelref templabel)))
      (*move (reg T00) ($fluid throwsignal*))
      (*move (Reg 5) (reg 1))
      (lapopt-fence (reg 1))
      (*lbl (label (labelref ganz-hinten)))
))
 
(if_system CRAY1 NIL
 (progn
   (put 'enable-operand-range-interrupt 'opencode  '((ERI)))
   (put 'disable-operand-range-interrupt 'opencode '((DRI)))
   (put 'enable-operand-range-interrupt  'destroys '(NIL))
   (put 'disable-operand-range-interrupt 'destroys '(NIL))
))
 
(put 'enable-floatingpoint-interrupt 'opencode  '((EFI)))
(put 'disable-floatingpoint-interrupt 'opencode '((DFI)))
(put 'enable-floatingpoint-interrupt  'destroys '(NIL))
(put 'disable-floatingpoint-interrupt 'destroys '(NIL))
 
(if_system CRAY1 NIL
 (progn
   (put 'eqcar-intern 'opencode
         '((*move (reg 1) (reg a5))
           (DRI)
           (*move (indexed (reg a5) 0) (reg 3))
           (*wshift (Reg 1) -59)
           (*move pair-tag (Reg S7))
           (Xor (Reg 4) (reg 1) (reg S7))
           (Xor (reg 5) (reg 3) (reg 2))
           (or  (reg 1) (reg 5) (reg 4))
           (ERI)))
 
   (put 'length 'opencode
         '((DRI)
           (*move 0 (reg 3))
           (*move (reg 1) (reg s6))
           (*move 1 (reg 4))
           (*move pair-tag (reg s7))
           (*lbl (label (labelgen templabel)))
           (*move (reg s6) (reg 1))
           (*move (reg s6) (reg a6))
           (*move (indexed (reg a6) 1) (reg s6))
           (*wrshift (reg 1) 59)
           (*move (reg 3) (reg 2))   %update count
           (*wxor3 (reg s0) (reg 1) (reg s7))
           (*wplus2 (reg 3) (reg 4))  % increment
           (*jumpeq (labelref templabel) (reg s0) 0)
           (*move (reg 2) (reg 1))  % value
           (*move (reg s6) (reg s7))
           (ERI)
)))   )
 
(if_system CRAY1
 (progn
   (put 'eqcar-intern 'opencode
         '((*move (reg 1) (reg a5))
           (*move (reg 1) (reg 4))
           (*wrshift (Reg 4) 59)
           (*move pair-tag (Reg S7))
           (Move (reg 1) 1)   % No
           (*jumpnoteq (labelgen templabel) (reg 4) (reg s7))
           (*move (indexed (reg a5) 0) (reg 3))
           (Xor (reg 1) (reg 3) (reg 2))
           (*lbl (label (labelref templabel)))))
))
