%
% 370LDEP.RED - IBM 370 PSL LAP depositing routines
%
% Author:      James McDonald
%              IMSSS
%              Stanford University
% Date:        12 February 1984
% Copyright (c) 1984 Stanford University
%

(compiletime (load fasl-decls))

(fluid '(*listlap *depositlap))
        

(fluid '(codebase* 
         currentoffset*     % additional offset
         targetcharacterset* 
         lapevenbyte* 
         lapbytepositioninword* 
         addressesknowntolap*))

(setq targetcharacterset* 'ascii)       % even though IBM is generally EBCDIC

(deflist '((byte lapdepositbytelist) 
           (halfword lapdeposithalfwordlist) 
           (fullword lapdepositfullwordlist) 
           (float lapdepositfloat) 
           (string lapdepositstring) 
           (ebcdic lapdepositebcdic) 
           (ascii lapdepositascii) 
           (samelinecomment lapignorecomment)) 
          'lapdepositfunction)

(on syslisp)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                           bytes                                      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapdepositbytelist (x) 
 (prog (l) 
  (setq l (cdr x)) 
  (while (and l (cdr l)) 
  (progn                                %  Deposit numbers exactly as given.
      %  If ASCII/EBCDIC conversion is needed, write a different routine.
      (lapdepositbytebyte (car l) (cadr l)) 
      (setq l (cddr l)))) 
  (cond (l (lapdepositbytebyte (car l) 0)))))
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                        strings                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapdepositstring (s) 
 (prog nil 
  (cond ((weq targetcharacterset* 'ascii) (lapdepositascii s))
%	((weq targetcharacterset* 'ebcdic) (lapdepositebcdic s))
	(t (errorprintf "LapDepositString: Unknown target character set: %w" 
                    targetcharacterset*)))))
 


(de lapdepositebcdic (x) 
 (prog (s n compilercharacterset 
              lapbytepositioninword* lapevenbyte*) 
  (setq s (cadr x)) 
  (setq n (strbyt "A" 0)) 
  (cond ((weq n 65) (setq compilercharacterset 'ascii))
        ((weq n 193) (setq compilercharacterset 'ebcdic))
        (t (errorprintf "LapDepositEBCDIC: Using unknown character set. A = %w" 
                    n))) 

  (setq lapbytepositioninword* 0) 
  (setq n (size s))
  (for (from i 0 n 1) (do 
  	(progn (cond ((weq compilercharacterset 'ascii) 
           			(lapdepositbyte (asciitoebcdic (strbyt s i))))
		     (t (lapdepositbyte (strbyt s i))))))) 

  (while (wneq lapbytepositioninword* 0) (lapdepositbyte 0))))
 

(de lapdepositascii (x) 
 (prog (s n compilercharacterset lapbytepositioninword* lapevenbyte*) 
  (setq s (cadr x)) 
  (setq compilercharacterset 'ascii) 

  (setq lapbytepositioninword* 0) 
  (setq n (wplus2 (size s) 1)) 
  (for (from i 0 n 1) (do (lapdepositbyte (strbyt (inf s) i))))

  (while (wneq lapbytepositioninword* 0) (lapdepositbyte 0))))
 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                      Deposit bytes                                    %
%-----------------------------------------------------------------------%
%                    LapDepositByte( B )                                %
%                                                                       %
%   Bytes for even addresses are remembered and then later deposited    %
%   as part of a pair with the following byte.                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapdepositbyte (b) 
  (cond ((zerop (wremainder lapbytepositioninword* 2)) (setq lapevenbyte* b))
        (t (lapdeposithalfword (wor (wshift lapevenbyte* 8) b)))) 
  (setq lapbytepositioninword* (wremainder (wplus2 lapbytepositioninword* 1) 4)))
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%           Deposit halfwords with various components                   %
%-----------------------------------------------------------------------%
%  LapDepositHalfWordList    ( HALFWORD . HalfWords )                   %
%  LapDepositByteNibbleNibble( HalfWordOffset, Byte, Nibble, Nibble )   %
%  LapDepositByteByte        ( HalfWordOffset, Byte, Byte           )   %
%  LapDepositHalfWord        ( HalfWordOffset, HalfWord             )   %
%                                                                       %
%  HalfWordOffset is to be added to offset for current instruction.     %
%  It will be converted to a byte offset.                               %
%                                                                       %
%    (Would DepositByte be better for these?)                           %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapdeposithalfwordlist (x) 
  (foreach hw in (cdr x) do 
  (progn (cond ((pairp hw) 
         (cond ((eqcar hw 'idoffset) (setq hw (idoffsetof (cadr hw))))))) 
      (lapdeposithalfword hw))))

(de lapdepositbytenibblenibble (byte n1 n2) 
 (lapdeposithalfword (wor (wshift byte 8) (wor (wshift n1 4) n2))))

(de lapdepositbytebyte (b1 b2) 
 (lapdeposithalfword (wor (wshift b1 8) b2)))

(de lapdeposithalfword (hw) 
  (cond ((not (fixp hw)) 
  (progn (cond ((eqcar hw 'hex) (setq hw (cadr hw)))
	       (t (laperrorprintf "LapDepositHalfWord: using 0 for %p" hw) 
                  (setq hw 0))))))
  (cond (*listlap (listlaphalfword hw))) 
  (cond (*depositlap (puthalfword codebase* currentoffset* hw))) 
  (setq currentoffset* (wplus2 currentoffset* 1)))
 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%            Deposit fullwords at current location                      %
%-----------------------------------------------------------------------%
%          LapDepositFullWordList     ( FULLWORD . Words )              %
%          LapDepositAlignedFullWord  ( Word )                          %
%          LapDepositFullWord         ( number )                        %
%                                                                       %
%  The word will be deposited at the location Lap is currently filling. %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapdepositfullwordlist (l) 
 (prog nil 
  (cond ((not (zerop (wremainder codebase* 4))) 
     (laperrorprintf "Unaligned Fullwords at %w : %p" codebase* l))) 
  (foreach word in (cdr l) do (lapdepositalignedfullword word))))
 


%
%  Possible forms here are:
%
%   <integer>
%   <other number> ???
%   (MkItem <tag> <inf>)        <inf> may be local label or ???
%   (ExternalRefernce <label>)
%   (InternalEntry <label>)
%   ??
%

(de lapdepositalignedfullword (x) 
 (prog (word tag inf address) 
  %  Note that the number to be deposited will sometimes exceed 24 bits.
  %  Addresses and Offsets will be leq 24 bits, though.
% (setq word (resolvewconstexpression x)) 	% expands macros, e.g. IDLOC
  (setq word x)

  (cond ((fixp word) (lapdepositfullword (int2sys word))) 
        ((atom word) (stderror (bldmsg 
           "LapDepositAlignedFullWord: can't handle reference to %p" x))) 
        ((eqcar word 'mkitem) 
		(progn (setq tag (cadr word)) 
			(setq inf (caddr word)) 
			(cond ((not (numberp tag)) 
	                (stderror (bldmsg
			 "LapDepositAlignedFullWord: unknown tag in %p" x)))) 
       % Prepare to relocate absolute ptrs
             (cond ((stringp inf)
			(setq inf (cdr (assoc inf label-alist*)))
			(unless *writingfaslfile
				(setq inf (wplus2 codebase* inf))))
		   ((eqcar inf 'idloc) 
				(if (not *writingfaslfile) 
				   (setq inf (id2int (cadr inf)))
				   (setq inf (findidnumber (cadr inf)))))
		   ((not (numberp inf))(setq inf (lapeffectiveaddressof inf))))
             (if (not *writingfaslfile)
             	(lapdepositfullword (wor (wshift tag 27) (inf inf)))
	 	(updatebittable 4 reloc-inf)
		(lapdepositfullword 
			(mkitem tag (makerelocinf 
 				(if (eq tag id-tag) reloc-id-number
						reloc-code-offset)
						inf))))))
  	((eqcar word 'entry) 
		(if (not *writingfaslfile)
                   (lapdepositfullword
		    (wplus2 SYMFNC (idoffsetof (cadr word))))
			(updatebittable 4 reloc-word)
			(lapdepositfullword
			    (makerelocword
				reloc-function-cell (findidnumber (cadr word)))
	 )))
  	((eqcar word 'internalentry)
			(updatebittable 4 reloc-word)
			(if (get (cadr word) 'internalentryoffset)
				(lapdepositfullword 
				  (makerelocword reloc-code-offset
					(get (cadr word) 'internalentryoffset)))
				(push (list 'call (cadr word) (wtimes2 2 currentoffset*))
					forwardinternalreferences*)
				(lapdepositfullword 0))) 
		%(lapeffectiveaddressof (findentrypoint (cadr word))))) 
	((eqcar word 'valuecell)
		(setq word (cadr word))
		(if (not *writingfaslfile)
                        (lapdepositfullword
			   (plus2 symval (idoffsetof (cadr word))))
                        (updatebittable 4 reloc-word)
                        (lapdepositfullword
                          (makerelocword reloc-value-cell
                                (findidnumber (cadr word))))))
  	((eqcar word 'hex) (lapdepositfullword (int2sys (cadr word)))) 
  	((or (eqcar word 'idloc) (eqcar word 'idoffset) )
		(if (not *writingfaslfile)
			(lapdepositfullword (id2int (cadr word)))
			(updatebittable 4 reloc-word)
			(lapdepositfullword
			  (makerelocword reloc-id-number 
				(findidnumber (cadr word))))))
	((eqcar word 'idoffsetc)
		(if (not *writingfaslfile)
                        (lapdepositfullword (id2int (cadr (cadr word))))
                        (updatebittable 4 reloc-word)
                        (lapdepositfullword
                          (makerelocword reloc-id-number
                                (findidnumber (cadr (cadr word)))))))
	(t (stderror (bldmsg "LapDepositAlignedFullWord: %p" x))))))
 

(de lapdepositfullword (word) 
 (prog nil 
  (cond (*listlap 
  (progn (listlaphalfword (wshift word -16)) 
      (listlaphalfword (wand word 65535))))) 
  (cond (*depositlap 
     (setf (getmem (wplus2 codebase* (wtimes2 2 currentoffset*))) word))) 
  (setq currentoffset* (wplus2 currentoffset* 2))))
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%      floating point numbers                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapdepositfloat (f) 
 (prog (ptr) 
  (setq ptr (fltinf (cadr f))) 
  (lapdepositfullword (getmem (wplus2 ptr addressingunitsperitem))) 
  (lapdepositfullword (getmem (wplus2 ptr (wtimes2 2 addressingunitsperitem))))))
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%      auxiliary routines for address calculations                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapeffectiveaddressof (reference) 
 (prog (address) 
  (setq address (findlapeffectiveaddressof reference)) 
  (when (null address) 
     (stderror (bldmsg "Lap cannot calculate address of %p" reference)))
  (return address)))
 

(de findlapeffectiveaddressof (reference) 
 (prog (expansion pr) 
  (cond ((atom reference) 
	(cond ((codep reference) (return (inf reference)))) 
	(setq expansion (expandatomicoperand reference)) 
	(cond ((atom expansion) (return nil))) 
	(setq reference expansion))) 
  (cond ((eqcar reference 'indexed) 
        (setq pr (assoc (cadr reference) addressesknowntolap*)) 
        (cond (pr 
	      (return (wplus2 (cdr pr) (operanddisplacementof reference)))))))
  % what about references to SymPrp and SymNam ?
  (return nil)))
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                        misc                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapignorecomment (x) nil)

(de lapallocatestorage (x) 
 (cond ((equal x "0F") (if (eq 1 (wand currentoffset* 1))
			 (setq currentoffset* (add1 currentoffset*))))
       ((member x '("0H" "0F" "0D")) nil)
       (t (errorprintf "***** LapAllocateStorage can't handle %p%n" x))))

(de lapproblem (x) 
 (errorprintf "***** Lap saw PROBLEM opcode: %p" x))

(off syslisp)
% of file
