%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for Sparc
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

(compiletime (flag '(signal-done build-trap-message) 'internalfunction))

(fluid '(*bruch* bruch_bruch *dump))

(de initializeinterrupts() 
  (setq bruch_bruch 0)
  (psl_sigset 2 (symfnc (id2int 'terminal-interrupt))))

(setq *bruch* 0)

(de terminal-interrupt()
 (setq bruch_bruch 0)
 % (console-print-string "Terminal interrupt")(console-newline)  
 (error 10002 "Terminal Interrupt"))

(de trap_handler (sigcode pc)
     (prog (mess)
(CONSOLE-PRINT-STRING "signal ")
(unixputn sigcode)(console-print-string " pc ")
(unixputn pc)(console-newline)
	(when (and (eq sigcode 2) (not (eq *bruch* 0)))
		(setq *bruch* -17)
		(return 0)) % a weaponed interrupt
    (setq *bruch* 0)
    (when (eq (wand 16#ffff sigcode) 16#fd)
          (return (stack-overflow pc)))
    (smalldump)
    (setq mess (atsoc (wand 16#ffff sigcode)
                        '((1 . "GuardPage")
                          (2 . "Alignment error")
                          (16#fd "Stack Overflow")
                          (16#1d "illegal instruction")
                          (16#95 "integer overflow")
                          (16#aa "instruction misalignment")
                          (16#94  "integer divide by zero")
                          (16#92  "floating point exception")
                          (5      "bad virtual address trap")
                           )))
    (setq mess (if mess (cdr mess)
                  (bldmsg " Unknown signal type %d"  sigcode)))
    (return (stderror (build-trap-message mess pc)))))

(de stack-overflow(n) 
    (console-print-string "++++ stack overflow at ")
    (unixputn n)
    (console-newline)
    (stderror "Stack Overflow"))

(de smalldump() nil)

(compiletime (flag '(x-code-address-to-symbol) 'internalfunction))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (setf extra-info (bldmsg "%w%w" " in "
                                 (x-code-address-to-symbol (inf trap-addr))))
      (bldmsg "%w%w" trap-type extra-info)))

(fluid '(code-address* closest-address* closest-symbol*))

(de x-code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0))
        (mapobl #'(lambda (symbol)
                 (when (fcodep symbol)
                       (let ((address (inf (getfcodepointer symbol))))
                            (when (and (ileq address code-address*)
                                       (igreaterp address closest-address*))
                                  (setq closest-address* address)
                                  (setq closest-symbol* symbol))))))
       closest-symbol*))

(de code-address-to-symbol (ad) (x-code-address-to-symbol ad))
