%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PNK:COMP-SUPPORT.SL
% Title:          Run-time support for assorted things
% Author:         Eric Benson
% Created:        20 September 1981
% Modified:       29-Aug-84 11:01:48 (Brian Beach)
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 14-Sep-90 (Herbert Melenk)
%   faster code for list1:5
% 01-Dec-83 14:23:53 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% NCONS, XCONS are defined in CONS-MKVECT.

(compiletime (ds nextitem(&x)
                 (prog2 (putmem a &x) (setq a (wplus2 a addressingunitsperitem)))))
 
(de list2 (u v) 
  % 2-argument EXPR for LIST 
   (let* ((a (gtheap 4))
          (b (mkitem pair-tag a)))
         (nextitem u)
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem)))
         (nextitem v)
         (nextitem nil)
         b))

(de list3 (u v w)  
   (let* ((a (gtheap 6))
          (b (mkitem pair-tag a)))
         (nextitem u) 
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem))) 
         (nextitem v) 
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem)))
         (nextitem w)
         (nextitem nil) 
         b)) 
 
(de list4 (u v w x)   
   (let* ((a (gtheap 8)) 
          (b (mkitem pair-tag a))) 
         (nextitem u)  
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem)))  
         (nextitem v)  
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem))) 
         (nextitem w) 
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem)))
         (nextitem x)
         (nextitem nil)
         b))   
  

(de list5 (u v w x y)    
   (let* ((a (gtheap 10))  
          (b (mkitem pair-tag a)))
         (nextitem u)
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem)))
         (nextitem v)
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem)))
         (nextitem w)
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem)))
         (nextitem x)
         (nextitem (mkitem pair-tag (wplus2 a addressingunitsperitem))) 
         (nextitem y)
         (nextitem nil)
         b))  
 


%
% Support for WARRAYS:
%

(on fast-integers)

(de clear-warray (address size)
  (for (from index 0 (- size 1))
       (do (wputv address index 0))
       ))

(de initialize-warray (address initial-values)
  (for (in value initial-values)
       (from index 0)
       (do (wputv address index value))
       ))

(de clear-wstring (address char-count word-count)
  (clear-warray address word-count)
  (wputv address 0 (- char-count 1))
  )

(de initialize-wstring (address initial-string word-count)
  (let ((inf (strinf initial-string)))
    (for (from i 0 (- word-count 1))
	 (do (wputv address i (wgetv inf i)))
	 )))

(off fast-integers)
