%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:CONS-MKVECT.SL
% Title:        Standard Lisp constructor functions
% Author:       Eric Benson
% Created:      20 August 1981
% Modified:     07-May-84 07:30:17
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:
% Runtime:
%
% (c) Copyright 1987, Konrad Zuse-Zentrum Berlin, for modifications
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 07-Nov-86 W. Neun (ZIB Berlin)
%  changed some presetting loops to vector code
% 07-May-84 07:27:46 (Brian Beach)
%  Added Mike's MAKE-MANUAL-EVECTOR
% 24-Feb-84 08:24:09 (Brian Beach)
%  Make into a loadable file.  Removed definition of CONS, since CONS is now
%  in the kernel.  DePROGged.
% 01-Dec-83 14:25:23 (Brian Beach)
%   Translated from Rlisp to Lisp.
% 19-Jul-83  Nancy Kendzierski
%   Fixed Allocate!-String to refer to Len instead of L in call to
%   NonPositiveIntegerError.
% 22 Jun 1983 1433-PDT, Cris Perdue
%   Added Allocate-String, uninitialized strings for speed, also
%   Allocate-Words.
% <PSL.KERNEL>CONS-MKVECT.RED.4, 28-Feb-83 11:41:46, Edit by PERDUE
%  Moved Make-Words, Make-Halfwords, etc. here from SEQUENCE.RED
%  Also moved STRING and VECTOR here from there.
% Edit by Cris Perdue, 23 Feb 1983 1045-PST
% Changed occurrences of HeapUpperbound to HeapTrapBound in optimized
% allocators to supported pre-GC traps.
%  <PSL.KERNEL>CONS-MKVECT.RED.2, 10-Jan-83 15:50:08, Edit by PERDUE
%  Added MkEVect
% Edit by GRISS: (?)
% Optimized CONS, XCONS and NCONS
%  <PSL.INTERP>CONS-MKVECT.RED.5,  9-Feb-82 06:25:51, Edit by GRISS
%  Added HardCons
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(on fast-integers)
 
(compiletime (load inum vector-instructions))
 
% CONS is defined in the kernel.  It will probably need to be changed if
% XCONS and NCONS need changing.
 
%(de xcons (b a)
%  (let ((ptr (gtheap (pairpack))))
%    (setf (wgetv ptr 0) a)
%    (setf (wgetv ptr 1) b)
%    (mkpair ptr)
%    ))
 
%(de ncons (a)
%  % Same as (CONS A NIL)
%  (let ((ptr (gtheap (pairpack))))
%    (setf (wgetv ptr 0) a)
%    (setf (wgetv ptr 1) nil)
%    (mkpair ptr)
%    ))
 
(de mkvect (upper-bound)
  % Allocate a vector, initialize all elements to NIL
  (if (intp upper-bound)
    (if (wlessp upper-bound -1)
      (nonpositiveintegererror upper-bound 'mkvect)
     (if (wlessp upper-bound 64)
      (let* ((v (gtvect upper-bound)))
         (VsetVL (aplus2 upper-bound 1))
         (Vset (Vreg 1) nil)
         (Vstore (Vreg 1) (aplus2 v 1) 1)
         (mkvec v))
 
      (let* ((v (gtvect upper-bound))
             (rem (wplus2 (wand upper-bound 63) 1))
             (v1  (wdifference (wplus2 v upper-bound) rem)))
        (VsetVL 64)
        (Vset (Vreg 1) nil)
 
        (for (from i (wplus2 v 1) v1 64)
                  (do (Vstore (Vreg 1) i 1))
        )
        (VsetVL rem)
        (Vstore (Vreg 1) (aplus2 v1 2) 1)
        (mkvec v)
     ))) 
    (nonintegererror upper-bound 'mkvect)
  ))
 
(de make-vector (upper-bound fill-value)
  % Allocate a vector, initialize all elements to FILL-VALUE.
  (if (intp upper-bound)
    (if (wlessp upper-bound -1)
      (nonpositiveintegererror upper-bound 'make-vector)
     (if (wlessp upper-bound 64)
      (let* ((v (gtvect upper-bound)))
         (VsetVL (aplus2 upper-bound 1))
         (Vset (Vreg 1) fill-value)
         (Vstore (Vreg 1) (aplus2 v 1) 1)
         (mkvec v))
 
      (let* ((v (gtvect upper-bound))
             (rem (wplus2 (wand upper-bound 63) 1))
             (v1  (wdifference (wplus2 v upper-bound) rem)))
        (VsetVL 64)
        (Vset (Vreg 1) fill-value)
 
        (for (from i (wplus2 v 1) v1 64)
                  (do (Vstore (Vreg 1) i 1))
        )
        (VsetVL rem)
        (Vstore (Vreg 1) (aplus2 v1 2) 1)
        (mkvec v)
     ))) 
    (nonintegererror upper-bound 'make-vector)
    ))
 
(de mkevector (upper-bound first-element-value)
  % Allocate an evector, initialize the first element, and set the rest to NIL.
  (if (intp upper-bound)
    (if (wlessp upper-bound 0)
      (nonpositiveintegererror upper-bound 'mkevector)
     (if (wlessp upper-bound 64)
      (let* ((v (gtevect upper-bound)))
         (VsetVL (aplus2 upper-bound 1))
         (Vset (Vreg 1) nil)
         (VsetElement (VReg 1) 0 first-element-value)
         (Vstore (Vreg 1) (aplus2 v 1) 1)
         (mkevect v))
 
      (let* ((v (gtevect upper-bound))
             (rem (wplus2 (wand upper-bound 63) 1))
             (v1  (wdifference (wplus2 v upper-bound) rem)))
        (VsetVL 64)
        (Vset (Vreg 1) nil)
 
        (for (from i (wplus2 v 1) v1 64)
                  (do (Vstore (Vreg 1) i 1))
        )
        (VsetVL rem)
        (Vstore (Vreg 1) (aplus2 v1 2) 1)
        (setf (evecitm v 0) first-element-value)
        (mkevect v)
     ))) 
    (nonintegererror upper-bound 'mkevector)
    ))
 
(de make-manual-evector (upper-bound etag value)
 
  %  Allocate Evect, init all to 'value'.
  %  Fill in the 0th element with 'etag'.
 
  (if (not (intp upper-bound))
    % THEN We have an error.
    (nonintegererror upper-bound 'mkevect)
    % ELSE allocate.
    (if (< upper-bound 0)
      % THEN We have an error.
      (stderror '"An  Evect with fewer than one element cannot be allocated")
      % ELSE
     (if (wlessp upper-bound 64)
      (let* ((v (gtevect upper-bound)))
         (VsetVL (aplus2 upper-bound 1))
         (Vset (Vreg 1) value)
         (VsetElement (Vreg 1) 0 etag)
         (Vstore (Vreg 1) (aplus2 v 1) 1)
         (mkevect v))
 
      (let* ((v (gtevect upper-bound))
             (rem (wplus2 (wand upper-bound 63) 1))
             (v1  (wdifference (wplus2 v upper-bound) rem)))
        (VsetVL 64)
        (Vset (Vreg 1) value)
 
        (for (from i (wplus2 v 1) v1 64)
                  (do (Vstore (Vreg 1) i 1))
        )
        (VsetVL rem)
        (Vstore (Vreg 1) (aplus2 v1 2) 1)
        (setf (evecitm v 0) etag)
        (mkevect v)
 ))  ))) 
 
(de allocate-string (len)
  % Make str with Len chars, uninitialized
  (if (intp len)
    (if (wlessp len 0)
      (nonpositiveintegererror len 'allocate-string)
      (mkstr (gtstr (- len 1)))
      )
    (nonintegererror len 'allocate-string)
    ))
 
(de mkstring (upper-bound fill-char)
  % Make string of UPPER-BOUND+1 characters, with all characters set to FILL-CHAR.
  (if (intp upper-bound)
    (if (wlessp upper-bound -1)
      (nonpositiveintegererror upper-bound 'mkstring)
      (let ((ptr (gtstr upper-bound)))
      (for (from i 0 upper-bound 1)
                              (do (setf (strbyt ptr i) fill-char)))
      (mkstr ptr)
      )) 
    (nonintegererror upper-bound 'mkstring)
    ))
 
(de make-bytes (upper-bound fill-value)
  % Make a byte vector of UPPER-BOUND+1 elements, will all elements set to FILL-VALUE.
  (if (intp upper-bound)
    (if (wlessp upper-bound -1)
      (nonpositiveintegererror upper-bound 'make-bytes)
      (let ((ptr (gtstr upper-bound)))
      (for (from i 0 upper-bound)(do (setf (strbyt ptr i) fill-value)))
      (mkbytes ptr)
      )) 
    (nonintegererror upper-bound 'make-bytes)
    ))
 
(de make-halfwords (upper-bound fill-value)
  % Make a halfword vector of UPPER-BOUND+1 elements, will all elements set to FILL-VALUE.
  (if (intp upper-bound)
    (if (wlessp upper-bound -1)
      (nonpositiveintegererror upper-bound 'make-halfwords)
      (let ((ptr (gthalfwords upper-bound)))
      (for (from i 0 upper-bound) (do (setf (halfworditm ptr i) fill-value)))
      (mkhalfwords ptr)
      )) 
    (nonintegererror upper-bound 'make-halfwords)
    ))
 
(de make-words (upper-bound fill-value)
  % Make a word vector of UPPER-BOUND+1 elements, will all elements set to FILL-VALUE.
  (if (intp upper-bound)
    (if (wlessp upper-bound -1)
      (nonpositiveintegererror upper-bound 'make-words)
     (if (wlessp upper-bound 64)
      (let* ((v (gtwrds upper-bound)))
         (VsetVL (aplus2 upper-bound 1))
         (Vset (Vreg 1) fill-value)
         (Vstore (Vreg 1) (aplus2 v 1) 1)
         (mkwrds v))
 
      (let* ((v (gtvect upper-bound))
             (rem (wplus2 (wand upper-bound 63) 1))
             (v1  (wdifference (wplus2 v upper-bound) rem)))
        (VsetVL 64)
        (Vset (Vreg 1) fill-value)
 
        (for (from i (wplus2 v 1) v1 64)
                  (do (Vstore (Vreg 1) i 1))
        )
        (VsetVL rem)
        (Vstore (Vreg 1) (aplus2 v1 2) 1)
        (mkwrds v)
     ))) 
    (nonintegererror upper-bound 'make-words)
    ))
 
(de allocate-words (len)
  % Make a word vector of UPPER-BOUND+1 elements, unititialized.
  (if (intp len)
    (if (wlessp len 0)
      (nonpositiveintegererror len 'allocate-words)
      (mkwrds (gtwrds (- len 1)))
      )
    (nonintegererror len 'allocate-words)
    ))
 
% Maybe we want to support efficient compilation of these, as with LIST,
% by functions String2, String3, Vector2, Vector3, etc.
 
(dn string (u)
  %. Analogous to LIST, string constructor
  (list2string u))
 
(dn vector (u)
  %. Analogous to LIST, vector constructor
  (list2vector u))
 
(off fast-integers)
 
