%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:PUTD-GETD.SL
% Title:        Standard Lisp function defining functions
% Author:       Eric Benson
% Created:      18 August 1981
% Modified:     31-May-84 10:51:14 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 10 Oct 1986 , H. Melenk and W. Neun , ZIB Berlin
%               addition of fntype information into
%               the rightmost bits of symfnc vector (very cray
%               specific)
% 23-May-84 16:16:15 (Mike Creech)
%  Rewrote code-number-of-arguments to deal with functions that were
%  defined to have any number of arguments (-1 in *entry definition).
%  Reformatted PUTD to make more readable.
%  Rewrote remd to be more readable.
%  Reformatted getd to be more readable (and changed u => func).
% 01-Dec-83 14:59:44 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (load sys-consts))
 
(de getd (func)
 
    % Lookup function definition of func.
 
   (prog (fntype symbits)
    (return (and
                   (idp func)
                   (not (funboundp func))
                   (setq fntype (getmem (wplus2 symfnc (inf func))))
                   (setq symbits (wand fntype 15))
                   (setq fntype (wrshift (wlshift fntype 6) 40))
                   (cons
                         (cond ((eq symbits 0) 'expr)
                               ((eq symbits 1) 'fexpr)
                               ((eq symbits 2) 'nexpr)
                               ((eq symbits 3) 'macro))
                         (if
                          (eq fntype lambda-link-address)
                          (get func '*lambdalink)
                          (mkcode fntype)))))))
 
(de remd (func)
 
    % Remove function definition of func.
 
    (let 
        ((oldgetd (getd func)))
      (when
           oldgetd
                                      % There was an old definition.
           (makefunbound func)
           (remprop func 'type)
           (remprop func '*lambdalink))
      % Return the old definition.
        oldgetd)
    )
 
(fluid '(*redefmsg % controls printing of redefined
       *usermode % controls query for redefinition
       ))
 
(loadtime
  (progn (setq *usermode nil)  % start in system mode
         (setq *redefmsg t)    % message in PutD
       ))
 
(fluid '(*comp % controls automatic compilation
       promptstring*))
 
(de putd (fnname fntype fnexp)
 
  % Install function definition
  %
  % this differs from the SL Report in 2 ways:
  % - function names flagged LOSE are not defined.
  % -       "      "   which are already fluid or global are defined anyway,
  % with a warning.
  %
 
  (cond ((not (idp fnname)) (noniderror fnname 'putd))
        ((not (memq fntype '(expr fexpr macro nexpr)))
         (conterror 1305 "%r is not a legal function type" fntype
                    (putd fnname fntype fnexp)))
        ((flagp fnname 'lose)
         (errorprintf "*** %r has not been defined, because it is flagged LOSE"
                      fnname)
         nil)
        (t (prog (vartype
              printredefinedmessage
              oldin
              promptstring*
                  queryresponse)
                 (unless
                   (funboundp fnname)
                   (when
                      *redefmsg
                      (setq printredefinedmessage t))
                   (when
                      (and *usermode
                         (not (flagp fnname 'user)))
                      (if
                         (not (yesp (bldmsg
                            "Do you really want to redefine the system function %r?"
                                                  fnname)))
                         (return nil)
                         (flag1 fnname 'user))))
                 (cond
                  ((codep fnexp) (makefcode fnname fnexp
                                   (cond ((eq fntype 'expr) 0)
                                         ((eq fntype 'fexpr) 1)
                                         ((eq fntype 'nexpr) 2)
                                         ((eq fntype 'macro) 3)
                                  ) )
                                 (remprop fnname '*lambdalink))
                  ((and (idp fnexp)
                      (not (funboundp fnexp)))
                   (return (putd fnname fntype (cdr (getd fnexp)))))
                  (*comp
                        (return (compd fnname fntype fnexp)))
                  ((eqcar fnexp 'lambda)
                         (put fnname '*lambdalink fnexp)
                         (makeflambdalink fnname
                                   (cond ((eq fntype 'expr)  0)
                                         ((eq fntype 'fexpr) 1)
                                         ((eq fntype 'nexpr) 2)
                                         ((eq fntype 'macro) 3)
                         )         ) )
                  (t
                        (return
                               (conterror 1105
                                "Ill-formed function expression in PutD"
                                (putd fnname fntype fnexp)
                                )
                         )
                  )
                  )
 
                 (if
                (neq fntype 'expr)
                (put fnname 'type fntype)
                (remprop fnname 'type))
 
                 (if
                *usermode
                (flag1 fnname 'user)
                (remflag1 fnname 'user))
 
             (when
                  printredefinedmessage
                  (errorprintf "*** Function %r has been redefined"
                           fnname))
                 (return fnname)
             )
         )
      )
  )
 
(de code-number-of-arguments (code-pointer)
 
    % Return the number of arguments within the range 0-maxargs, or
    % "any" if function is setup to have any number of arguments, or
    % NIL if not a code-pointer or too many arguments.
 
    (when
         (codep code-pointer)
                                 % We have a real codepointer.
       (let
           ((num-args (!%code-number-of-arguments (codeinf code-pointer))))
           (cond
                ((weq num-args -1) "any")
              ((and (wgeq num-args 0)
                  (wleq num-args maxargs)) num-args)
              (t NIL)
              )
           )
       ) 
    )
 
 
 
