%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:LISP-MACROS.SL 
% Title:        Various macros to make pure Lisp more tolerable 
% Author:       Eric Benson 
% Created:      5 October 1981 
% Modified:     28-Aug-84 13:13:02 (Brian Beach) 
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 01-Dec-83 14:49:53 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
  (flag '(inthiscase) 'internalfunction))

% Not a macro, but it belongs with these
(df case (u)
  %U is of form (CASE <integer exp> (<case-1> <exp-1>) . . .(<case-n> <exp-n>)).

  % If <case-i> is NIL it is default,
  %   else is list of INT or (RANGE int int)
  (prog (caseexpr def caselst bod)
        (setq caseexpr (eval (car u)))
   l
        (unless (pairp (setq u (cdr u)))
          (return (eval def)))
        (setq caselst (caar u))
        (setq bod (cadar u))
        (when (or (not (pairp caselst))
                  (memq (car caselst) '(otherwise default)))
          (setq def bod)
          (go l))
        (when (inthiscase caseexpr caselst)
          (return (eval bod)))
        (go l)))

(de inthiscase (caseexpr cases)
  (cond ((not (pairp cases)) nil)
        ((and (pairp (car cases)) (eq (caar cases) 'range)
              (geq caseexpr (cadar cases)) (leq caseexpr (caddar cases)))
         t)
        ((equal caseexpr (car cases)) t)
        (t (inthiscase caseexpr (cdr cases)))))

(dm setf (u)
  %. General assignment macro
  (expandsetf (cadr u) (caddr u)))

(de expandsetf (lhs rhs)
  (let (lhsop)
       (cond ((atom lhs) (list 'setq lhs rhs))
             ((setq lhsop (get (car lhs) 'assign-op))
              (cons lhsop (append (cdr lhs) (list rhs))))
             % simple substitution case
             ((setq lhsop (get (car lhs) 'setf-expand))
              (apply lhsop (list lhs rhs)))
             % more complex transformation
             ((and (setq lhsop (getd (car lhs)))
                   (equal (car lhsop) 'macro))
              (expandsetf (apply (cdr lhsop) (list lhs)) rhs))
             (t (stderror
                 (bldmsg "%r is not a known form for assignment"
                         (list 'setf lhs rhs)))))))

(loadtime
  (deflist '((getv putv) (car rplaca) (cdr rplacd) (indx setindx)
             (sub setsub) (nth (lambda (l i x)
                                       (rplaca (pnth l i) x)
                                       x)) (eval set) (value set))
           'assign-op))

