%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for Mips
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 10-Mar-89 (Winfried Neun)
%  Handling for Mips as for Sparc (SUN4). Complete redesign.
%  Also installed fluid *bruch* to allow gc (etc) to turn off
%  cntrl-c interrupt detection, which destroy system. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (flag '(sigsetup signal-done build-trap-message) 
     			 'internalfunction))

(fluid '(*bruch* Mips-sigcp Mips-sigaddr Mips-sigcode Mips-signumber
         *dump Mips-sigpc Mips-dumparray*))

% this is the interface for toploop

(de initializeinterrupts ()
    (ieee_handler (strbase (strinf "set"))
                  (strbase (strinf "common"))
                  (symfnc (id2int 'signalhaendler))) 

          % all signals are initiated expect for those iin the followin list
          % they are ignored

    (for (from i 1 31) (do
       (when (not (memq i '(1 18 19 20 21 22 25 28))) (sigsetup i)))))

% this installs the handler

(de sigsetup (signumber)
    (sun3_sigset signumber (wtimes2 4 (symfnc (id2int 'signalhandler)))))
 
% this is the 'handler'

(setq Mips-dumparray* (gtwarray 127))
 
(de signal-done ()
     (prog (handel)
         (initializeinterrupts)
         (when (eq 2#10101111101 (wshift (wgetv mips-sigpc 0) -21))
                (console-print-string "***** Stack overflow *****")
                (console-newline)
                (wputv (ashift Mips-sigcp -2) 32
                       (wplus2 (wgetv (ashift Mips-sigcp -2) 32) 10000)))
         (setq handel (wshift (symfnc (id2int 'lisprpv)) 2))
         (wputv (ashift Mips-sigcp -2) 2 handel)
)    )            %        'Brutal' version like Cray X-MP for UNICOS
 

(lap '((*entry signalhandler expr 3)
%       We save the information which is in 'scratch' registers.
%
         (*move (indexed (reg 3) 272) (reg 5))
         (*move 8#777774 (reg 4))
         (*wnot (reg 4) (reg 4))
         (*wand (reg 5) (reg 4))  % clear fp status mask
         (*wor (reg 5) 8#3000)
         (*move (reg 5) (indexed (reg 3) 272))

         (*move 1 (reg 5))    % checking whether reg 19 is usable
         (*wshift (reg 5) 28) % if not, go on
         (*jumpwlessp (label goon1) (reg r19) (reg 5))
         (*move (reg r19) (reg 5))
         (*wand (reg 5) 3)
         (*jumpnoteq (label goon1)  (reg 5) 0)

         (*jumpnoteq (label goon)  ($fluid *kernelmode) 'nil)

         (*move (reg 1) (fluid Mips-signumber))
         (*move (reg 2) (fluid Mips-sigcode))
         (*move (reg 3) (fluid Mips-sigcp))   % save signal context 
         (*move (indexed (reg 3) 8) (reg 4))
	 (*wshift (reg 4) -2)
	 (*move (reg 4) (fluid Mips-sigpc))

% No alloc, trap handler will never come back!

         (*jumpnoteq (label templab) (reg 1) (quote 2)) % interrupt
         (*jumpeq    (label templab) ($fluid *bruch*) 0)
%
% Now we have a 'weaponed' interrupt. This is used e.g. by gc to remain
% operable. Return Immediately and mark with a negative value in *bruch* !!         
%
    goon
         (*move -17 ($fluid *bruch*))
         (*move 0 (fluid Mips-sigcp))
    goon1
         (jr (reg returnaddr))
         (nop)

    templab
         (*move ($fluid Mips-sigcp) (reg t1))
         (*move ($fluid Mips-dumparray*) (reg t2))
	 (*wshift (reg t2) 2)
	 (*move 77 (reg t4))
      lab
         (*move (indexed (reg t1) 0) (reg t5))
         (*move (reg t5) (indexed (reg t2) 0))
         (*wplus2 (reg t1) 4)
         (*wplus2 (reg t2) 4)
         (*wplus2 (reg t4) -1)
         (*jumpwgreaterp (label lab) (reg t4) 0)
         
         (*move  ($fluid mips-sigpc) (reg t4))
         (*wshift (reg t4) 2)
         (*move (indexed (reg t4) 0) (reg t4))
         (*move 2#10101111101 (reg t5))
         (*wshift (reg t4) -21)
         (*jumpeq (label nodump) (reg t4) (reg t5))

         (*move 50 (reg t4))
         (*move ($fluid Mips-dumparray*) (reg t1))
         (*wshift (reg t1) 2)
         (*move (indexed (reg t1) 128) (reg t1))
         (*wdifference (reg t1) 40)
     stack-lab                % copy a centerdump of stack
         (*move (indexed (reg t1) 0) (reg t5))
         (*move (reg t5) (indexed (reg t2) 0))
         (*wplus2 (reg t1) 4)
         (*wplus2 (reg t2) 4)
         (*wplus2 (reg t4) -1)
         (*jumpwgreaterp (label stack-lab) (reg t4) 0)
          
     nodump
         (*jcall signal-done)
))

(setq *bruch* 0)

(de lisprpv()
   (prog (mess)
    (setq *bruch* 0)
    (smalldump)
    (setq mess (atsoc Mips-signumber '((1 . "Hangup")
                          (2 . "Interrupt")
                          (3 . "Quit") 
                          (4 . "Illegal Instruction")
                          (5 . "Trace Trap") 
                          (6 . "IOT Instruction")
                          (7 . "EMT Instruction")
                          (8 . "Floating Point Exception")
                          (10 . "Bus Error")
			  (11 . "Segmentation Violation")
                          (12 . "Bad Args to System Call")
                          (13 . "Write on Pipe With None to Read")
                          (14 . "Alarm Clock")
                          (15 . "Software termination signal")  )))
    (when (eq Mips-signumber 8)
          (cond ((eq Mips-sigcode 16#c8)
                         (setq mess '( 0 . "Floating Point divide by zero")))
                ((eq Mips-sigcode 16#d4)
                         (setq mess '( 0 . "Floating Point overflow")))
                ((eq Mips-sigcode 16#cc)
                         (setq mess '( 0 . "Floating Point underflow")))))
    (setq mess (if mess (cdr mess)
                  (bldmsg " Unknown signal type %d"  Mips-signumber)))
    (build-trap-message mess Mips-sigpc)
    (setq Mips-sigaddr 0 Mips-sigcp 0) % for gc
    (return (stderror))))

(de smalldump()
  (when *dump 
    (console-print-string "*** UNEXPECTED INTERRUPT ***")
    (console-newline)
    (console-print-string "*** System signal number: 16#")
    (unixputn Mips-signumber)
    (console-print-string " , signal code: 16#")
    (unixputn Mips-sigcode)
    (console-newline)
    (console-print-string "*** signal pc: 16#")
    (unixputn Mips-sigpc)
    (console-print-string " , signal Address: 16#")
    (unixputn Mips-sigaddr)
    (console-print-string " , context: 16#")
    (unixputn Mips-sigcp)
    (console-newline)
    (setq Mips-sigaddr 0 Mips-sigcp 0) % for gc
    (console-print-string " RegMask ") (unixputn (wgetv Mips-dumparray* 0))
    (console-newline)
    (console-print-string " sigMask ") (unixputn (wgetv Mips-dumparray* 1)) 
    (console-newline)
    (console-print-string " sig_pc  ") (unixputn (wgetv Mips-dumparray* 2))
    (console-newline) (console-newline)

    (console-print-string " Reg 0-3 ") (unixputn (wgetv Mips-dumparray* 3))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 4))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 5))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 6)) 
    (console-newline)
    (console-print-string " Reg 4-7 ") (unixputn (wgetv Mips-dumparray* 7))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 8))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 9))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 10)) 
    (console-newline)
    (console-print-string "Reg 8-11 ") (unixputn (wgetv Mips-dumparray* 11))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 12))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 13))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 14)) 
    (console-newline)
    (console-print-string "Reg12-15 ") (unixputn (wgetv Mips-dumparray* 15))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 16))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 17))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 18)) 
    (console-newline)
    (console-print-string "Reg16-19 ") (unixputn (wgetv Mips-dumparray* 19))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 20))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 21))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 22)) 
    (console-newline)
    (console-print-string "Reg20-23 ") (unixputn (wgetv Mips-dumparray* 23))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 24))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 25))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 26))
    (console-newline)
    (console-print-string "Reg24-27 ")
    (unixputn (wgetv Mips-dumparray* 27))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 28))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 29))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 30))
    (console-newline)
    (console-print-string "Reg28-31 ") 
    (unixputn (wgetv Mips-dumparray* 31))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 32))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 33))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 34))
    (console-newline) (console-newline)
    (console-print-string " ownedfp ") (unixputn (wgetv Mips-dumparray* 35))
    (console-newline)
    (console-print-string "FReg 0-3 ") (unixputn (wgetv Mips-dumparray* 36))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 37))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 38))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 39))
    (console-newline)
    (console-print-string "FReg 4-7 ") (unixputn (wgetv Mips-dumparray* 40))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 41))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 42))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 43))
    (console-newline)
    (console-print-string "Freg 8-11 ") (unixputn (wgetv Mips-dumparray* 44))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 45))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 46))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 47))
    (console-newline)
    (console-print-string "FReg12-15 ") (unixputn (wgetv Mips-dumparray* 48))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 49))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 50))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 51))
    (console-newline)
    (console-print-string "FReg16-19 ") (unixputn (wgetv Mips-dumparray* 52))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 53))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 54))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 55))
    (console-newline)
    (console-print-string "FReg20-23 ") (unixputn (wgetv Mips-dumparray* 56))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 57))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 58))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 59)) 
    (console-newline)
    (console-print-string "FReg24-27 ") (unixputn (wgetv Mips-dumparray* 60))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 61))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 62))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 63)) 
    (console-newline)
    (console-print-string "FReg28-31 ") (unixputn (wgetv Mips-dumparray* 64))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 65))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 66))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 67))
    (console-newline) (console-newline)
    (console-print-string " fpc_csr ")
    (unixputn (wgetv Mips-dumparray* 68)) (console-newline)
    (console-print-string " fpc_eir ")
    (unixputn (wgetv Mips-dumparray* 69)) (console-newline)
    (console-print-string " md_hi   ") 
    (unixputn (wgetv Mips-dumparray* 70)) (console-newline)
    (console-print-string " md_lo   ") 
    (unixputn (wgetv Mips-dumparray* 71)) (console-newline)
    (console-print-string " cause   ")
    (unixputn (wgetv Mips-dumparray* 72)) (console-newline)
    (console-print-string " badvadd ")
    (unixputn (wgetv Mips-dumparray* 73)) (console-newline)
    (console-print-string " badpadd ")
    (unixputn (wgetv Mips-dumparray* 74)) (console-newline)
    (console-print-string " issigst ")
    (unixputn (wgetv Mips-dumparray* 75)) (console-newline)
    (console-print-string " trigger ")
    (unixputn (wgetv Mips-dumparray* 76)) (console-newline)
    (console-newline) 
    (console-print-string " ++++ Dump of Stack ++++") (console-newline)
    (console-newline) 
    (for (from i 77 126 5)
      (do (unixputn (wgetv Mips-dumparray* i)) (console-print-string " ")
          (unixputn (wgetv Mips-dumparray* (wplus2 1 i)))
	  (console-print-string " ")
          (unixputn (wgetv Mips-dumparray* (wplus2 2 i)))
	  (console-print-string " ")
          (unixputn (wgetv Mips-dumparray* (wplus2 3 i)))
	  (console-print-string " ")
          (unixputn (wgetv Mips-dumparray* (wplus2 4 i))) 
          (console-newline)))
))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (setf extra-info (bldmsg "%w%w" " in " 
                                 (code-address-to-symbol (inf trap-addr))))
      (bldmsg "%w%w" trap-type extra-info)))

(compiletime (fluid '(code-address* closest-address* closest-symbol*)))

(de code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0))
       (mapobl #'(lambda (symbol)
		         (when (fcodep symbol)
			       (let ((address (inf (getfcodepointer symbol))))
				    (when (and (ileq address
						     code-address*)
					       (igreaterp address
							  closest-address*))
					  (setq closest-address*
						address)
					  (setq closest-symbol* symbol))))))
       closest-symbol*))

