%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           pvm.sl
% Description:    Interface for pvm (3.1) connections
% Author:         Winfried Neun , ZIB Berlin
% Created:        13-August 1993
% Status:         Experimental (Do Not Distribute)
%
% Copyright (c) 1993, Konrad Zuse Zentrum Berlin, all rights reserved
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pvm constants
% return codes
 
(define-constant PvmOk               0)
(define-constant PvmBadparam        -2)
(define-constant PvmMismatch        -3)
(define-constant PvmNoData          -5)
(define-constant PvmNoHost          -6)
(define-constant PvmNoFile          -7)
(define-constant PvmNoMem           -10)
(define-constant PvmBadMsg          -12)
(define-constant PvmSysErr          -14)
(define-constant PvmNoBuf           -15)
(define-constant PvmNoSuchBuf       -16)
(define-constant PvmNullGroup       -17)
(define-constant PvmDupGroup        -18)
(define-constant PvmNoGroup         -19)
(define-constant PvmNotInGroup      -20)
(define-constant PvmNoInst          -21)
(define-constant PvmHostFail        -22)
(define-constant PvmNoParent        -23)
(define-constant PvmNotImpl         -24)
(define-constant PvmDSyserr         -25)
(define-constant PvmBadVersion      -26)
(define-constant PvmOutofRes        -27)
(define-constant PvmDupHost         -28)
(define-constant PvmCantStart       -29)
(define-constant PvmAlready         -30)
(define-constant PvmNoTask          -31)
(define-constant PvmNoEntry         -32)
(define-constant PvmDupEntry        -33)

%Option encoding

(define-constant PvmDataDefault     0)
(define-constant PvmDataRaw         1)
(define-constant PvmDataInPlace     2)

(define-constant PvmTaskDefault     0)
(define-constant PvmTaskHost        1)
(define-constant PvmTaskArch        2)
(define-constant PvmTaskDebug       4)

(define-constant PvmTaskExit        1)
(define-constant PvmHostDelete      2)
(define-constant PvmHostAdd         3)

% examples

(de master (nproc slave )

   (prog (mytid tids i who n msgtype data result hilf)

	(setq data (gtwarray 100))
	(setq result (gtwarray 32))
	(setq tids (gtwarray 32))
        (setq n 100 hilf (gtwarray 1))  % for &nproc etc

	(setq mytid (pvm_mytid))
	
        (pvm_spawn (strbase (mkvec slave))
		0 0 (strbase (mkvec "")) nproc tids)

	(for (from i 0 (isub1 nproc)) (do
		(print (list " tid " (wgetv tids i)))))

	(for (from i 0 100) (do (wputv data i i)))

	(pvm_initsend pvmdataraw)
	(wputv hilf 0 nproc) (pvm_pkint hilf 1 1)
	(pvm_pkint tids nproc 1)
	(wputv hilf 0 n) (pvm_pkint hilf 1 1)
	(pvm_pkint data n 1)
	(pvm_mcast tids nproc 0)
	
	(setq msgtype 5)

	(for (from i 0 (isub1 nproc) 1) (do
		(pvm_recv -1 msgtype)
		(pvm_upkint hilf 1 1)
		(pvm_upkint (wplus2 result (wtimes2 (wgetv hilf 0)
					addressingunitsperitem)) 1 1)
		(printf "I got %w from %w %n" (wgetv result (wgetv hilf 0))
			(wgetv hilf 0))
		))
	(pvm_exit)))

(de slave ()

   (prog (nproc me mytid tids i n master msgtype data result hilf)

        (setq data (gtwarray 100))
        (setq tids (gtwarray 32))
        (setq i 0 hilf (gtwarray 1))  % for &nproc etc

        (setq mytid (pvm_mytid))

	(setq msgtype 0)

        (pvm_recv -1 msgtype)
        (pvm_upkint hilf 1 1) (setq nproc (wgetv hilf 0))
        (pvm_upkint tids nproc 1)
        (pvm_upkint hilf 1 1) (setq n (wgetv hilf 0))
        (pvm_upkint data n 1)

 % determine which slave I am : 0 .. nproc -1

	(while (not me) 
		(if (eq mytid (wgetv tids i))
			(setq me i)
			(setq i (iadd1 i))))

	(setq result (for (from i 0 n ) (sum (wgetv data i))))
	(setq result (times2 result me))

 % result is  sum(data) mod me

 % send it back

        (pvm_initsend pvmdatadefault)
        (wputv hilf 0 me) (pvm_pkint hilf 1 1)
        (wputv hilf 0 result) (pvm_pkint hilf 1 1)
	(setq msgtype 5)
	
	(setq master (pvm_parent))
        (pvm_send master msgtype)

        (pvm_exit)))
