%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pascal_compat.red - Pascal compatability functions for Apollo PSL.
%
% Author:      J.W. Peterson 
%              Computer Science Dept.
%              University of Utah
% Date:        14 January 1983
%
% Modified 5-Apr-87 by Hucka to un-rlisp and convert for PSL 3.4.
% updated 30-jan-83 to add bitpad functions
% Based on GPR_INIT.RED by Fuh-Meei Chen
%
% This package provides identical facilities to some system functions that
% were originally implemented as Pascal ForiegnFunctions.  By using the
% SysCall package, they have all been coverted to work exclusivly in Lisp.
%
% This package contains additional facilities, that transform arguments into
% (or from) a form more pleasant to the normal LISP user.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% create a pad on top of standard transcript pad
% 
(de gpr_package_init nil 
  (prog (init_id fonts_id)  
	% graphics_channel is name of edited file, blank for tempory file
	(setq graphics_channel (pad_create "" 0 1 2 0 75))
                  % 0 = transcript style pad
                  % 1 = the window on which this pad is open, standoutput
                  % 2 = position this pad at the top of the current window
                  % 0 = Pane size is relative
                  % 75 = pane size is 75%
      % sets a scale factor for cursor operations
      (pad_set_scale graphics_channel 1 1) 	% stream id where pad is open
                     % 1 = the scale factor for the x-coordinate  
                     % 1 = the scale factor for the y-coordinate
      % creates a frame in a pad
      (pad_create_frame graphics_channel 800 1000) 	% stream id of open pad
                        % the width of the frame
                        % the height of the frame
   
                      % initialize the graphics primitives package
      (setq init_id 
         (gpr_init 1                    % frame mode
                   graphics_channel     % stream id of the pad
                   (list 800 1000) 0))  % size of the initial bitmap
                   % currently is 1
                       
         % establishes a bitmap as the current bit map for subsequent operation
      (gpr_set_bitmap init_id) 
         
         % loads a font file for the graphics to use

      (setq fonts_id (pad_load_font graphics_channel "/sys/dm/fonts/std")) 

      (setq fonts_id (gpr_load_font_file "/sys/dm/fonts/std" 17)) 

         % establishes a new font for subsequent text operation
      (gpr_set_text_font fonts_id) 
      (return (cons graphics_channel fonts_id))))


%------------------------------------------------------------------------------
% rlisp calling procedures
%------------------------------------------------------------------------------

(de l_erase nil                         % Erase the graphics frame
(gpr_clear 0))

(de l_move (x y)                        % Move to location X,Y on the screen
(gpr_move x y))

(de l_draw (x y)                        % Draw a line to location X,Y on the frame.
(gpr_line x y))

(de l_text (string)                     % Place the String at the current cursor pos.
(gpr_text string (plus (size string) 1)))

(de l_inq_position (streamid)           % Inquire the postion of the current
(prog (xposyposchar) 
  (setq xposyposchar (pad_locate streamid)) 	% cursor (i.e. touchpad).  NOTE: a PFK must be defined to
  (return (cons (cons (car xposyposchar) (cadr xposyposchar)) (caddr xposyposchar)))))

                    % produce a hex code F1...F4 for this function to return.


(de l_initplot nil                      % Initialized the graphics...MUST be called first.
(gpr_package_init))

(de l_inq_text_extent (string)          % Inquire the extent of the string.
(gpr_inq_text_extent string (plus (size string) 1)))

(de l_quit nil                          % Exit the graphics package...(EndPlot is better)
(stream_close graphics_channel))

(de l_endplot nil                       % added for Pascal compatablity
(progn (gpr_terminate 1)                % exit and delete display
%   GPR_UnLoad_Font_File(1);   % possibly unload std font (?need global fontid??)
    (stream_close graphics_channel) nil))	% Graphics_channel defined in initplot
%   MBX_close( ..whatever the bitpad uses..);  % not implemented yet


(de l_inq_cur nil                       % Get the last position MOVEd or DRAWn to.
(prog (pos x) 
  (setq pos (gpr_inq_cursor x)) 
  (return (car x))))


(de l_raw (panestreamid)                % Set pad to raw mode
(progn (pad_raw panestreamid) nil))

(de l_cooked (panestreamid)             % Set pad on StreamID to Cooked
(progn (pad_cooked panestreamid) nil))

(de l_set_draw_value (number)           % Set Draw value (AEGIS bug?)
(progn (gpr_set_draw_value number) nil))

(de l_set_text_value (number)           % Set color for text
(progn (gpr_set_text_value number) nil))

(de l_load_font_file (fontname)         % Load file specifying filename
 (gpr_load_font_file fontname (stringlength fontname)))

(de l_unload_font_file (fontindex)      % Unload font via FontID
(progn (gpr_unload_font_file fontindex) nil))

(de l_set_text_font (fontindex)         % Set Font to FontIndex (ID)
(progn (gpr_set_text_font fontindex) nil))

(de l_pad_create (filename type side panesize) 	% Create a pad
  (pad_create filename type 1           % Stdout
             side 0 panesize))

(de l_define_program_function_key (thestream whichkey definition) 
  (pad_def_pfk thestream whichkey definition))

% Invoke an inferiour process.
%
(de pgm_invoke (name                    % Filename of process
                             len        % Length of filename
                             argc       % # of arguments
                             argv       % pointers (array of) to said args.
                             connc      % # of streams being passed
                             connv      % array of said streams
                             mode)      % Currently PGM_$WAIT (?)
(prog (reserved                         % ...for future use
             status) 
  (putarg 1 name '(string)) 
  (putarg 2 len '(integer16)) 

  (putarg 3 argc '(integer16)) 
  (putobjpointer 4 argv)                % Note: argV structure of 16 bit count followed
                                % by string (pascal array of char type) must
                                % be set up by calling routine.  A combintion
                                % of the functions StringLength, PutHalfWord,
                                % GTwrds and xx[index] can be used.
  (putarg 5 connc '(integer16)) 
  (putobjpointer 6 connv)               % much the same as ArgV, but less tricky.
  (putarg 7 mode '(integer16))          % assume 1 for the current deflt value??

  (cleararg 8) 
  (cleararg 9) (syscall 

  'pgm_$invoke 9) 

  (setq status (getarg 9 '(integer32) '(fixn))) 

  (cond ((eq status 0) (return status)) (t (return (apolloerror status))))))


(de l_invoke_shell nil 
%
% Invokes an inferior shell underneath the running lisp system
%
(prog (shstring shell_arg shell_arg_vector streamv comstring) 
%
% first step is to build an 'ArgV' for the arguments to the invoked command.
% This consists of an ("UNIV") array of pointers to a group of arguments, where
% the arguments are a 16 bit length count followed by a pascal array of char
%
  (setq shell_arg (gtwrds 1))           % 32 bits is enough for 'sh'
  (setq shstring "SH") 
  (puthalfword shell_arg 0 (stringlength shstring)) 
  (puthalfword shell_arg 1 (plus (times (indx shstring 0) 256) (indx shstring 1))) 
%
% now that we've built an argument, put it's address into a vector 
% (length of one in this case).
%
  (setq shell_arg_vector (gtwrds 1)) 
  (putword shell_arg_vector 0 shell_arg) 
%
  (setq streamv (gtwrds 1))             % Dummy array for StreamV
%
  (setq comstring "/com/sh") 
  (pgm_invoke comstring (stringlength comstring) 1 	% argument count
             shell_arg_vector 0         % stream count
             streamv 1)))               % goofy mode stuff

% %-------------------------------
% %   Bitpad procedures
% %------------------------------- 
% %    
% % For these procedures to work, a Bitpad Server process MUST be running.
% % Invoking the command /utah/com/start_bitpad will fire up a process running
% % the bitpad server.
% %
% % Requires the following to be fluid for mbx access:
% %
% fluid '(BP_mbx_buff,        % Actual mailbox buffer
%         BP_handle);         % pointer to mailbox
% 
% %
% % This opens a line of communication to the bitpad server
% %
% symbolic procedure L_InitBitPad();
% <<
%   BP_mbx_buff:=gtbps(64);   % allocate space for mbx
%   BP_handle:=mbx_open("/utah/bitpadbox",15,BP_mbx_buff,0);  % start connection
% >>;
% 
% %
% % Returns a list of integers containing (Xpos Ypos KeyCode)
% %
% symbolic procedure L_GetPuckPosition();
%    begin              
%      putword(BP_mbx_buff,0,0);
%      putword(BP_mbx_buff,1,0);  % stuff message placeholder into buffer.
%      putword(BP_mbx_buff,2,0);
%      mbx_put_rec(BP_handle,BP_mbx_buff,64);
%      MBX_GET_REC(BP_handle,BP_mbx_buff,64);
%      return list(gethalfword(BP_mbx_buff,1),  % returns X coord...
%                  gethalfword(BP_mbx_buff,2),  % Y coord...
%                  getbyte(BP_mbx_buff,6));     % and button code (0..15).
%    end;
% 
% %
% % Stop communication and close bitpad channel
% %
% symbolic procedure L_CloseBitPad();
%   MBX_Close(BP_handle);    % That's all!
% %
% %-----------------------------------------------
% %
% 
% 